﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {

    // Defines configurations of lazors. 
    abstract class CFG : IDisposable{

        public Rect rect = new Rect(Screen.width / 2, Screen.height / 2, 200, 10);
        public bool docked = true;
        protected DoubleSwitch enabled;
		public bool Enabled { get { return enabled; } }
        protected DoubleSwitch enabledMain, enabledEVA, configuredEVA;
        public bool ConfiguredEVA {
            get { return configuredEVA; }
            set { 
                configuredEVA.set = value;
                swapConfig();
            }
        }
        
        public bool hasEvaSystem;
        protected LazorSystem system;
        protected HashSet<int> claimedLazors = new HashSet<int>();
        protected int claimedLazorsCount = 0;
        public int id;
        public string name;
        public string nameEVA;
        protected Color color;
        public Color Color {
            get { return color; }
        }
        protected Texture2D colorTexture;
        public Texture2D ColorTexture {
            get { return colorTexture; }
        }
        public int systemActive = 1;
        public int systemCount = 1;


        public CFG(LazorSystem system) {
            this.system = system;
            id = system.guiController.GenerateID();
        }

        virtual public void Dispose() {}

        public HashSet<int> GetClaimedLazors() {
            return claimedLazors;
        }

        virtual public bool UnclaimLazors(int i) {
			foreach(LazorAimable l in system.config.LazorCounterparts[i])
				if(l != null)
					l.Lazor.Off();
            if(claimedLazors.Remove(i)){
                claimedLazorsCount -= system.config.LazorCounterparts[i].Count;
                return true;
            }
            return false;
        }

        virtual public bool ClaimLazors(int i) {
			foreach(LazorAimable l in system.config.LazorCounterparts[i])
				if(l != null && enabled)
					l.Lazor.On();
            if (claimedLazors.Add(i)) {
                claimedLazorsCount += system.config.LazorCounterparts[i].Count;
                return true;
            }
            return false;
        }

        //optional methods
        virtual public void DrawGUI() { } //popups
        virtual public void DrawGUIbuttonLeader() { } //Button goes under targetwindow on main GUI
        virtual public void Inject_EVA() { } //does this lazor add to the EVA support system?
        virtual public void OnTargetSwitch() { } //called when target changes
        virtual public void OnEnabledFixedUpdate() { } //fixed update, system.lazoron && cfg.enabled
        virtual public void swapConfig() { }
        virtual public void OnEnable() { }
        virtual public void OnDisable() { }
        virtual public void Reset() { }
		virtual public void RemoveLazorRef(LazorAimable l) { }

		public int OnFixedUpdate() {
			if (enabled.In()) {
				OnEnable();
				foreach (int i in GetClaimedLazors()) //show lazor beams
					foreach (LazorAimable l in system.config.LazorCounterparts[i])
						if (l != null)
							l.Lazor.On();
				system.config.ActivateCFG(this);
			} else if (enabled.Out()) {
				OnDisable();
				foreach (int i in GetClaimedLazors()) //hide lazor beams
					foreach (LazorAimable l in system.config.LazorCounterparts[i])
						if (l != null)
							l.Lazor.Off();
				system.config.DeactivateCFG(this);
			} 
			int activeLazorCount = 0;
			if (enabled) {
				foreach (int i in GetClaimedLazors()) {
					if(system.config.LazorCounterparts.ContainsKey(i)){
						foreach (LazorAimable l in system.config.LazorCounterparts[i]) {
							/*if (l == null){
								system.config.UnRegister(l);
								continue;
							}*/
							if(l != null){
								ExecuteCommand(l);
								l.Lazor.SetRange(system.lazorRange);
								activeLazorCount++;
							}else
								MonoBehaviour.print("CFG.OnFixedUpdate : LazorCounterparts returned null");
						}
					}else
						MonoBehaviour.print("CFG.OnFixedUpdate : key not found in LazorCounterparts");
				}
				OnEnabledFixedUpdate();
			} 
			return activeLazorCount;
		}

        //required methods  
        abstract public void ExecuteCommand(LazorAimable lazor); //Execute command on each claimed lazors
        abstract public bool Detect(Dictionary<int, List<LazorAimable>> lazors); //Detect if the given list contains the correct lazors this configuration require.
        abstract public void DrawGUIbutton();
        abstract public void ContentGUI(int windowID);

        protected int ClaimedLazorNotHittingOwnVessel(){
            int count = 0;
            foreach (int i in claimedLazors) {
                foreach (LazorAimable l in system.config.LazorCounterparts[i]) {
                    if (!l.Lazor.LazorIsHittingOwnVessel)
                        count++;
                }
            }
            return count;
        }

        public HashSet<int> StdDetect(Color c, Dictionary<int, List<LazorAimable>> lazors) {
            claimedLazorsCount = 0;
            foreach (int i in lazors.Keys) {
                if (lazors[i][0].Lazor.GetColor() == c) {
                    claimedLazorsCount += lazors[i].Count;
                    return new HashSet<int>(new[] { i });
                }
            }
            return new HashSet<int>();
        }

        public HashSet<int> StdDetectAll(Color c, Dictionary<int, List<LazorAimable>> lazors) {
            claimedLazorsCount = 0;
            HashSet<int> foundLazors = new HashSet<int>();
            foreach (int i in lazors.Keys)
                if (lazors[i][0].Lazor.GetColor() == c) {
                    foundLazors.Add(i);
                    claimedLazorsCount += lazors[i].Count;
                }
            if(foundLazors.Count>0)
                return foundLazors;
            return new HashSet<int>();
        }

        public bool Claimable(Color c) {
            return c == color;
        }

        protected void onFlightStateLoadX(Dictionary<string, KSPParseable> parsedData, string identifier) {
            Util.LoadKSPParseableRect(parsedData, identifier + "_rect", ref rect);
            if (parsedData.ContainsKey(identifier + "_docked"))
                docked = parsedData[identifier + "_docked"].value_bool;
            if (parsedData.ContainsKey(identifier + "_enabled"))
                enabled.set = parsedData[identifier + "_enabled"].value_bool;
        }

        protected void onFlightStateSaveX(Dictionary<string, KSPParseable> partDataCollection, string identifier) {
            Util.SaveKSPParseableRect(partDataCollection, identifier + "_rect", rect);
            partDataCollection.Add(identifier + "_docked", new KSPParseable((bool)docked, KSPParseable.Type.BOOL));
            partDataCollection.Add(identifier + "_enabled", new KSPParseable((bool)enabled, KSPParseable.Type.BOOL));
        }
        abstract public void onFlightStateLoad(Dictionary<string, KSPParseable> parsedData);
        abstract public void onFlightStateSave(Dictionary<string, KSPParseable> partDataCollection);
    }
}
