﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;
using System.Timers;

namespace Romfarer {
    class CFG_cyan : CFG, IDisposable {

        private bool turnedOff;
        private FlightCtrlState ctrlState;


        public CFG_cyan(LazorSystem system): base(system) {
            ctrlState = new FlightCtrlState();
            name = "?";
            new Color(0, 0.5f, 0.5f, 1);
            colorTexture = Util.getTexture(color);
        }

        override public bool Detect(Dictionary<int, List<LazorAimable>> lazors) {
            claimedLazors = StdDetect(new Color(0,0.5f,0.5f,1), lazors);
            return claimedLazors.Count > 0;
        }

        override public void ExecuteCommand(LazorAimable lazor) {
            if (lazor != null)
                lazor.Lazor.Aim(lazor.transform.up - (lazor.transform.position - lazor.vessel.mainBody.position).normalized);
        }

        override public void DrawGUIbutton() {
            enabled = GUILayout.Toggle(enabled, "Stabilize", Util.StdStyle, GUILayout.ExpandWidth(true));
            if (enabled) {
                if (turnedOff) {
                    //system.upAxis = Romfarer.LazorSystem.Axis.gUp;
                    turnedOff = false;
                    FlightInputHandler.OnFlyByWire += new FlightInputHandler.FlightInputCallback(fly);
                    MonoBehaviour.print("[Stabilizer On]");
                }
            } else {
                if (!turnedOff) {
                    ctrlState.roll = 0;
                    ctrlState.yaw = 0;
                    ctrlState.pitch = 0;
                    system.vessel.rootPart.propagateControlUpdate(ctrlState);
                    turnedOff = true;
                    FlightInputHandler.OnFlyByWire -= new FlightInputHandler.FlightInputCallback(fly);
                    MonoBehaviour.print("[Stabilizer Off]");
                }
            }
        }

        //To-DO
        override public void ContentGUI(int windowID) {

        }


        private void fly(FlightCtrlState s) {
            /*if (system.upAxis == Romfarer.LazorSystem.Axis.gUp) {
                if (!system.IsHorizontal) {
                    s.roll += system.vessel.angularVelocity.y;
                    Vector3 vOffset = system.vessel.transform.InverseTransformDirection(system.GetUpAxis());
                    s.yaw += Mathf.Clamp(system.vessel.angularVelocity.z + system.vessel.angularMomentum.z / 5 + vOffset.x, -1, 1);
                    s.pitch += Mathf.Clamp(system.vessel.angularVelocity.x + system.vessel.angularMomentum.x / 5 - vOffset.z, -1, 1);
                    //MonoBehaviour.print(vOffset + " " + s.roll.ToString("f2") + " " + s.yaw.ToString("f2") + " " + s.pitch.ToString("f2"));
                } else {
                    float o = -1;
                    if (Vector3.Angle(system.vessel.transform.right, -system.GetUpAxis()) > 90)
                        o = 1;
                    float angle = Vector3.Angle(system.vessel.transform.forward, -system.GetUpAxis()) * Mathf.Deg2Rad * o;
                    s.roll += Mathf.Clamp(angle, -1, 1);  //system.vessel.angularVelocity.y - 
                    //MonoBehaviour.print(angle.ToString("f2") + " " + s.roll.ToString("f2"));
                }
            }*/
        }

        override public void Dispose() {
            FlightInputHandler.OnFlyByWire -= new FlightInputHandler.FlightInputCallback(fly);
        }


    }
}
