﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;
using System.Timers;

namespace Romfarer {
    class CFG_red : CFG, IDisposable {

        private GUItargets guiTargets;
        private DoubleSwitch hEnabled;
        private DoubleSwitch vEnabled;
        private FlightCtrlState ctrlStateRemote;
        
        //airplanes
        private float bearing;
        private float yaw;
        private float mark;
        private float roll;
        private DoubleSwitch toggleBearing;
        private DoubleSwitch toggleYaw;
        private DoubleSwitch toggleMark;
        private DoubleSwitch toggleRoll;
        private bool toggleFollowLeader;
        private bool toggleDampen = true;
        private bool toggleAdvanced;

        private bool toggleAltitude;
        double cruiseAlt = 1000d;
        float maxClimb = 25f;
        private LinkedList<float> altcoeff;
        private float dPDinv = 0f;
        private float pitch = 0f;
        double lastMark = 0d;
        float tuner = 0f;
        UtilAverage avgAcc;
        UtilAverage avgMark;
        UtilAverage avgMarkAcc;
        UtilAverage avgVspd;
        private float lastTime = 0f;
        private float diff = 0;

        private float sideSlipAngle;
        Vector3d surfacePos = Vector3.zero;
        private float actualBearing_Velocity_last;
        private UtilAverage TurnRate = new UtilAverage(25);
        private float actualBearing_Velocity; //bearing of velocity vector
        private float velocityOffset;

        //great circle
        private DoubleSwitch toggle_GS_travel;
        private bool toggleMapSelect;
        private double GS_destination_LATITUDE = 0;
        private double GS_destination_LONGITUDE = 0;
        private double GS_destination_distance = -5000;
        private double GS_destination_bearing = -5000;
        private bool GS_destination_set;
        private double fLon = 0d;
        private double fLat = 0d;
        private string fLonSex = "";
        private string fLatSex = "";

		private PIDclamp pid_plane_roll_manual { get { return LazorGlobals.GLOBALCONFIG_pid_plane_roll_manual; }
												 set { LazorGlobals.GLOBALCONFIG_pid_plane_roll_manual = value; } }
		private PIDclamp pid_plane_roll_auto { get { return LazorGlobals.GLOBALCONFIG_pid_plane_roll_auto; }
											   set { LazorGlobals.GLOBALCONFIG_pid_plane_roll_auto = value; } }
		private PIDclamp pid_plane_yaw_auto { get { return LazorGlobals.GLOBALCONFIG_pid_plane_yaw_auto; }
											  set { LazorGlobals.GLOBALCONFIG_pid_plane_yaw_auto = value; } }
		private PIDclamp pid_plane_pitch { get { return LazorGlobals.GLOBALCONFIG_pid_plane_pitch; }
										   set { LazorGlobals.GLOBALCONFIG_pid_plane_pitch = value; } }
        /*private PIDclamp pid_plane_roll_manual = new PIDclamp("Roll manual", 0.25f, 0.001f, 0.085f, 90f);       
        private PIDclamp pid_plane_roll_auto = new PIDclamp("Roll auto", 0.25f, 0.001f, 0.095f, 45f);
        private PIDclamp pid_plane_yaw_auto = new PIDclamp("Yaw auto", 0.102f, 0.055f, 0.06f, 1f);
        private PIDclamp pid_plane_pitch = new PIDclamp("Pitch auto/manual", 0.3f, 0.4f, 0.021f, 1f);*/

		private PIDclamp pid_rocket_yaw { get { return LazorGlobals.GLOBALCONFIG_pid_rocket_yaw; }
			set { LazorGlobals.GLOBALCONFIG_pid_rocket_yaw = value; } }
		private PIDclamp pid_rocket_pit { get { return LazorGlobals.GLOBALCONFIG_pid_rocket_pit; }
			set { LazorGlobals.GLOBALCONFIG_pid_rocket_pit = value; } }
		private PIDclamp pid_rocket_rol { get { return LazorGlobals.GLOBALCONFIG_pid_rocket_rol; }
			set { LazorGlobals.GLOBALCONFIG_pid_rocket_rol = value; } }
		private PIDclamp pid_rocket_lr { get { return LazorGlobals.GLOBALCONFIG_pid_rocket_lr; }
			set { LazorGlobals.GLOBALCONFIG_pid_rocket_lr = value; } }
		private PIDclamp pid_rocket_ud { get { return LazorGlobals.GLOBALCONFIG_pid_rocket_ud; }
			set { LazorGlobals.GLOBALCONFIG_pid_rocket_ud = value; } }
		private PIDclamp pid_rocket_fb { get { return LazorGlobals.GLOBALCONFIG_pid_rocket_fb; }
			set { LazorGlobals.GLOBALCONFIG_pid_rocket_fb = value; } }
        //rockets
		/*private PIDclamp pid_rocket_yaw = new PIDclamp("Torque yaw", 8.3f, 0.3f, 0.5f, 1f);
		private PIDclamp pid_rocket_pit = new PIDclamp("Torque pitch", 8.3f, 0.3f, 0.5f, 1f);
		private PIDclamp pid_rocket_rol = new PIDclamp("Torque roll", 6.0f, 0.25f, 0.025f, 1f);

		private PIDclamp pid_rocket_lr = new PIDclamp("Translate left/right", 4.3f, 0.01f, 0.02f, 1f);
		private PIDclamp pid_rocket_ud = new PIDclamp("Translate up/down", 4.3f, 0.01f, 0.02f, 1f);
		private PIDclamp pid_rocket_fb = new PIDclamp("Translate fwd/back", 4.3f, 0.01f, 0.02f, 1f);*/

        private enum RocketModes { 
			Free = 0, 
            RSAS = 1, 
            CurrentDirection = 2, 

            Prograde = 3, 
            Retrograde = 4,

            ProGee = 5, 
            RetroGee = 6,

            Vplus = 7,
            Vminus = 8,

            Nplus = 9,
            Nminus = 10,

            Maneuver = 11,

			TargetDirPlus = 12,
			TargetDirMinus = 13,

			TargetVelPlus = 14,
			TargetVelMinus = 15,

			TargetRelDirPlus =16,
			TargetRelDirMinus = 17,

			TargetRelVelPlus = 18,
			TargetRelVelMinus = 19,
        };
		private enum RocketModesThrust { 
			Free = 0,
			RCSBrake = 1,
			RCSBrakeRelVel = 2,
		};
        
        private bool lockDirectionSet;
        private Vector3 vectorCustom = Vector3.zero;
        private Vector3 vectorLock = Vector3.zero;
        private Ray ray;
        private float torque;
        private bool comingFromEVA;
        private float throttleEVA;
		private Vector3 target_angle = Vector3.zero;
		private Vector3 target_angle_thrust = Vector3.zero;
		private RocketModes rocketMode;
		EnumToggleGroup<RocketModes> toggleRocketModes = new EnumToggleGroup<RocketModes>(RocketModes.Free);
		private RocketModesThrust rocketModeThrust;
		EnumToggleGroup<RocketModesThrust> toggleRocketModesThrust = new EnumToggleGroup<RocketModesThrust>(RocketModesThrust.Free);
		private bool rcsFwd;
		private bool rcsBack;

        public CFG_red(LazorSystem system) : base(system) {
            guiTargets = system.guiTargets;
            roll = 0;
            mark = 0;
            bearing = 0;
            rocketMode = RocketModes.Free;
            altcoeff = new LinkedList<float>();
            avgAcc = new UtilAverage(10);
            avgMark = new UtilAverage(10);
            avgMarkAcc = new UtilAverage(10);
            avgVspd = new UtilAverage(10);
            ctrlStateRemote = new FlightCtrlState();
            /*if (system.vessel.rootPart is CommandPod)
                torque = ((CommandPod)system.vessel.rootPart).rotPower - ((CommandPod)system.vessel.rootPart).rotPower * (1 / 10);
            else
                torque = 0.000001f;*/
			torque = 5;

            name = "Lazor Guided Flight";
            color = Color.red;
            colorTexture = Util.LightTexture(Util.getTexture(new Color(color.r, color.g, color.b, 0.9f)), 0.6f);
        }

        override public bool Detect(Dictionary<int, List<LazorAimable>> lazors) {
            claimedLazors = StdDetect(Color.red, lazors);
            return claimedLazors.Count>0;
        }

		override public void Dispose() {
			system.vessel.OnFlyByWire -= new FlightInputCallback(fly);
			system.vessel.OnFlyByWire -= new FlightInputCallback(flyRocket);
		}

        override public void OnEnabledFixedUpdate() {
            system.RequestEnergyDelta(0.15d);

            if (system.IsHorizontal) {
                Quaternion orientationVelocity = Quaternion.Inverse(Quaternion.Euler(90, 0, 0) * Quaternion.Inverse(system.vessel.transform.rotation) *
                    Quaternion.LookRotation(system.rotatingFrameVelocity, system.vessel.upAxis));

                if (system.rotatingFrameVelocity.magnitude > 20) {
                    velocityOffset = -Util.TwoPiToSignedAngle(orientationVelocity.eulerAngles.y);
                    sideSlipAngle = Util.AngleAroundNormal(system.LocalForward, system.rotatingFrameVelocity, system.LocalUp);
                } else {
                    velocityOffset = 0;
                    sideSlipAngle = 0;
                }
                actualBearing_Velocity = (float)Util.NormalizeAngle(system.actualBearing + velocityOffset); //bearing of velocity vector

                //Change of velocity vector.
                TurnRate.Average((actualBearing_Velocity - actualBearing_Velocity_last) / Time.deltaTime);
                actualBearing_Velocity_last = actualBearing_Velocity;

                if (toggle_GS_travel) {
                    GS_destination_bearing = Util.GreatCircleBearing(system.LATITUDE, system.LONGITUDE, fLat, fLon);
                    this.bearing = (float)GS_destination_bearing;
                    GS_destination_distance = Util.GreatCircleDistance(system.LATITUDE, system.LONGITUDE, GS_destination_LATITUDE, GS_destination_LONGITUDE, system.vessel.mainBody.Radius);   
                }

                //leave g-force tracking for later
                /*Vector3 gee = FlightGlobals.getGeeForceAtPosition(system.vessel.transform.position);
                if (Vector3.Dot(system.averageAcceleration, system.LocalUp) < 0)
                    TC = Util.AngleAroundNormal(system.averageAcceleration + gee, -system.LocalUp, system.LocalForward);
                else
                    TC = Util.AngleAroundNormal(system.averageAcceleration - gee, system.LocalUp, -system.LocalForward);*/
            }
        }

		//What the lazor will do, purely graphical
		override public void ExecuteCommand(LazorAimable lazor) {
			//for airplanes
			if (system.IsHorizontal) {
				ray = new Ray(lazor.Lazor.GetOrigin(), system.vessel.transform.up);
				if (toggleFollowLeader && system.toggleFreeAim){
					ray.direction = system.Leader.GetAim();
				//else if (toggleFollowLeader && !system.toggleFreeAim) { 
				} else if(toggleBearing || toggle_GS_travel){
					ray.direction = Quaternion.LookRotation(system.vessel.transform.up, (system.vessel.rootPart.transform.position - system.vessel.mainBody.position).normalized) *
						Quaternion.Euler(system.actualMark - mark, diff, 0) * Vector3.forward;
				}else //if(toggleYaw)
					ray.direction = Quaternion.LookRotation(system.vessel.transform.up, (system.vessel.rootPart.transform.position - system.vessel.mainBody.position).normalized) *
						Quaternion.Euler(system.actualMark - mark, yaw / 7, 0) * Vector3.forward;
				/*else
					ray.direction = Quaternion.LookRotation(system.vessel.transform.up, (system.vessel.rootPart.transform.position - system.vessel.mainBody.position).normalized) *
						Quaternion.Euler(0, 0, 0) * Vector3.forward;*/
				lazor.Lazor.Aim(ray.direction);
			}
			//for rockets
			else {
				lazor.Lazor.Aim(target_angle);
				if (rocketMode != RocketModes.CurrentDirection && lockDirectionSet) {
					vectorCustom = Vector3.zero;
					lockDirectionSet = false;
				}
			}
		}

        override public void DrawGUIbutton() {
            if (system.IsHorizontal) { //for airplanes
                hEnabled.set = system.guiSubSystems.SubsystemButton(hEnabled, name, color, out docked, ContentGUI, 0);
                vEnabled.set = false;
            } else { //For ROCKETS
                vEnabled.set = system.guiSubSystems.SubsystemButton(vEnabled, name, color, out docked, ContentGUI, 0);
                hEnabled.set = false;
            }
            enabled.set = hEnabled || vEnabled;
        }

        override public void OnEnable() {
            if (system.IsHorizontal){ //for airplanes
                hEnabled.set = true;
                vEnabled.set = false;
            } else { // For rockets
                hEnabled.set = false;
                vEnabled.set = true;
            }
            OnOff();
        }

        override public void OnDisable() {
            hEnabled.set = false;
            vEnabled.set = false;
            toggleFollowLeader = false;
            OnOff();
        }

        private void OnOff() {
            //for airplanes
            if (hEnabled.In())
                PlaneEnable();
            else if (hEnabled.Out())
                PlaneDisable();

            // For rockets
            if (vEnabled.In())
                RocketEnable();
            else if (vEnabled.Out())
                RocketDisable();
        }

        public override void Reset() {
            if (hEnabled) {
                PlaneDisable();
                PlaneEnable();
            } else if (vEnabled) {
                RocketDisable();
                RocketEnable();
            }
        }

        private void PlaneEnable() {
            system.vessel.OnFlyByWire += fly;
            toggleBearing.set = false;
            toggle_GS_travel.set = false;
            toggleYaw.set = true;
            toggleRoll.set = true;
            toggleMark.set = true;
            PlaneOnOff();
            MonoBehaviour.print("[Plane LGF On]");
        }

        private void PlaneDisable() {
            system.vessel.OnFlyByWire -= fly;
            toggleBearing.set = false;
            toggle_GS_travel.set = false;
            toggleYaw.set = false;
            toggleRoll.set = false;
            toggleMark.set = false;
            system.guiOverlays.toggleLGF = false;
            system.guiOverlays.toggleLGFnewBearing = false;
            PlaneOnOff();
            MonoBehaviour.print("[Plane LGF Off]");
        }

        private void RocketEnable() {
            system.vessel.OnFlyByWire += flyRocket;
            MonoBehaviour.print("[Rocket LGF On]");
        }

        private void RocketDisable() {
            system.vessel.OnFlyByWire -= flyRocket;
            rocketMode = RocketModes.Free;
            MonoBehaviour.print("[Rocket LGF Off]");
        }

        private void PlaneOnOff() {
            system.sysLock_LEADER_ignorehits = toggleFollowLeader;

            if (toggleYaw.In())
                this.yaw = FlightUIController.fetch.stgYaw.Value * 100;
            else if (toggleYaw.Out())
                this.yaw = 0;

            if (toggleRoll.In())
                this.roll = system.actualRoll;
            else if (toggleRoll.Out())
                this.roll = 0;

            if (toggleMark.In())
                this.mark = system.actualMark;
            else if (toggleMark.Out())
                this.mark = 0;

            if (toggleBearing.In()) {
                this.bearing = system.actualBearing;
                toggle_GS_travel.set = false;
            } else if (toggleBearing.Out())
                this.bearing = 0;

            if (toggle_GS_travel.In())
                toggleBearing.set = false;
            else if (toggle_GS_travel.Out()) { }
        }


		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// GUI
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


        override public void ContentGUI(int windowID) {
            if (system.IsHorizontal) {//for airplanes
                PlaneOnOff();
                docked = system.guiSubSystems.Header(docked, rect, hEnabled, ref hEnabled);
                vEnabled.set = false;

                // Yaw, Pitch, Roll readouts
                GUILayout.BeginVertical(Util.ListItemBG_compact);
                GUILayout.BeginHorizontal(Util.Space_Compact);
                GUILayout.Label("Yaw: ", Util.StdLabelSmallCustom(LazorSystem.COLOR_list_identifier), GUILayout.Width(29));
                GUILayout.Label((FlightUIController.fetch.stgYaw.Value * 100).ToString("f0") + "%", Util.StdLabelSmallCustom(Color.white), GUILayout.Width(35));
                GUILayout.Label("Pit: ", Util.StdLabelSmallCustom(LazorSystem.COLOR_list_identifier), GUILayout.Width(26));
                GUILayout.Label((FlightUIController.fetch.stgPitch.Value * 100).ToString("f0") + "%", Util.StdLabelSmallCustom(Color.white), GUILayout.Width(35));
                GUILayout.Label("Rol: ", Util.StdLabelSmallCustom(LazorSystem.COLOR_list_identifier), GUILayout.Width(26));
                GUILayout.Label((FlightUIController.fetch.stgRoll.Value * 100).ToString("f0") + "%", Util.StdLabelSmallCustom(Color.white), GUILayout.Width(35));
                GUILayout.EndHorizontal();
                GUILayout.EndVertical();

                GUILayout.BeginHorizontal(Util.Space_Compact);
                GUILayout.Label("Current Bearing: ", Util.StdLabelSmallCustom(LazorSystem.COLOR_list_identifier));
                GUILayout.Label(system.actualBearing.ToString("f2"));
                GUILayout.Label(" °", Util.StdLabelSmallCustom(LazorSystem.COLOR_list_identifier));
                GUILayout.EndHorizontal();

                GUILayout.BeginHorizontal(Util.Space_Compact);
                GUILayout.Label("Turn rate: ", Util.StdLabelSmallCustom(LazorSystem.COLOR_list_identifier));
                GUILayout.Label(TurnRate.value.ToString("f2"));
                GUILayout.Label(" °/sec", Util.StdLabelSmallCustom(LazorSystem.COLOR_list_identifier));
                GUILayout.EndHorizontal();

                GUILayout.BeginHorizontal(Util.Space_Compact);
                GUILayout.Label("Side slip: ", Util.StdLabelSmallCustom(LazorSystem.COLOR_list_identifier));
                GUILayout.Label(sideSlipAngle.ToString("f2"));
                GUILayout.Label(" °", Util.StdLabelSmallCustom(LazorSystem.COLOR_list_identifier));
                GUILayout.EndHorizontal();

                GUILayout.HorizontalSlider((float)sideSlipAngle, -5f, 5f, Util.Slider_balance, Util.Slider_balance_button, GUILayout.Width(190), GUILayout.Height(24));

                GUILayout.Space(5);

                // Bearing, Mark, Roll readouts and setters
                GUI.enabled = !toggleBearing && !toggle_GS_travel;

                GUILayout.BeginHorizontal();
                GUILayout.Label("Yaw: ", GUILayout.ExpandWidth(true));
                GUILayout.Label((FlightUIController.fetch.stgYaw.Value * 100).ToString("f2") + " %", GUILayout.Width(58));
                this.yaw = float.Parse(GUILayout.TextField(this.yaw.ToString("f2"), GUILayout.Width(58)).Replace("\n", "").Replace("\r", ""));
                toggleYaw.set = GUILayout.Toggle(toggleYaw, "", GUILayout.Width(20), GUILayout.Height(20));
                GUILayout.EndHorizontal();

                GUILayout.BeginHorizontal();
                GUILayout.Label("Roll: ", GUILayout.ExpandWidth(true));
                GUILayout.Label(system.actualRoll.ToString("f2") + " °", GUILayout.Width(58));
                this.roll = float.Parse(GUILayout.TextField(this.roll.ToString("f2"), GUILayout.Width(58)).Replace("\n", "").Replace("\r", ""));
                toggleRoll.set = GUILayout.Toggle(toggleRoll, "", GUILayout.Width(20), GUILayout.Height(20));
                GUILayout.EndHorizontal();

                GUI.enabled = !toggleAltitude;

                GUILayout.BeginHorizontal();
                GUILayout.Label("Pitch: ", GUILayout.ExpandWidth(true));
                GUILayout.Label(system.actualMark.ToString("f2") + " °", GUILayout.Width(58));
                this.mark = float.Parse(GUILayout.TextField(this.mark.ToString("f2"), GUILayout.Width(58)).Replace("\n", "").Replace("\r", ""));
                toggleMark.set = GUILayout.Toggle(toggleMark, "", GUILayout.Width(20), GUILayout.Height(20));
                GUILayout.EndHorizontal();

                GUI.enabled = true;

                GUILayout.BeginHorizontal();
                toggleFollowLeader = GUILayout.Toggle(toggleFollowLeader, "Mouse Look");

                //Dampen input
                toggleDampen = GUILayout.Toggle(toggleDampen, "Dampen Input");
                GUILayout.EndHorizontal();

                GUILayout.Space(5);

                GUILayout.BeginHorizontal();
                GUILayout.FlexibleSpace();
                GUILayout.Label("- - - - -", Util.StdLabelSmallCustom(Color.white, TextAnchor.MiddleCenter));
                GUILayout.FlexibleSpace();
                GUILayout.EndHorizontal();


                //////////////////////////////
                // PITCH / ALTITUDE Control
                //////////////////////////////


                // Altitude Control
                GUILayout.Label("Pitch and Altitude Control: ", Util.StdLabelSmallCustom(Color.yellow));
                toggleAltitude = GUILayout.Toggle(toggleAltitude, "Automatic Pitch");
                GUILayout.Space(3);
                GUILayout.BeginHorizontal();
                GUILayout.Label("Altitude: ");
                //GUILayout.TextField(system.vessel.orbit.altitude.ToString("f2"));
                cruiseAlt = Double.Parse(GUILayout.TextField(cruiseAlt.ToString("f0"), 5).Replace("\n", "").Replace("\r", ""));

                GUILayout.Label("Max °: ");
                maxClimb = float.Parse(GUILayout.TextField(maxClimb.ToString("f0"), GUILayout.Width(45)).Replace("\n", "").Replace("\r", ""));
                GUILayout.EndHorizontal();


                GUILayout.Space(3);
                GUILayout.BeginHorizontal();
                GUILayout.FlexibleSpace();
                GUILayout.Label("- - - - -");
                GUILayout.FlexibleSpace();
                GUILayout.EndHorizontal();


                //////////////////////////////
                // YAW / ROLL Control
                //////////////////////////////

                
                //Angular velocity along horizon
                GUILayout.Label("Yaw/Roll and Bearing Control: ", Util.StdLabelSmallCustom(Color.yellow));

                toggleBearing.set = GUILayout.Toggle(toggleBearing, "Fly towards bearing");
                GUILayout.BeginHorizontal();
                GUILayout.Label("Bearing: ", GUILayout.ExpandWidth(true));
                GUILayout.Label(system.actualBearing.ToString("f2") + " °", GUILayout.Width(58));
                this.bearing = float.Parse(GUILayout.TextField(this.bearing.ToString("f2"), GUILayout.Width(58)).Replace("\n", "").Replace("\r", ""));
                GUILayout.EndHorizontal();


                // Travel to coordinate using great circle trigonometry
                GUILayout.BeginHorizontal();
                toggle_GS_travel.set = GUILayout.Toggle(toggle_GS_travel, "Fly to coordinate");
                GUILayout.FlexibleSpace();
                toggleMapSelect = GUILayout.Toggle(toggleMapSelect, !toggleMapSelect ? "Select on map" : "Cancel select", Util.Button_Toggle);
                GUILayout.EndHorizontal();

                GUILayout.BeginHorizontal();
                GUILayout.BeginVertical();
                GUILayout.Label("Latitude: ");
                fLat = Double.Parse(GUILayout.TextField(fLat.ToString("f4")).Replace("\n", "").Replace("\r", ""));
                GUILayout.Label(fLatSex);
                GUILayout.EndVertical();
                GUILayout.BeginVertical();
                GUILayout.Label("Longitude: ");
                fLon = Double.Parse(GUILayout.TextField(fLon.ToString("f4")).Replace("\n", "").Replace("\r", ""));
                GUILayout.Label(fLonSex);
                GUILayout.EndVertical();
                GUILayout.EndHorizontal();

                //Travel to coordinate
                if (toggleMapSelect) {
                    if (MapView.MapIsEnabled) {
                        Ray ray = system.planetariumCamera.camera.ScreenPointToRay(Input.mousePosition);
                        ray.origin = ScaledSpace.ScaledToLocalSpace(ray.origin);
                        Vector3d origin = ray.origin - system.vessel.mainBody.position;

                        string latitudeSexagesimal = "";
                        string longitudeSexagesimal = "";
                        if (PQS.LineSphereIntersection(origin, ray.direction, system.vessel.mainBody.Radius, out surfacePos)) {
                            surfacePos += system.vessel.mainBody.position;
                            GS_destination_LATITUDE = system.vessel.mainBody.GetLatitude(surfacePos);
                            GS_destination_LONGITUDE = system.vessel.mainBody.GetLongitude(surfacePos);
                            latitudeSexagesimal = Util.DegToSex(GS_destination_LATITUDE) + (GS_destination_LATITUDE > 0 ? " N" : " S");
                            longitudeSexagesimal = Util.DegToSex(GS_destination_LONGITUDE) + (GS_destination_LONGITUDE > 0 ? " E" : " W");
                            double newBearing = Util.GreatCircleBearing(system.LATITUDE, system.LONGITUDE, GS_destination_LATITUDE, GS_destination_LONGITUDE);
                            double newDistance = Util.GreatCircleDistance(system.LATITUDE, system.LONGITUDE, GS_destination_LATITUDE, GS_destination_LONGITUDE, system.vessel.mainBody.Radius);
                            system.guiOverlays.FeedLGF(GS_destination_LATITUDE.ToString("f4"), latitudeSexagesimal, GS_destination_LONGITUDE.ToString("f4"), longitudeSexagesimal, newBearing, newDistance);
                            if (Input.GetMouseButtonDown(0)) {
                                system.guiOverlays.FeedLGFTarget(surfacePos, GS_destination_LATITUDE, GS_destination_LONGITUDE);
                                fLat = GS_destination_LATITUDE;
                                fLon = GS_destination_LONGITUDE;
                                fLatSex = latitudeSexagesimal;
                                fLonSex = longitudeSexagesimal;
                                system.guiOverlays.toggleLGF = false;
                                GS_destination_set = true;
                                toggleMapSelect = false;
                            }
                            system.guiOverlays.toggleLGFnewBearing = true;
                        } else {
                            system.guiOverlays.toggleLGFnewBearing = false;
                        }
                    } else {
                        MapView.EnterMapView();
                    }
                } else { 
                    system.guiOverlays.toggleLGF = false;
                    system.guiOverlays.toggleLGFnewBearing = false;
                }
                if (GS_destination_set) {
                    GUILayout.BeginHorizontal();
                    Util.DistanceFormatted(GS_destination_distance, Color.white, "Distance remaining: ");
                    GUILayout.EndHorizontal();
                    GUILayout.Label("Great cirle bearing: " + GS_destination_bearing.ToString("f2") + "°");
                }


                //////////////////////////////
                // ADVANCED OPTIONS / PID
                //////////////////////////////

				#region options plane
                toggleAdvanced = GUILayout.Toggle(toggleAdvanced, "Advanced Controller Options");
                if (toggleAdvanced) {
                    /*pidRoll_manual.DebugGui("Roll manual");
                    pidRoll_auto.DebugGui("Roll auto");
                    pidYaw_auto.DebugGui("Yaw auto");
                    pidPitch.DebugGui("Pitch auto/manual");*/
					PIDclamp.DebugGuiPanel(pid_plane_roll_manual, pid_plane_roll_auto, pid_plane_yaw_auto, pid_plane_pitch);
                }
				#endregion

                //////////////////////////////
                // KEYBOARD INPUT
                //////////////////////////////


                if (!FlightGlobals.ActiveVessel.isEVA) {
                    bool pm = FlightInputHandler.fetch.precisionMode;

                    if (toggleBearing) {
                        if (Input.GetKey(GameSettings.YAW_LEFT.primary))
                            this.bearing -= pm ? 0.01f : 0.1f;
                        else if (Input.GetKey(GameSettings.YAW_RIGHT.primary))
                            this.bearing += pm ? 0.01f : 0.1f;
                        this.yaw = FlightUIController.fetch.stgYaw.Value;
                        this.roll = system.actualRoll;
                    } else {
                        if (!toggle_GS_travel) {
                            this.bearing = system.actualBearing;
                            if (toggleYaw) {
                                if (Input.GetKey(GameSettings.YAW_LEFT.primary))
                                    this.yaw -= pm ? 0.02f : 0.2f;
                                else if (Input.GetKey(GameSettings.YAW_RIGHT.primary))
                                    this.yaw += pm ? 0.02f : 0.2f;
                            }
                            if (toggleRoll) {
                                if (Input.GetKey(GameSettings.ROLL_LEFT.primary))
                                    this.roll -= pm ? 0.01f : 0.1f;
                                else if (Input.GetKey(GameSettings.ROLL_RIGHT.primary))
                                    this.roll += pm ? 0.01f : 0.1f;
                            }
                        }else{
                            this.yaw = FlightUIController.fetch.stgYaw.Value;
                            this.roll = system.actualRoll;
                        }
                    }

                    if (!toggleAltitude) {
                        if (toggleMark) {
                            if (Input.GetKey(GameSettings.PITCH_DOWN.primary))
                                this.mark -= pm ? 0.0075f : 0.075f;
                            else if (Input.GetKey(GameSettings.PITCH_UP.primary))
                                this.mark += pm ? 0.0075f : 0.075f;
                        }else
                            this.mark = system.actualMark;
                    }
                }

                //Fix input
                if (this.bearing >= 360)
                    this.bearing = this.bearing - 360;
                else if (this.bearing < 0)
                    this.bearing = 360 + this.bearing;
                else if (this.bearing > 720 || this.bearing < -720)
                    this.bearing = 0;

                if (this.mark > 90)
                    this.mark = 90;
                else if (this.mark < -90)
                    this.mark = -90;

                if (this.roll > 90)
                    this.roll = 90;
                else if (this.roll < -90)
                    this.roll = -90;

                if (this.yaw > 100)
                    this.yaw = 100;
                else if (this.yaw < -100)
                    this.yaw = -100;

            } else { // For ROCKETS
                docked = system.guiSubSystems.Header(docked, rect, vEnabled, ref vEnabled);
                hEnabled.set = false;

				GUILayout.BeginHorizontal(Util.ListItemBG_compact);
                switch (rocketMode) {
                    case RocketModes.Prograde:
                        GUILayout.Label("Direction of rotating frame");
                        target_angle = system.vessel.mainBody.getRFrmVel(system.vessel.transform.position).normalized;
                        break;
                    case RocketModes.Retrograde:
                        GUILayout.Label("Opposite dir. rotating frame");
                        target_angle = system.vessel.mainBody.getRFrmVel(system.vessel.transform.position).normalized * -1;
                        break;
                    case RocketModes.ProGee:
                        GUILayout.Label("Direction of gravity");
                        target_angle = (system.vessel.mainBody.position - system.vessel.transform.position).normalized;
                        break;
                    case RocketModes.RetroGee:
                        GUILayout.Label("Opposite direction of gravity");
                        target_angle = (system.vessel.transform.position - system.vessel.mainBody.position).normalized;
                        break;
                    case RocketModes.Vplus:
                        GUILayout.Label("Direction of Velocity");
                        target_angle = system.vessel.orbit.GetVel().normalized;
                        break;
                    case RocketModes.Vminus:
                        GUILayout.Label("Opposite dir. of Velocity");
                        target_angle = system.vessel.orbit.GetVel().normalized * -1;
                        break;
                    case RocketModes.Nplus:
                        GUILayout.Label("Orbit normal '+' inclination");
                        target_angle = Vector3.Cross(system.vessel.orbit.GetVel(), system.vessel.mainBody.position - system.vessel.transform.position).normalized * -1;
                        break;
                    case RocketModes.Nminus:
                        GUILayout.Label("Orbit normal '-' inclination");
                        target_angle = Vector3.Cross(system.vessel.orbit.GetVel(), system.vessel.mainBody.position - system.vessel.transform.position).normalized;
                        break;
                    case RocketModes.RSAS:
                        GUILayout.Label("Romfarer SAS");
                        target_angle = system.LocalUp;
                        break;
                    case RocketModes.CurrentDirection:
                        GUILayout.Label("Current direction");
                        if (!lockDirectionSet) {
                            target_angle = vectorCustom = system.vessel.transform.up;
                            lockDirectionSet = true;
                        }
                        break;
                    case RocketModes.Free:
                        GUILayout.Label("Moving freely");
                        target_angle = system.LocalUp;
                        break;
                    case RocketModes.TargetDirPlus:
                        GUILayout.Label("Targets heading +");
                        if (system.Target != null)
                            target_angle = vectorLock = system.Target.transform.up;
                        else
							rocketMode = RocketModes.Free;
                        break;
					case RocketModes.TargetDirMinus:
						GUILayout.Label("Targets heading -");
						if (system.Target != null)
							target_angle = vectorLock = -system.Target.transform.up;
						else
							rocketMode = RocketModes.Free;
						break;

					case RocketModes.TargetRelDirPlus:
						GUILayout.Label("Face target +");
						if (system.Target != null)
							target_angle = vectorLock = system.Target.transform.position - system.vessel.transform.position;
						else
							rocketMode = RocketModes.Free;
						break;
					case RocketModes.TargetRelDirMinus:
						GUILayout.Label("Face target -");
						if (system.Target != null)
							target_angle = vectorLock = system.vessel.transform.position - system.Target.transform.position;
						else
							rocketMode = RocketModes.Free;
						break;


                    case RocketModes.Maneuver:
                        GUILayout.Label("Maneuver Node");
                        if (system.vessel.patchedConicSolver.maneuverNodes.Count > 0)
                            target_angle = system.vessel.patchedConicSolver.maneuverNodes[0].GetBurnVector(system.vessel.orbit);
                        else
                            rocketMode = RocketModes.Free;
                        break;

					case RocketModes.TargetRelVelPlus:
						GUILayout.Label("Relative Velocity +");
						if (system.Target != null && system.Target is ITargetable)
							target_angle = (system.Target as ITargetable).GetObtVelocity() - system.vessel.GetObtVelocity();
						else
							rocketMode = RocketModes.Free;
						break;

					case RocketModes.TargetRelVelMinus:
						GUILayout.Label("Relative Velocity -");
						if (system.Target != null && system.Target is ITargetable)
							target_angle =  system.vessel.GetObtVelocity() - (system.Target as ITargetable).GetObtVelocity();
						else
							rocketMode = RocketModes.Free;
						break;

					case RocketModes.TargetVelPlus:
						GUILayout.Label("Targets Velocity +");
						if (system.Target != null && system.Target is ITargetable)
							target_angle = (system.Target as ITargetable).GetObtVelocity();
						else
							rocketMode = RocketModes.Free;
						break;
						
					case RocketModes.TargetVelMinus:
						GUILayout.Label("Targets Velocity -");
						if (system.Target != null && system.Target is ITargetable)
							target_angle =  -(system.Target as ITargetable).GetObtVelocity();
						else
							rocketMode = RocketModes.Free;
						break;
                }

				switch (rocketModeThrust) {
					case RocketModesThrust.RCSBrakeRelVel:
						//GUILayout.Label("Kill relative velocity");
						if (system.Target != null && system.Target is ITargetable)
							target_angle_thrust =  system.vessel.GetObtVelocity() - (system.Target as ITargetable).GetObtVelocity();
						else
							rocketModeThrust = RocketModesThrust.Free;
						break;
					case RocketModesThrust.RCSBrake:
						//GUILayout.Label("Kill surface velocity");
						target_angle_thrust =  system.rotatingFrameVelocity;
						break;
				}
				GUILayout.EndHorizontal();

				GUILayout.Label("Orbit: ", Util.StdLabelSmallCustom(Color.yellow));
				
				GUILayout.BeginHorizontal();
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.Prograde, "Prograde", 90);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.Retrograde, "Retrograde", 90);
				GUILayout.EndHorizontal();
				
				GUILayout.BeginHorizontal();
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.ProGee, "ProGee", 90);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.RetroGee, "RetroGee", 90);
				GUILayout.EndHorizontal();
				
				GUILayout.BeginHorizontal();
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.Vplus, "Velocity +", 90);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.Vminus, "Velocity -", 90);
				GUILayout.EndHorizontal();
				
				GUILayout.BeginHorizontal();
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.Nplus, "Normal +", 90);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.Nminus, "Normal -", 90);
				GUILayout.EndHorizontal();
				
				GUILayout.Space(5);
				
				GUILayout.Label("Target: ", Util.StdLabelSmallCustom(Color.yellow));

				GUI.enabled = system.Target != null;
				GUILayout.BeginHorizontal();
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.TargetDirPlus, "Direction +", 90);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.TargetDirMinus, "Direction -", 90);
				GUILayout.EndHorizontal();

				GUI.enabled = system.Target != null && system.Target is ITargetable;
				GUILayout.BeginHorizontal();
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.TargetVelPlus, "Velocity +", 90);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.TargetVelMinus, "Velocity -", 90);
				GUILayout.EndHorizontal();

				GUI.enabled = system.Target != null;
				GUILayout.BeginHorizontal();
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.TargetRelDirPlus, "Rel Dir +", 90);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.TargetRelDirMinus, "Rel Dir -", 90);
				GUILayout.EndHorizontal();

				GUI.enabled = system.Target != null && system.Target is ITargetable;
				GUILayout.BeginHorizontal();
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.TargetRelVelPlus, "Rel Vel +", 90);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.TargetRelVelMinus, "Rel Vel -", 90);
				GUILayout.EndHorizontal();

				GUI.enabled = system.vessel.patchedConicSolver.maneuverNodes.Count > 0;
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.Maneuver, "Maneuver Node", 150);
				GUI.enabled = true;
				
				GUILayout.Space(5);
				
				GUILayout.Label("Generic: ", Util.StdLabelSmallCustom(Color.yellow));
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.RSAS, "RSAS", 150);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.CurrentDirection, "Current direction", 150);
				rocketMode = toggleRocketModes.SelectMode(rocketMode, RocketModes.Free, "Free", 150);
				
				GUILayout.Space(5);
				GUILayout.Label("..........................................................", Util.StdLabelSmallCustom(Color.gray));
				GUILayout.Label("RCS Translation: ", Util.StdLabelSmallCustom(Color.yellow));
				GUI.enabled = system.Target != null && system.Target is ITargetable;
				rocketModeThrust = toggleRocketModesThrust.SelectMode(rocketModeThrust, RocketModesThrust.RCSBrakeRelVel, "Kill Relative Velocity (obt)", 150);
				GUI.enabled = true;
				rocketModeThrust = toggleRocketModesThrust.SelectMode(rocketModeThrust, RocketModesThrust.RCSBrake, "Kill Velocity (srf)", 150);
				GUILayout.Space(3);
				GUILayout.BeginHorizontal();
				rcsFwd = GUILayout.RepeatButton("RCS Fwd");
				rcsBack = GUILayout.RepeatButton("RCS Back");
				GUILayout.EndHorizontal();

				GUILayout.Space(5);

				#region options rocket
				toggleAdvanced = GUILayout.Toggle(toggleAdvanced, "Advanced Controller Options");
				if (toggleAdvanced) {
					/*pidRocket_yaw.DebugGui("torque yaw");
					pidRocket_pit.DebugGui("torque pitch");
					pidRocket_rol.DebugGui("torque roll");
					pidRocket_lr.DebugGui("translate left/right");
					pidRocket_ud.DebugGui("translate up/down");
					pidRocket_fb.DebugGui("translate fwd/back");*/
					PIDclamp.DebugGuiPanel(pid_rocket_yaw, pid_rocket_pit, pid_rocket_rol, pid_rocket_lr, pid_rocket_ud, pid_rocket_fb);
				}
				#endregion
            }
            system.guiSubSystems.Footer(docked, rect);
        }


		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// FLYBY-WIRES
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   
        private void fly(FlightCtrlState s) {
            if (system.vessel != FlightGlobals.ActiveVessel) {
                FlightGlobals.ActiveVessel.OnFlyByWire -= fly;
                return;
            }
            diff = bearing - system.actualBearing;
            if (diff > 180)
                diff = diff - 360;
            else if (diff < -180)
                diff = diff + 360;

            float dampFactorX = system.vessel.angularVelocity.x;
            float dampFactorY = system.vessel.angularVelocity.y;
            float dampFactorZ = system.vessel.angularVelocity.z;

            avgMark.Average(system.actualMark);
            if (toggleAltitude) {
                double vAcc = Math.Sin(avgMark * Mathf.Deg2Rad) * avgAcc.Average(system.vessel.acceleration.magnitude);
                double vSpd = avgVspd.Average(system.vessel.verticalSpeed);
                double vDisp = (cruiseAlt - system.vessel.orbit.altitude);
                if(Time.fixedTime > lastTime){
                    avgMarkAcc.Average((system.actualMark - lastMark) / (Time.fixedTime - lastTime));
                }
                float dens = (float)system.vessel.atmDensity;
                //important values
                dPDinv = Mathf.Clamp((2 - (float)system.vessel.staticPressure), 1, 2); //density multiplier
                //main mark tuning
                double b = vDisp;
                double a = 900 * dPDinv * dPDinv * dPDinv;

                //fine tune the mark
                if ((Math.Abs(vSpd) < 15 || Math.Abs(vDisp) < (1500*dPDinv)) && system.vessel.orbit.altitude>100) {
                    if (vAcc > 0 ? avgMarkAcc < vAcc : avgMarkAcc > vAcc) {//hmm fix
                        if (vDisp > 0) {
                            if (vSpd < 0) { //going gown
                                tuner += ((float)vDisp / 1500) * dens; //invert
                                //MonoBehaviour.print("vspd < 0");
                            } else {
                                if (Math.Abs(vDisp) - (vSpd * 10) < 0) {
                                    tuner -= ((float)-(Math.Abs(vDisp) - (vSpd * 10)) / 1750) * dens; //too high speed
                                    //MonoBehaviour.print("too high");
                                } else if (Math.Abs(vDisp) - (vSpd * 50) > 0) {
                                    tuner += ((float)(Math.Abs(vDisp) - (vSpd * 50)) / 1750) * dens; //too low speed
                                    //MonoBehaviour.print("too low");
                                }
                            }
                        } else {
                            //MonoBehaviour.print("vDisp < 0");
                            if (vSpd > 0) { // going up
                                tuner += ((float)vDisp / 1000) * dens; //invert
                                //MonoBehaviour.print("vspd > 0");
                            } else {
                                if (Math.Abs(vDisp) + (vSpd * 10) < 0) {
                                    tuner += ((float)-(Math.Abs(vDisp) + (vSpd * 10)) / 1000) * dens; //too high speed ...remember vDisp is negative.
                                    //MonoBehaviour.print("too high");
                                } else if (Math.Abs(vDisp) + (vSpd * 50) > 0) {
                                    tuner -= ((float)(Math.Abs(vDisp) + (vSpd * 50)) / 1000) * dens; //too low speed ...remember vDisp is negative.
                                    //MonoBehaviour.print("too low");
                                }
                            }
                        }
                    } else {
                        if (avgMarkAcc > 0)
                            tuner -= ((float)Math.Sqrt(Math.Abs(vDisp)) / 500) * dens * (float)avgMarkAcc;
                        else
                            tuner += ((float)Math.Sqrt(Math.Abs(vDisp)) / 500) * dens * (float)avgMarkAcc;
                        //MonoBehaviour.print("vAcc check failed");
                    }
                } else {
                    //MonoBehaviour.print("tuner=0");
                    tuner = 0;
                }
                tuner = Mathf.Clamp(tuner, -maxClimb, maxClimb);

                //tuner should not increase faster than acc

                //set the mark
                mark = Mathf.Clamp((float)Math.Atan2(b, a) * Mathf.Rad2Deg + tuner, -maxClimb, maxClimb);
                lastMark = system.actualMark;
                lastTime = Time.fixedTime;

                //MonoBehaviour.print("vAcc: " + vAcc.ToString("f3") + " spd: " + vSpd.ToString("f3") + " tuner: " + tuner.ToString("f3") + " aMA: " + ((float)avgMarkAcc).ToString("f6"));
            }

   
            float climb = Mathf.Clamp(mark - system.actualMark, -5f, 5f); //climb factor
            float dPitch = Mathf.Clamp(pid_plane_pitch.Update(climb, Time.deltaTime), -1, 1);

            if (toggleFollowLeader) {
                
                Vector3 v = Quaternion.Inverse(Quaternion.Euler(90, 0, 0) * Quaternion.Inverse(system.vessel.transform.rotation) *
                    Quaternion.LookRotation(system.Leader.GetAim(), system.vessel.upAxis)).eulerAngles;
                float b = v.y > 180 ? v.y - 360 : v.y;
                float m = v.x > 180 ? v.x - 360 : v.x;

                this.bearing = actualBearing_Velocity - b;
                this.mark = system.actualMark + m;
                float diffV = bearing - actualBearing_Velocity; //
                if (diffV > 180)
                    diffV = diffV - 360;
                else if (diffV < -180)
                    diffV = diffV + 360;

                double turnRateModifier = 3 * (system.HeightFromTerrain() / 1200) + 1;

                float desiredRoll = Mathf.Clamp(Mathf.Clamp(diffV * (float)turnRateModifier, -50, 50) - system.actualRoll, -50, 50);
                float dF = pid_plane_roll_auto.Update(desiredRoll, Time.deltaTime);
                s.roll += Mathf.Clamp(dF * Mathf.Deg2Rad, -1f, 1f);
                s.yaw += pid_plane_yaw_auto.Update(sideSlipAngle, Time.deltaTime);

                s.pitch += dPitch;


            } else {
                if (toggleBearing || toggle_GS_travel) {
                    if (system.HeightFromTerrain() < 10)
                        s.yaw = Mathf.Clamp(diff * Mathf.Deg2Rad * 10, -1, 1);
                    else {
                        float diffV = bearing - actualBearing_Velocity; //
                        if (diffV > 180)
                            diffV = diffV - 360;
                        else if (diffV < -180)
                            diffV = diffV + 360;

                        double turnRateModifier = ((system.HeightFromTerrain() / 6000) + 1);

                        float desiredRoll = Mathf.Clamp(Mathf.Clamp(diffV * (float)turnRateModifier, -45, 45) - system.actualRoll, -45, 45);
                        float dF = pid_plane_roll_auto.Update(desiredRoll, Time.deltaTime);
                        s.roll = Mathf.Clamp(dF * Mathf.Deg2Rad, -1f, 1f);

                        s.yaw = pid_plane_yaw_auto.Update(sideSlipAngle, Time.deltaTime); //adjusts yaw according to TC quite fast.

                        //MonoBehaviour.print(velocityOffset.ToString("f4") + "  " + actualBearing_Velocity.ToString("f4") + "  " + diffV.ToString("f4") + " des " + desiredRoll.ToString("f4") + " dF " + dF.ToString("f4"));
                    }
                } else {
                    if (toggleYaw)
                        s.yaw = Mathf.Clamp(yaw / 100, -1, 1);
                    else
                        s.yaw += Mathf.Clamp(yaw, -1, 1);

                    if (toggleRoll) {
                        float dF = pid_plane_roll_manual.Update(roll - system.actualRoll, Time.deltaTime);
                        s.roll = Mathf.Clamp(dF * Mathf.Deg2Rad, -1f, 1f);
                    } else
                        s.roll -= Mathf.Clamp((system.actualRoll - roll) * Mathf.Deg2Rad, -1, 1);
                }

                if (toggleMark || toggleAltitude)
                    s.pitch = dPitch;
            }

            
            if (toggleDampen) {
                s.yaw = Mathf.Clamp(s.yaw + dampFactorZ * 2, -1, 1);
                //s.roll = Mathf.Clamp(s.roll + dampFactorY / 2, -1, 1);
                s.pitch = Mathf.Clamp(s.pitch + (toggleAltitude ? dampFactorX * 4 : dampFactorX * 2), -1, 1); 
            }
            pitch = s.pitch;
            
            if (FlightGlobals.ActiveVessel.isEVA) { //switch to remote control on EVA
                ctrlStateRemote.pitch = s.pitch;
                ctrlStateRemote.yaw = s.yaw;
                ctrlStateRemote.roll = s.roll;
                ctrlStateRemote.mainThrottle = throttleEVA;
                system.vessel.rootPart.propagateControlUpdate(ctrlStateRemote);
                comingFromEVA = true;
            } else {
                if (comingFromEVA) {
                    FlightInputHandler.state.mainThrottle = throttleEVA;
                    comingFromEVA = false;
                }
                throttleEVA = s.mainThrottle;
            }
        }


        //makes the effect value curve towards 0 at values('m') below 5
        private float Dampen(float m) {
            float mf = m;
            if (m > -5 && m < 0)
                mf = -(5 / Mathf.Pow(1.25f, ((Math.Abs(m) - 5) * (Math.Abs(m) - 5))));
            else if (m > 0 && m < 5)
                mf = 5 / Mathf.Pow(1.25f, ((m - 5) * (m - 5)));
            return mf;
        }
        private float DampenSoft(float m) {
            float mf = m;
            if (m > -4.5 && m < 0)
                mf = -((5 / Mathf.Pow(1.12f, ((Math.Abs(m) - 5) * (Math.Abs(m) - 6)))) - 0.15f);
            else if (m > 0 && m < 4.5)
                mf = (5 / Mathf.Pow(1.12f, ((m - 5) * (m - 6)))) - 0.15f;
            return mf;
        }


        private void flyRocket(FlightCtrlState s) {
            if (system.vessel != FlightGlobals.ActiveVessel) { //HMM
                FlightGlobals.ActiveVessel.OnFlyByWire -= flyRocket; //HMM
                return;
            }
            float needed_pitch = 0;
            float needed_yaw = 0;
            Util.PitchYawAngle(system.vessel.ReferenceTransform, target_angle, out needed_pitch, out needed_yaw);
            float stop_pitch = (system.vessel.angularVelocity.x * system.vessel.angularVelocity.x) / ((2 * torque * system.vessel.angularVelocity.x) / system.vessel.angularMomentum.x);
            float stop_yaw = (system.vessel.angularVelocity.z * system.vessel.angularVelocity.z) / ((2 * torque * system.vessel.angularVelocity.z) / system.vessel.angularMomentum.z);
            needed_pitch *= Mathf.Deg2Rad;
            needed_yaw *= Mathf.Deg2Rad;
            if (rocketMode != RocketModes.Free) {
                if (rocketMode == RocketModes.RSAS) {
					if(!(Input.GetKey(GameSettings.YAW_LEFT.primary) || Input.GetKey(GameSettings.YAW_RIGHT.primary)))
						s.yaw += pid_rocket_yaw.Update((system.vessel.angularVelocity.z), Time.deltaTime);
					if(!(Input.GetKey(GameSettings.PITCH_DOWN.primary) || Input.GetKey(GameSettings.PITCH_UP.primary)))
						s.pitch += pid_rocket_pit.Update((system.vessel.angularVelocity.x), Time.deltaTime);
					if (!(Input.GetKey(GameSettings.ROLL_LEFT.primary) || Input.GetKey(GameSettings.ROLL_RIGHT.primary)))
						s.roll += pid_rocket_rol.Update(system.vessel.angularVelocity.y, Time.deltaTime);
				} else {
                    //find the axis with the shortest path to 0 or |180|
                    float p = 0; 
                    float y = 0;
                    float a = Math.Abs(needed_pitch);
                    float ai = Mathf.PI - a;
                    float b = Math.Abs(needed_yaw);
                    float bi = Mathf.PI - b;
                    float min = Math.Min(Math.Min(a, ai), Math.Min(b, bi));
                    if (min == a) {
                        p = needed_pitch;
                        y = needed_yaw;
                    } else if (min == b) {
                        p = needed_pitch;
                        y = needed_yaw;
                    } else if (min == ai) {
                        p = -needed_pitch;
                        y = needed_yaw;
                    } else if (min == bi) {
                        p = needed_pitch;
                        y = -needed_yaw;
                    }
                    //yaw to the right = negative
                    //pitch under = negative
                    if (!(Input.GetKey(GameSettings.YAW_LEFT.primary) || Input.GetKey(GameSettings.YAW_RIGHT.primary))){
						if(stop_yaw < b)
							s.yaw += pid_rocket_yaw.Update((system.vessel.angularVelocity.z + y), Time.deltaTime);
						else
							s.yaw += pid_rocket_yaw.Update((system.vessel.angularVelocity.z), Time.deltaTime);
					}
					if (!(Input.GetKey(GameSettings.PITCH_DOWN.primary) || Input.GetKey(GameSettings.PITCH_UP.primary))){
						if(stop_pitch < a)
							s.pitch += pid_rocket_pit.Update((system.vessel.angularVelocity.x - p), Time.deltaTime);
						else
							s.pitch += pid_rocket_pit.Update((system.vessel.angularVelocity.x), Time.deltaTime);
					}
					if (!(Input.GetKey(GameSettings.ROLL_LEFT.primary) || Input.GetKey(GameSettings.ROLL_RIGHT.primary))) {
						s.roll += pid_rocket_rol.Update(system.vessel.angularVelocity.y, Time.deltaTime);
					}

                }
                if (FlightGlobals.ActiveVessel.isEVA) { //switch to remote control on EVA
                    ctrlStateRemote.pitch = s.pitch;
                    ctrlStateRemote.yaw = s.yaw;
                    ctrlStateRemote.roll = s.roll;
                    ctrlStateRemote.mainThrottle = throttleEVA;
                    system.vessel.rootPart.propagateControlUpdate(ctrlStateRemote);
                    comingFromEVA = true;
                } else {
                    if (comingFromEVA) {
                        FlightInputHandler.state.mainThrottle = throttleEVA;
                        comingFromEVA = false;
                    }
                    throttleEVA = s.mainThrottle;
                }
            }
			if(rocketModeThrust != RocketModesThrust.Free){
				Vector3 dir = system.transform.InverseTransformDirection(target_angle_thrust);
				if (!(Input.GetKey(GameSettings.TRANSLATE_LEFT.primary) || Input.GetKey(GameSettings.TRANSLATE_RIGHT.primary))) {
					s.X = pid_rocket_lr.Update(dir.x, Time.deltaTime); //1 translate left
				}if(!(Input.GetKey(GameSettings.TRANSLATE_UP.primary) || Input.GetKey(GameSettings.TRANSLATE_DOWN.primary))){
					s.Y = pid_rocket_ud.Update(dir.z, Time.deltaTime); //1 translate up //-target_angle.y;
				}if(!(Input.GetKey(GameSettings.TRANSLATE_FWD.primary) || Input.GetKey(GameSettings.TRANSLATE_BACK.primary))){
					s.Z = pid_rocket_fb.Update(dir.y, Time.deltaTime); //1 translate back //-target_angle.x;
				}
				//MonoBehaviour.print(dir.x + "  " + dir.y + "  "+ dir.z);
			}
			if(rcsFwd)
				s.Z = -1;
			else if(rcsBack)
				s.Z = 1;

        }

        override public void onFlightStateLoad(Dictionary<string, KSPParseable> parsedData) {
            onFlightStateLoadX(parsedData, "CFG_Red");
            if (system.IsHorizontal) {//for airplanes
                hEnabled.set = enabled;
            } else { // For ROCKETS
                vEnabled.set = enabled;
            }
        }

        override public void onFlightStateSave(Dictionary<string, KSPParseable> partDataCollection) {
            onFlightStateSaveX(partDataCollection, "CFG_Red");
        }


    }
}

