﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {
    class CFG_yellow : CFG{

        private Vessel transferVessel;
        VesselIteratorOwn.Resource rT; //current transfer resource in TARGET
        VesselIteratorOwn.Resource r; //current transfer resource in own vessel
        private DoubleSwitch toggleTransfer;
        private DoubleSwitch toggleAfterburner;

		//Fuel transfer
        private bool toggleTransferFrom;
        private bool toggleTransferTo;
        private bool toggleTransferDump;
        private const float maxTransferRange = 10000f;
        private Dictionary<LazorAimable,GameObject> effects = new Dictionary<LazorAimable,GameObject>();
		private VesselIteratorTarget transferVesselIterator;
		private Vessel actualTransferVessel = null;
		private VesselIteratorTarget actualTransferVesselIterator = null;
		private double targetDistance;

        //Afterburner
        private Dictionary<Part, GameObject> afterBurnerList = new Dictionary<Part, GameObject>();
        private LinkedList<string> currentTransfer = new LinkedList<string>();
		private float afterburnerYield = 1;
		private float afterburnerYield_min = 0;
		private float afterburnerYield_max = 5;


        public CFG_yellow(LazorSystem system) : base(system) {
            name = "Engineering";
            color = Color.yellow;
            colorTexture = Util.LightTexture(Util.getTexture(new Color(color.r, color.g, color.b, 0.9f)), 0.6f);
            currentTransfer = new LinkedList<string>(system.VI.resourceTypes);
        }

        override public void Dispose() {
            foreach (GameObject go in effects.Values)
                UnityEngine.Object.Destroy(go);
            foreach (GameObject go in afterBurnerList.Values)
                UnityEngine.Object.Destroy(go);
            system.lazorSound.StopLoop();
        }

        override public bool Detect(Dictionary<int, List<LazorAimable>> lazors) {
            claimedLazors = StdDetectAll(color, lazors);
            effects.Clear();
            foreach (int i in claimedLazors) {
                foreach (LazorAimable l in lazors[i]) {
                    //fuel pouring anim
                    GameObject o = Part.Instantiate(UnityEngine.Resources.Load("Effects/fx_exhaustFlame_blue")) as GameObject;
                    o.transform.parent = l.transform;
                    o.transform.localEulerAngles = Vector3.zero;
                    o.transform.localPosition = new Vector3(0, 0.11f, 0);
                    ParticleEmitter emitter = o.particleEmitter;
                    emitter.emit = true;
                    emitter.minSize = 0F;
                    emitter.maxSize = 0F;
                    emitter.minEnergy = 0.1F;
                    emitter.maxEnergy = 0.3F;
                    emitter.rndVelocity = Vector3.one * 0.15f;
                    emitter.localVelocity = Vector3.zero;
                    emitter.useWorldSpace = true;
                    (o.GetComponent<ParticleAnimator>() as ParticleAnimator).sizeGrow = -0.8F;
                    ParticleAnimator flare_emitterAnimator = emitter.GetComponent<ParticleAnimator>();
                    Color[] emitterColors = flare_emitterAnimator.colorAnimation;
                    emitterColors[0] = new Color(0.5f, 0.5f, 0, 1);
                    emitterColors[1] = new Color(1f, 1f, 0, 1);
                    emitterColors[2] = new Color(0.5f, 0.5f, 0, 1);
                    emitterColors[3] = new Color(1f, 1f, 0, 1);
                    flare_emitterAnimator.colorAnimation = emitterColors;
                    effects.Add(l, o);
                }
            }
            return claimedLazors.Count > 0;
        }


        private GameObject MakeAfterburnerEffect(Part p) {
            GameObject afterBurner = Part.Instantiate(UnityEngine.Resources.Load("Effects/fx_exhaustFlame_blue")) as GameObject;
            afterBurner.transform.parent = p.transform;
            afterBurner.transform.localEulerAngles = Vector3.zero;
            afterBurner.transform.localPosition = new Vector3(0, -2, 0);
            ParticleEmitter emitter = afterBurner.particleEmitter;
            emitter.emit = true;
            emitter.minSize = 0F;
            emitter.maxSize = 0F;
            emitter.minEnergy = 0.5F;
            emitter.maxEnergy = 0.7F;
            emitter.rndVelocity = Vector3.one * 0.3f;
            emitter.localVelocity = Vector3.zero;
            emitter.useWorldSpace = false;
            ParticleAnimator emitterAnimator = emitter.GetComponent<ParticleAnimator>();
            emitterAnimator.sizeGrow = -0.91F;
            Color[] emitterColors = emitterAnimator.colorAnimation;
            emitterColors[0] = new Color(0, 0, 0.7f, 0.9f);
            emitterColors[1] = new Color(0.1f, 0.1f, 1f, 0.8f);
            emitterColors[2] = new Color(0.25f, 0.25f, 0.9f, 0.7f);
            emitterColors[3] = new Color(0.5f, 0.5f, 0.8f, 0.6f);
            emitterColors[4] = new Color(1f, 0f, 0f, 0.6f);
            emitterAnimator.colorAnimation = emitterColors;
            return afterBurner;
        }

        //aim lazors at fuel tanks on target vessel
        override public void ExecuteCommand(LazorAimable lazor) {
            if (lazor != null){
                if(system.Target != null && system.Target is Vessel)
                    lazor.Lazor.Aim(system.Target.transform.position - lazor.Lazor.GetOrigin());
                else if(toggleTransferLOCK && actualTransferVessel != null)
                    lazor.Lazor.Aim(actualTransferVessel.transform.position - lazor.Lazor.GetOrigin());
            }
        }
		
        override public void OnDisable() { 
            toggleTransfer.set = false;
            toggleAfterburner.set = false;
            toggleTransferFrom = false;
            toggleTransferTo = false;
            toggleTransferDump = false;
        }

        override public void OnEnabledFixedUpdate() {
            if (!system.VI.hasFuel)
                toggleAfterburner.set = false;
            if (toggleAfterburner.In()) { //toggle effects ON when dumping
                Vector3 vel = FlightGlobals.getGeeForceAtPosition(system.vessel.transform.position);
                foreach (int i in claimedLazors) {
                    foreach (LazorAimable l in system.config.LazorCounterparts[i]) { //set up liquid fuel engine particle effects
                        ParticleEmitter pe = effects[l].particleEmitter;
                        pe.minSize = 0.04f;
                        pe.maxSize = 0.08f;
                        pe.worldVelocity = vel; //wrong
                    }
                }
                foreach (Part p in VesselIteratorOwn.GetEngines(system.vessel)) { //set up afterburners particle effects
                    if (!afterBurnerList.ContainsKey(p)) {
                        afterBurnerList.Add(p, MakeAfterburnerEffect(p));
                    }
                }
                system.lazorSound.PlayLoop(SoundType.AFTERBURNER);
            } else if (toggleAfterburner.Out()) {
                foreach (int i in claimedLazors) {
                    foreach (LazorAimable l in system.config.LazorCounterparts[i]) {
                        ParticleEmitter pe = effects[l].particleEmitter;
                        pe.minSize = 0.0f;
                        pe.maxSize = 0.0f;
                    }
                }
                foreach (Part p in afterBurnerList.Keys) { //shut down the afterburner effect on each engine
                    if (p != null){
                        ParticleEmitter pe = afterBurnerList[p].particleEmitter;
                        pe.minSize = 0;
                        pe.maxSize = 0;
                    }
                }
                afterBurnerList.Clear();
                system.lazorSound.StopLoop();
            }

            if (toggleAfterburner && system.VI.hasFuel && system.VI.resourceParts["LiquidFuel"].amount > 0) {
                //system.VI.DrainResource("LiquidFuel", 0.075f * claimedLazors.Count);
                //draw the afterburner effect on each engine
                float modifier = FlightInputHandler.state.mainThrottle < 0.01f ? 0 : FlightInputHandler.state.mainThrottle > 0.5 ? 1 : FlightInputHandler.state.mainThrottle * 2;
                foreach (Part p in afterBurnerList.Keys) {
                    float angleModifier = 1 - Mathf.Clamp(Vector3.Angle(system.rotatingFrameVelocity, p.transform.up) / 20, 0, 1);
                    if (p != null && p.inverseStage == system.vessel.currentStage && angleModifier > 0) {
                        ParticleEmitter pe = afterBurnerList[p].particleEmitter;
                        pe.minSize = 7F * modifier * angleModifier;
                        pe.maxSize = 7.5F * modifier * angleModifier;
                        pe.worldVelocity = -p.transform.up * (2 + 12 * FlightInputHandler.state.mainThrottle);

						system.RequestFuelDelta( ((ModuleEngines)p.Modules["ModuleEngines"]).propellants.Find(x => x.name == "LiquidFuel").currentAmount/Time.fixedDeltaTime * afterburnerYield);// * (1/Time.fixedDeltaTime));

						//system.RequestFuelDelta( ((ModuleEngines)p.Modules["ModuleEngines"]).fuelFlowGui * afterburnerYield * 100);// * (1/Time.fixedDeltaTime));
						p.rigidbody.AddForce(p.transform.up * ((ModuleEngines)p.Modules["ModuleEngines"]).maxThrust * FlightInputHandler.state.mainThrottle * angleModifier * afterburnerYield); // * ((float)claimedLazors.Count / (float)afterBurnerList.Keys.Count));
                        system.lazorSound.SetVolumeLoop(modifier * angleModifier);
                    }
                }
            } else {
                toggleAfterburner.set = false;
            }


            //do fuel transfer stuff
            float amount = 0;
            float amountPerLazor = 0;
            if (actualTransferVessel != null) {
                //Vessel v = actualTransferVessel;
                //determine fuel transfer rate
                if (toggleTransferFrom || toggleTransferTo) {
                    amountPerLazor = (3000f / maxTransferRange ) / (((float)targetDistance / (maxTransferRange / 50f)) + 1f); //transfer speed is 51x as slow at max range
                    //MonoBehaviour.print("apl " + amountPerLazor.ToString("f7"));
                    amount = ClaimedLazorNotHittingOwnVessel() * amountPerLazor;
                }
                //fuel transfer
                if (toggleTransferFrom && rT != null && rT.amount > 0 && r.capacity > 0) {
                    amount = Mathf.Clamp(amount, 0, (float)r.capacity);
                    //MonoBehaviour.print("a1 " + amount.ToString("f7") + " cap "+((float)r.capacity).ToString("f7"));
                    amount = (float)actualTransferVesselIterator.DrainResource(currentTransfer.First.Value, amount);
                    //MonoBehaviour.print("a2 " + amount.ToString("f7"));
                    system.VI.AddResource(currentTransfer.First.Value, amount);
                } else if (toggleTransferTo && r.amount > 0 && rT.capacity > 0) {
                    amount = Mathf.Clamp(amount, 0, (float)rT.capacity);
                    amount = (float)system.VI.DrainResource(currentTransfer.First.Value, amount);
                    actualTransferVesselIterator.AddResource(currentTransfer.First.Value, amount);
                } 
            }
            if (toggleTransferDump && r.amount > 0) {
                amount = ClaimedLazorNotHittingOwnVessel() * 0.3f;
                amount = Mathf.Clamp(amount, 0, (float)r.amount);
                amount = (float)system.VI.DrainResource(currentTransfer.First.Value, amount);
            }
            if (toggleTransfer.In()) {

            } else if (toggleTransfer.Out()) {
                OnTargetSwitch();
                toggleTransferLOCK.set = false;
            }
            if (toggleTransferLOCK.In()) {
                transferVessel = (Vessel)system.Target;
                transferVesselIterator = new VesselIteratorTarget(system, transferVessel);
            } else if (toggleTransferLOCK.Out()) {
                transferVessel = null;
                transferVesselIterator = null;
            }
                  
        }

        override public void OnTargetSwitch() {
            toggleTransferFrom = false;
            toggleTransferTo = false;
            toggleTransferDump = false;
        }

        override public void DrawGUIbutton() {
            enabled.set = system.guiSubSystems.SubsystemButton(enabled, name, color, out docked, ContentGUI, 0);
        }

        ////////////////////////////////////////////////////////////////////
        /// GUI
        ////////////////////////////////////////////////////////////////////
        private DoubleSwitch toggleTransferLOCK;
        override public void ContentGUI(int windowID) {
            docked = system.guiSubSystems.Header(docked, rect, enabled, ref enabled);
            //enable transfer and lock buttons
            GUILayout.BeginHorizontal(Util.Space_Compact);
            toggleTransfer.set = GUILayout.Toggle(toggleTransfer, "Resource Transfer");
            if (toggleTransfer) {
                if (system.Target != null && system.Target is Vessel || toggleTransferLOCK)
                    toggleTransferLOCK.set = GUILayout.Toggle(toggleTransferLOCK, toggleTransferLOCK ? "Locked" : "Lock");
            }
            GUILayout.EndHorizontal();
            GUILayout.Space(5);

            if (toggleTransfer) {
                //Target display
                GUILayout.BeginVertical(Util.TxtArea);
                
                if (toggleTransferLOCK) {
                    actualTransferVessel = transferVessel;
                    actualTransferVesselIterator = transferVesselIterator;
                } else if (system.Target != null && system.Target is Vessel) {
                    actualTransferVessel = ((Vessel)system.Target);
                    actualTransferVesselIterator = system.VIT;
                } else {
                    actualTransferVessel = null;
                    actualTransferVesselIterator = null;
                }
                if(actualTransferVessel != null){
                    GUILayout.BeginHorizontal();
                    GUILayout.Label("Target: ", Util.StdLabelSmallCustom(Color.green));
                    GUILayout.Label(actualTransferVessel.vesselName, Util.StdLabelSmallCustom(Color.white));
                    GUILayout.EndHorizontal();
                    targetDistance = (actualTransferVessel.transform.position - system.vessel.transform.position).magnitude;
                    if (targetDistance < maxTransferRange) {
                        if (system.lazorRange > targetDistance) {
                            if (actualTransferVesselIterator.resourceParts.ContainsKey(currentTransfer.First.Value)) {
                                rT = actualTransferVesselIterator.resourceParts[currentTransfer.First.Value];
                                GUILayout.BeginHorizontal();
                                GUILayout.Label("Amount: ", Util.StdLabelSmallCustom(Color.green));
                                GUILayout.Label(rT.amount.ToString("f1") + " / " + rT.maxAmount.ToString("f1"), Util.StdLabelSmallCustom(Color.white));
                                GUILayout.Label("U", Util.StdLabelSmallCustom(Color.green));
                                GUILayout.EndHorizontal();
                                GUILayout.BeginHorizontal();
                                GUILayout.Label("Capacity: ", Util.StdLabelSmallCustom(Color.green));
                                GUILayout.Label(rT.capacity.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
                                GUILayout.Label("U", Util.StdLabelSmallCustom(Color.green));
                                GUILayout.EndHorizontal();
                            } else {
                                rT = null;
                                GUILayout.BeginHorizontal();
                                GUILayout.Label("Resource not present in target vessel. ", Util.StdLabelSmallCustom(Color.red));
                                GUILayout.EndHorizontal();
                            }
                        } else {
                            rT = null;
                            GUILayout.BeginHorizontal();
                            GUILayout.Label("Target is within transfer range but lazor range is too short.", Util.StdLabelSmallCustom(Color.red));
                            GUILayout.EndHorizontal();
                        }
                    } else {
                        rT = null;
                        GUILayout.BeginHorizontal();
                        GUILayout.Label("Target is too far away! ", Util.StdLabelSmallCustom(Color.red));
                        GUILayout.EndHorizontal();
                    }
                } else {
                    rT = null;
                    GUILayout.BeginHorizontal();
                    GUILayout.Label("No target", Util.StdLabelSmallCustom(Color.red));
                    GUILayout.EndHorizontal();
                }
                GUILayout.EndVertical();

                //Select resource
                GUILayout.BeginHorizontal();
                if (GUILayout.Button("◄", Util.Buttons_arrows_mainPanel, GUILayout.Width(22))) {
                    string t = currentTransfer.Last.Value;
                    currentTransfer.AddFirst(t);
                    currentTransfer.RemoveLast();
                    if (actualTransferVessel != null)
                        actualTransferVesselIterator.Update(true);
                    OnTargetSwitch();
                }
                GUILayout.BeginVertical(Util.TxtArea);
                GUILayout.Label(currentTransfer.First.Value, Util.StdLabelSmallCustom(Color.green, TextAnchor.MiddleCenter), GUILayout.Width(110), GUILayout.Height(17));
                GUILayout.EndVertical();
                if (GUILayout.Button("►", Util.Buttons_arrows_mainPanel, GUILayout.Width(22))) {
                    string t = currentTransfer.First.Value;
                    currentTransfer.AddLast(t);
                    currentTransfer.RemoveFirst();
                    if (actualTransferVessel != null)
                        actualTransferVesselIterator.Update(true);
                    OnTargetSwitch();
                }
                GUILayout.EndHorizontal();

                GUILayout.Space(5);

                //Transfer buttons
                GUILayout.BeginHorizontal();
                GUILayout.FlexibleSpace();
                GUILayout.Label("Transfer: ");
                GUILayout.Space(3);
                if (actualTransferVessel == null) {
                    GUI.enabled = false;
                    toggleTransferFrom = false;
                    toggleTransferTo = false;
                    GUILayout.Toggle(toggleTransferFrom, "▼", Util.Buttons_arrows_mainPanel, GUILayout.Width(22));
                    GUILayout.Toggle(toggleTransferTo, "▲", Util.Buttons_arrows_mainPanel, GUILayout.Width(22));
                    GUI.enabled = true;
                } else {
                    if (rT != null)
                        GUI.enabled = true;
                    else {
                        GUI.enabled = false;
                        toggleTransferTo = false;
                        toggleTransferFrom = false;
                    }
                    toggleTransferFrom = GUILayout.Toggle(toggleTransferFrom, "▼", Util.Buttons_arrows_mainPanel, GUILayout.Width(22));
                    if (toggleTransferFrom) {
                        toggleTransferTo = false;
                        toggleTransferDump = false;
                    }
                    toggleTransferTo = GUILayout.Toggle(toggleTransferTo, "▲", Util.Buttons_arrows_mainPanel, GUILayout.Width(22));
                    if (toggleTransferTo) {
                        toggleTransferFrom = false;
                        toggleTransferDump = false;
                    }
                }
                GUI.enabled = true;
                toggleTransferDump = GUILayout.Toggle(toggleTransferDump, "Dump", Util.Button_Toggle, GUILayout.Width(38), GUILayout.Height(22));
                if (toggleTransferDump) {
                    toggleTransferTo = false;
                    toggleTransferFrom = false;
                }
                GUILayout.FlexibleSpace();
                GUILayout.EndHorizontal();


                //Own vessel display
                r = system.VI.resourceParts[currentTransfer.First.Value];
                GUILayout.BeginVertical(Util.TxtArea);
                GUILayout.BeginHorizontal();
                GUILayout.Label("Amount: ", Util.StdLabelSmallCustom(Color.green));
                GUILayout.Label(r.amount.ToString("f1") + " / " + r.maxAmount.ToString("f1"), Util.StdLabelSmallCustom(Color.white));
                GUILayout.Label("U", Util.StdLabelSmallCustom(Color.green));
                GUILayout.EndHorizontal();
                GUILayout.BeginHorizontal();
                GUILayout.Label("Capacity: ", Util.StdLabelSmallCustom(Color.green));
                GUILayout.Label(r.capacity.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
                GUILayout.Label("U", Util.StdLabelSmallCustom(Color.green));
                GUILayout.EndHorizontal();
                GUILayout.EndVertical();
            }

            GUILayout.Space(5);


            //Afterburner
            if (system.VI.hasFuel) {
                toggleAfterburner.set = GUILayout.Toggle(toggleAfterburner, "Afterburner");
				afterburnerYield = GUILayout.HorizontalSlider(afterburnerYield, afterburnerYield_min, afterburnerYield_max);
				GUILayout.Label("Afterburner yield: "+afterburnerYield*100+"%");
            } else {
                GUI.enabled = false;
                toggleAfterburner.set = GUILayout.Toggle(toggleAfterburner, "Afterburner");
                GUI.enabled = true;
                GUILayout.Label("Afterburner requires Liquid Fuel");
            }
            system.guiSubSystems.Footer(docked, rect);
        }

        override public void onFlightStateLoad(Dictionary<string, KSPParseable> parsedData) {
            onFlightStateLoadX(parsedData, "CFG_Yellow");
        }

        override public void onFlightStateSave(Dictionary<string, KSPParseable> partDataCollection) {
            onFlightStateSaveX(partDataCollection, "CFG_Yellow");
        }



    }
}
