﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

namespace Romfarer {

    public enum CameraModes { Normal = 0, NightVision = 1, Invert = 2, Thermal = 3, Silhouette = 4 };

    //[RequireComponent(typeof(Camera))]
    public class CameraFilter : MonoBehaviour {

        private Material currentMaterial;

        public static Material normal = new Material(
                "Shader \"Unlit/Texture\" {" +
                "   Properties {" +
                "       _MainTex (\"MainTex\", 2D) = \"black\" {}" +
                "   }" +
                "   SubShader {" +
                "       Tags { \"RenderType\"=\"Opaque\" }" +
                "       LOD 100"+
                "       Pass {" +
                "           Lighting Off" +
                "           SetTexture [_MainTex] { combine texture }" +
                "       }" +
                "   }" +
                "}"
            );


        public static Material nightVision = new Material(
                "Shader \"Nightvision\" {" +
                "    Properties {" +
                "        _MainTex (\"MainTex\", 2D) = \"white\" {}" +
                "    }" +
                "    SubShader {" +
            //"        Tags { \"RenderType\"=\"Opaque\" }"+ works
            //"        Tags { \"RenderType\"=\"Opaque\" }" + 
            //"        Tags { \"LightMode\"=\"PixelOrNone\" }" + no
                "        Pass {" +
                "            Blend SrcColor DstColor" +
                "            SetTexture [_MainTex] {" +
                "                constantColor (0.5,0.7,0.5,0.5)" +
                "                combine constant + texture" +
                "            }" +
                "        }" +
                "        Pass {" +
                "            Blend SrcColor DstColor" +
                "            SetTexture [_MainTex] {" +
                "                combine texture * previous" +
                "            }" +
                "        }" +
                "    }" +
                "} "
            );

        public static Material invert = new Material(
                "Shader \"Invert\" {" +
                "   Properties {" +
                "       _MainTex (\"MainTex\", 2D) = \"white\" {}" +
                "   }" +
                "   SubShader {" +
                "       Pass {" +
                "           Lighting Off" +
                "           SetTexture [_MainTex] {" +
                "               constantColor (1,1,1,1)" +
                "               combine constant - texture " +
                "           }" +
                "       }" +
                "   }" +
                "}"
            );
        public static Material thermal = new Material(
                "Shader \"ObjectView\" {" +
                "    Properties {" +
                "        _MainTex (\"MainTex\", 2D) = \"white\" {}" +
                "    }" +
                "    SubShader {" +
                "        Tags { \"Queue\" = \"Transparent\" }" +
                "        Pass {" +
                "            Blend DstAlpha OneMinusSrcAlpha" +
                "            SetTexture [_MainTex] {" +
                "                constantColor (0.3,0.3,0.3,1)" +
                "                combine constant - texture double" +
                "            }" +
                "        }" +
                "    }" +
                "} "
            );
        public static Material silhouette = new Material(
                "Shader \"Objectvision\" {" +
                "    Properties {" +
                "        _MainTex (\"MainTex\", 2D) = \"white\" {}" +
                "    }" +
                "    SubShader {" +
                "        Tags { \"Queue\" = \"Transparent\" }" +
                "        Pass {" +
                "            Blend SrcAlpha OneMinusSrcAlpha" +
                "            SetTexture [_MainTex] {" +
                "                combine texture * primary double, texture * primary" +
                "            }" +
                "        }" +
                "    }" +
                "} "
            );

        public static Material Material(CameraModes mode) {
            switch (mode) {
                case CameraModes.Normal:
                    return normal;
                case CameraModes.NightVision:
                    return nightVision;
                case CameraModes.Invert:
                    return invert;
                case CameraModes.Silhouette:
                    return silhouette;
                case CameraModes.Thermal:
                    return thermal;
            }
            return normal;
        }

        public void SetMaterial(CameraModes mode){
            if(mode == CameraModes.Normal)
                SetMaterial(null);
            else
                SetMaterial(Material(mode));
        }

        public void SetMaterial(Material mat){
            currentMaterial = mat;
            //MonoBehaviour.print("XXcamXX" + (mat == null ? "null" : mat.ToString()) + " X " + (currentMaterial == null ? "null" : currentMaterial.ToString()));
        }

        /*public void ResetMaterial() {
            currentMaterial = null;
        }*/

        void OnRenderImage(RenderTexture source, RenderTexture dest) {
            if(currentMaterial != null)
                Graphics.Blit(source, dest, currentMaterial);
            //Graphics.Blit(source, dest, nightVision);
        }
    }
















    //test crap



    public static class MultiMaterial {
        public static Material Lighten = new Material(
            "Shader \"Lighten\" {" +
            "    Properties {" +
            "        _MainTex (\"MainTex\", 2D) = \"white\" {}" +
            "    }" +
            "    SubShader {" +
            "        Pass {" +
            //"            Blend SrcColor DstColor" +
            "            SetTexture [_MainTex] {" +
            "                constantColor (0.4,0.4,0.4,1)" +
            "                combine constant + texture" +
            "            }" +
            "        }" +
            "    }" +
            "} "
        );

        public static Material Grayscale{
            get{
                Material mat = new Material(Shader.Find("Hidden/Grayscale Effect"));
                mat.SetTexture("_RampTex", Util.LoadTexture(256, 1, "ramp_grayscale.png"));
                mat.SetFloat("_RampOffset", 0);
                return mat;
            }
        }
        
        public static Material ColorCorrection {
            get {
                Material mat = new Material(Shader.Find("Hidden/Color Correction Effect"));
                mat.SetTexture("_RampTex", Util.LoadTexture(256, 1, "ramp_grayscale.png"));
                return mat;
            }
        }

        public static Material EdgeDetect {
            get {
                Material mat = new Material(Shader.Find("Hidden/Edge Detect X"));
                mat.SetFloat("_Treshold", 0.2f * 0.2f);
                return mat;
            }
        }

        //MotionBlur y;
        //SepiaToneEffect x;
        public static Material Sepiatone {
            get {
                Material mat = new Material(Shader.Find("Hidden/Sepiatone Effect"));
                return mat;
            }
        }
        //TwirlEffect x; "Hidden/Twirt Effect Shader"
        //VortexEffect x;
    }







	public class CameraFilterDummy3 : MonoBehaviour {
		void OnPostRender(RenderTexture source, RenderTexture dest) {
			Graphics.Blit(source, dest);
		}
	}

	public class CameraFilterDummy2 : MonoBehaviour {
		void OnRenderImage(RenderTexture source, RenderTexture dest) {
			Graphics.Blit(source, dest);
		}
	}

    public class CameraFilterDummy : MonoBehaviour {
        /*Material mat;
        public void Awake() {
            mat = CameraFilter.normal;
        }*/
        void OnRenderImage(RenderTexture source, RenderTexture dest) {
            //Graphics.Blit(source, dest, mat);
        }

        //Clear alpha      
        /*
        private Material mat2;
        void OnPostRender() {
            if (!mat2) {
                mat2 = new Material ("Shader \"Hidden/Alpha\" {" +
                    "SubShader {" +
                    "    Pass {" +
                    "        ZTest Always Cull Off ZWrite Off" +
                    "        ColorMask A" +
                    "        Color (0,0,0,1)" +
                    "    }" +
                    "}" +
                    "}"
                );
                mat2.shader.hideFlags = HideFlags.HideAndDontSave;
                mat2.hideFlags = HideFlags.HideAndDontSave;
            }
            GL.PushMatrix ();
            GL.LoadOrtho ();
            for (var i = 0; i < mat.passCount; ++i) {
                mat.SetPass (i);
                GL.Begin (GL.QUADS);
                GL.Vertex3 (0, 0, 0.1f);
                GL.Vertex3 (1, 0, 0.1f);
                GL.Vertex3 (1, 1, 0.1f);
                GL.Vertex3 (0, 1, 0.1f);
                GL.End ();
            }
            GL.PopMatrix ();
        }*/

    }

    public class CameraFilterGrayscale : ImageEffectBase {
        public Texture textureRamp;
        public float rampOffset;
        private bool supported;

        public void Awake() {
            shader = Shader.Find("Hidden/Grayscale Effect");
            textureRamp = Util.LoadTexture(256, 1, "ramp_grayscale.png");
            material.SetTexture("_RampTex", textureRamp);
            material.SetFloat("_RampOffset", rampOffset);
            supported = shader.isSupported;
        }
        public void OnRenderImage(RenderTexture source, RenderTexture dest) {
            if (supported)
                Graphics.Blit(source, dest, material);
        }
    }

    public class CameraFilterSepiatone : ImageEffectBase {
        private bool supported;
        public void Awake() {
            shader = Shader.Find("Hidden/Sepiatone Effect");
            supported = shader.isSupported;
        }
        void OnRenderImage(RenderTexture source, RenderTexture dest) {
            if(supported)
                Graphics.Blit(source, dest, material);
        }
    }

    /*public class RenderTextureGrayscale : ImageEffectBase {
        public Texture textureRamp;
        public float rampOffset;
        

        public RenderTextureGrayscale() {
            shader = Shader.Find("Hidden/Grayscale Effect");
            textureRamp = Util.LoadTexture(256, 1, "ramp_grayscale.png");
        }

        public void Blit(RenderTexture source, ref RenderTexture dest) {
            //RenderTexture dest = new RenderTexture(source.width, source.height, source.depth);
            material.SetTexture("_RampTex", textureRamp);
            material.SetFloat("_RampOffset", rampOffset);
            Graphics.Blit(source, dest, material);
            //return dest;
        }
    }*/
}

