﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {
    class GUIcontroller {

        private LazorSystem system;
        private Rect windowPos;
        public LinkedList<GUItab> guiButtons = new LinkedList<GUItab>();
        public Rect popupPos;
        private int id = 1337;
        private int totalTabs;
        private bool toggleCollapsed;


        public GUIcontroller(LazorSystem system) {
            this.system = system;
            popupPos = new Rect(Screen.width / 2, Screen.height / 2, 10, 10);
        }

        public void CountTabs() {
            totalTabs = 0;
            Count(system.guiTargets);
            Count(system.guiSubSystems);
            Count(system.guiAchievements);
            Count(system.guiModules);
            Count(system.guiConfig);
            Count(system.guiParameters);
            Count(system.guiTargetCam);
        }
        private void Count(GUItab t){
            if (t.hasContent)
                totalTabs++;
        }

        public void Shuffle(GUItab tab) {
            guiButtons.Remove(tab);
            guiButtons.AddLast(tab);
        }
        public void Pop(GUItab tab) {
            guiButtons.Remove(tab);
        }

        //Generates "unique" window id
        public int GenerateID() {
            return id++;
        }

        private void DrawTargetIndicator(Component c, TargetIndicatorType tit, bool drawName) {
            if(c is Vessel)
                system.guiOverlays.DrawTargetIndicator(c.transform.position, tit, drawName ? (c as Vessel).vesselName : "");
            else
                system.guiOverlays.DrawTargetIndicator(c.transform.position, tit, drawName ? c.name : "");
        }

        public void DrawGUIs(){
			GUI.skin = Util.LazorSkin;
			float xOffSet = 0;

			//draw target indicator
			if(system.leaderLocked && system.Target != null)
				DrawTargetIndicator(system.Target, TargetIndicatorType.Target, true);
          
			//draw tracking indicator
			for(int i = 0; i<system.targetsTracked.Count; i++) {
				if(system.targetsTracked[i]==null){
					system.targetsTracked.RemoveAll(c => c == null);
					break;
				}
				if(system.targetsTracked[i] != system.Target)
					DrawTargetIndicator(system.targetsTracked[i], TargetIndicatorType.Tracking, true);
				else
					DrawTargetIndicator(system.targetsTracked[i], TargetIndicatorType.Tracking, false);
			}

            //draw crosshair
            if (system.toggleFreeAim && !FlightDriver.Pause)
                system.guiOverlays.DrawCrosshair();
           


            //draw 'Config Tab' aim target rings
            if (system.guiOverlays.toggleImpactDist || system.guiOverlays.toggleImpactAngle || system.guiOverlays.toggleExitDist || system.guiOverlays.toggleExitAngle) {
                foreach(int i in system.config.claimedAndActiveLazorGroups)
                    foreach (LazorAimable l in system.config.LazorCounterparts[i])
                        system.guiOverlays.DrawOverlayLazor(l.Lazor);
                system.guiOverlays.DrawOverlayLazor(system.Leader);
            }

            //draw in-game overlays
            system.guiOverlays.DrawOverlay();

            //draw map overlay 
            if (!system.GAMEPAUSED) {
                system.guiOverlays.DrawMapOverlay(); //CFG_blue
            }


            //draw Tabs (docked or not)
            foreach (GUItab gb in guiButtons)
                xOffSet = gb.GUIbutton(xOffSet);

            //draw undocked subsystem windows
            system.guiSubSystems.DrawWindows();


            //draw floating menu button window
            if (Event.current.type == EventType.Layout)
                windowPos = new Rect(system.windowPos.x - windowPos.width - xOffSet, system.windowPos.y, 120, 78);
            windowPos = GUILayout.Window(4, windowPos, WindowGUI, "", GUIStyle.none);

            //draw popups
			system.config.DrawCFGGuis();
            
            if (system.guiTargets.visiblePopUp) {
                GUI.skin = Util.LazorSkin;
                Popup(system.guiTargets.SDPopup);
            }
            if(system.guiTargets.visiblePopUp.In()){
                GUI.BringWindowToFront(10);
            }else if(system.guiTargets.visiblePopUp.Out()){ }
        }
        

        //Draw floating buttons
        private void WindowGUI(int windowID) {
            GUILayout.BeginHorizontal(Util.Space_Compact);
            GUILayout.BeginHorizontal(Util.Space_Compact);
            if (!toggleCollapsed) {
                GUILayout.BeginVertical(Util.Space_Compact);
                if (!system.LazorMainWindowOn && GUILayout.Button("Main Panel", Util.Button_bold))
                    system.LazorMainWindowOn.set = true;
                DrawFloatingButton(system.guiTargets);
                DrawFloatingButton(system.guiSubSystems);
                DrawFloatingButton(system.guiAchievements);
                DrawFloatingButton(system.guiModules);
                DrawFloatingButton(system.guiConfig);
                DrawFloatingButton(system.guiParameters);
                DrawFloatingButton(system.guiTargetCam);
                GUILayout.EndVertical();
            } else
                GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            if (totalTabs - guiButtons.Count != 0) {
                GUILayout.BeginHorizontal(Util.Space_Compact, GUILayout.Width(13));
                if (toggleCollapsed) {
                    if (GUILayout.Button("◄", Util.Button_floatingToggle, GUILayout.ExpandHeight(true), GUILayout.Width(10), GUILayout.MinHeight(78)))
                        toggleCollapsed = false;
                } else {
                    if (GUILayout.Button("►", Util.Button_floatingToggle, GUILayout.ExpandHeight(true), GUILayout.Width(10), GUILayout.MinHeight(78)))
                        toggleCollapsed = true;
                }
                GUILayout.Space(3);
                GUILayout.EndHorizontal();
            }
            GUILayout.EndHorizontal();
        }

        private void DrawFloatingButton(GUItab t){
            if (t.hasContent && !t.Enabled && GUILayout.Button(t.GetTabName(), Util.Button_bold))
                t.Enabled = true;
        }

        public void Popup(GUI.WindowFunction p) {
            popupPos = Util.ClampToScreen(GUILayout.Window(10, popupPos, p, "Warning", GUILayout.MinWidth(250)));
        }


        public void onFlightStateLoad(Dictionary<string, KSPParseable> parsedData) {
            if (parsedData.ContainsKey("GUIcontroller_toggleCollapsed"))
                toggleCollapsed = parsedData["GUIcontroller_toggleCollapsed"].value_bool;
            if (parsedData.ContainsKey("GUIcontroller_guiButtons")) {
                //MonoBehaviour.print("xxLOADxx " + parsedData["GUIcontroller_guiButtons"].value);
                List<GUItab> a = guiButtons.ToList<GUItab>();
                List<GUItab> b = new List<GUItab>();
                foreach (string s in parsedData["GUIcontroller_guiButtons"].value.Split('#'))
                    b.Add(a.Find(x => x.Name == s));
                guiButtons = new LinkedList<GUItab>(b);
            }

        }
        public void onFlightStateSave(Dictionary<string, KSPParseable> partDataCollection) {
            partDataCollection.Add("GUIcontroller_toggleCollapsed", new KSPParseable((bool)toggleCollapsed, KSPParseable.Type.BOOL));
            //tab ordering
            if (guiButtons != null && guiButtons.Count > 0) {
                string[] a = new string[guiButtons.Count];
                List<GUItab> b = guiButtons.ToList<GUItab>();
                for (int i = 0; i < guiButtons.Count; i++)
                    a[i] = b[i].Name;
                partDataCollection.Add("GUIcontroller_guiButtons", new KSPParseable(string.Join("#", a), KSPParseable.Type.STRING));
                //MonoBehaviour.print("xxSAVExx " + string.Join("#", a));
            }
        }




    }
}

