﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;
using KSP.IO;

namespace Romfarer {
    class GUIsubsystems : GUItab {

        public delegate void Content(int i);

        public GUIsubsystems(LazorSystem system) : base(system) { }

        override public string Name { get { return "Subsystems"; } }
        override public string GetTabName() {
			return "Subsystems (" + (system.config.cfgActive.Count) + ")";
        }

        override public float GUIbutton(float indent, int width = 200, int height = 100) {
            return base.GUIbutton(indent, 210, 50);
        }

        //Draw the GUI window.
        override public void DrawGUI(string name) {
            GUI.skin = Util.LazorSkin;
            windowPos = GUILayout.Window(id, windowPos, WindowGUI, name, Util.Window_border);
            if (!docked)
                windowPos = Util.ClampToScreen(windowPos);
        }

        //undocked subsystems windows and stuff
        public void DrawWindows(){
			hasContent = system.config.cfgActive.Count > 0;
			foreach (CFG c in system.config.cfgActive) {  //subsystem CFG windows
                if (!c.docked && !c.ConfiguredEVA) {
                    c.rect = Util.ClampToScreen(GUILayout.Window(c.id, c.rect, c.ContentGUI, c.name));
                    if (Event.current.type == EventType.Repaint) 
                        c.rect.height = 10;
                }
            }
        }

        override public void WindowGUI(int windowID) {
            Header();
            //Draw CFG expandable buttons
			foreach (CFG c in system.config.cfgActive)
                if(c.docked)
                    c.DrawGUIbutton();
            GUILayout.Space(1);
            Footer();
        }

        ///////////////////////////////////////////
        // Generic stuff
        //////////////////////////////////////////

        public bool SubsystemButton(bool toggle, string name, Color c, out bool docked, Content content, int id) {
            GUIStyle style = Util.Button_color_white;
            if (c == Color.red)
                style = Util.Button_color_red;
            else if (c == Color.green)
                style = Util.Button_color_green;
            else if (c == Color.blue)
                style = Util.Button_color_blue;
            else if (c == Color.yellow)
                style = Util.Button_color_yellow;
            else if (c == Color.magenta)
                style = Util.Button_color_magenta;
            GUILayout.BeginHorizontal(Util.Space_Compact);
                GUILayout.BeginVertical(Util.Space_Compact);
                    GUILayout.BeginHorizontal(toggle?Util.Divider:Util.Divider_gray);
                        GUILayout.BeginHorizontal(Util.Space_Compact, GUILayout.Width(20), GUILayout.Height(20));
                        docked = true;
                        if (toggle) {
                            if (GUILayout.Button("", Util.Divider_gray_button, GUILayout.Width(20), GUILayout.Height(20)))
                                docked = false;
                        } else {
                            if (GUILayout.Button(" ", GUIStyle.none, GUILayout.Width(20), GUILayout.Height(20)))
                                toggle = true;
                        }
                        GUILayout.EndHorizontal();
                        GUILayout.BeginHorizontal(Util.Space_Compact);
                        GUILayout.Label("  "+name, Util.LabelBold);
                        GUILayout.EndHorizontal();
                    GUILayout.EndHorizontal();
                if (toggle) {
                    GUILayout.Space(3);
                    content(id); //draw the meat of the panel
                    GUILayout.Space(7);
                }
                GUILayout.EndVertical();
                GUILayout.BeginHorizontal(Util.Space_Compact, GUILayout.Width(20));
                toggle = GUILayout.Toggle(toggle, "", style, GUILayout.ExpandHeight(true), GUILayout.Width(20));
                GUILayout.EndHorizontal();
            GUILayout.EndHorizontal();

            return toggle;
        }

        public bool Header(bool docked, Rect rect, bool enabled, ref DoubleSwitch enable) {
            enable.set = enabled;
            if (!docked) {
                if (GUI.Button(new Rect(rect.width - 20, 0, 20, 20), "", Util.Button_header_minimize))
                    return true;
                enable.set = GUI.Toggle(new Rect(rect.width - 40, 0, 20, 20), enable, "");
                GUI.enabled = enable;
            }
            GUI.backgroundColor = Color.white;
            return docked;
        }

        public void Footer(bool docked, Rect rect) {
            if (!docked)
                GUI.DragWindow(new Rect(0, 0, rect.width, 30));
            GUI.enabled = true;
        }

        public static void DividerTop() {
            GUILayout.BeginHorizontal();
            GUILayout.Space(10);
            GUILayout.BeginVertical();
        }

        public static void DividerBottom() {
            GUILayout.EndVertical();
            GUILayout.Space(10);
            GUILayout.EndHorizontal();
        }

        public void onFlightStateLoad(Dictionary<string, KSPParseable> parsedData) {
            onFlightStateLoad(parsedData, "GUIsubsystems");
        }

        public void onFlightStateSave(Dictionary<string, KSPParseable> partDataCollection) {
            onFlightStateSave(partDataCollection, "GUIsubsystems");
        }

    }
}
