﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {
    class GUItargets : GUItab {

        private bool[] visibleTargetTypes;
        private string focusedListItem = "";
        private Vector2 v2 = new Vector2(0,0);
        private Rect area = new Rect(0, 0, 100, 100);
        private List<GUItargetsList.Buttons> standardButtons = new List<GUItargetsList.Buttons>(); //buttons displayed for every item
        
        private Texture2D textureHighlight = Util.getTexture(Color.yellow);
        private Texture2D textureNormal = Util.getTexture(Color.clear);
        private GUIStyle listItemRight;
        private GUIStyle listItem;
        private GUIStyle listItemSelected;
        private GUIStyle listItemHeader;
        private int index = 23;
        private Color colorHeader = new Color(0.282f, 1f, 0.803f, 1);

        public CFGI_KPS locationList;
        public bool toggleKSPmap;

        private TargetModes targetMode = 0;
        private enum TargetModes { Targets = 0, Locations = 1 };

        private Stack<Part> deathStack;
        public DoubleSwitch visiblePopUp;
        private GUItargetsListItem activeGUItargetsListItem;
        private Vessel deathVessel;
        private List<Part> deathList;



        public GUItargets(LazorSystem system) : base(system) {
            visibleTargetTypes = new bool[Enum.GetValues(typeof(TargetType)).Length];
            for (int i = 0; i < visibleTargetTypes.Length; i++)
                visibleTargetTypes[i] = true;
            AddStandardButton(StdButtonTarget);
            AddStandardButton(StdButtonTrack);
            AddStandardButton(StdButtonFly);
            AddStandardButton(ButtonSefDestruct);
            
            locationList = new CFGI_KPS(system);
        }

        override public string Name { get { return "Targets"; } }
        override public string GetTabName() {
            return system.targetList.list.Count > 0 ? "Targets (" + system.targetList.list.Count + ")" : "Targets";
        }

        override public float GUIbutton(float indent, int width = 200, int height = 100) {
            return base.GUIbutton(indent, 210);
        }

        private void AddStandardButton(GUItargetsList.Buttons b) {
            standardButtons.Add(b);
        }

        /*public void RemoveStandardButton(GUItargetsList.Buttons b) {
            standardButtons.Remove(b);
        }*/

        

        //Draw the GUI window.
        override public void DrawGUI(string name) {
            GUI.skin = Util.LazorSkin;

            //KPS
            if (toggleKSPmap) {
                double lat, lon;
                if (system.guiOverlays.MapSelect(out lat, out lon)) {
                    toggleKSPmap = false;
                    locationList.StdButtonAdd(new KPSmapLocation("", "", lat, lon, system.vessel.mainBody));
                }
            }
            locationList.ExecuteTasks(); //Execute KPS queued tasks.


            windowPos = GUILayout.Window(id, windowPos, WindowGUI, name, Util.Window_border);
            if (!docked)
                windowPos = Util.ClampToScreen(windowPos);

/*

            if (visiblePopUp) {
                GUI.skin = Util.LazorSkin;
                //windowPos = GUILayout.Window(10, windowPos, SDPopup, "Warning", GUILayout.MinWidth(250));
                system.guiController.Popup(SDPopup);
            }*/
        }





        override public void WindowGUI(int windowID) {
            Header();

            if (listItem == null) {
                listItem = Util.Label(12, Color.white, Color.yellow, new Color(0.1f, 0.1f, 0.1f, 0.9f));
                listItem.stretchWidth = true;
            }
            if (listItemSelected == null) {
                listItemSelected = Util.Label(12, Color.white, Color.yellow, new Color(0.1f, 0.2f, 0.2f, 0.9f));
                listItemSelected.stretchWidth = true;
            }
            if (listItemHeader == null) {
                listItemHeader = Util.Label(12, colorHeader, Color.yellow, new Color(0.2f, 0.2f, 0.2f, 0.9f));
                listItemHeader.stretchWidth = true;
            }

            if (listItemRight == null) {
                listItemRight = new GUIStyle(listItem);
                listItemRight.alignment = TextAnchor.MiddleRight;
                listItemRight.stretchWidth = false;
            }

            GUILayout.BeginVertical(Util.Space_Formatted);

            targetMode = (TargetModes)GUILayout.SelectionGrid((int)targetMode, Enum.GetNames(typeof(TargetModes)), 2);
            if (targetMode == TargetModes.Targets) {
                //Target checkboxes
                GUILayout.BeginHorizontal(Util.Space_Compact);
                for (int i = 0; i < visibleTargetTypes.Length; i++)
                    visibleTargetTypes[i] = GUILayout.Toggle(visibleTargetTypes[i], Enum.GetNames(typeof(TargetType))[i].ToString().Substring(0, 3), Util.Button_Toggle);
                GUILayout.EndHorizontal();

            }
            GUILayout.EndVertical();

            //GUILayout.Space(5);

            //BEGIN LIST
            float height = area.height > 400 ? 410 : area.height+10;
            v2 = GUILayout.BeginScrollView(v2, false, false, GUILayout.MinHeight(height), GUILayout.ExpandHeight(false));//, Util.TxtAreaBars, Util.TxtAreaBars, Util.TxtArea, GUILayout.MinHeight(height), GUILayout.ExpandHeight(false));
            GUILayout.BeginVertical(GUILayout.MaxWidth(windowPos.width - 22));

            if (targetMode == TargetModes.Locations) { //Draw list of locations
                GUILayout.Label("Locations", listItemHeader);
                index++;
                foreach (KPSmapLocation mapLoc in locationList.Get()) {
                    GUILayout.BeginHorizontal();
                    if (GUILayout.Button(mapLoc.name, focusedListItem == mapLoc.id ? listItemSelected : listItem) || GUILayout.Button(mapLoc.cb.name, listItemRight))
                        focusedListItem = mapLoc.id;
                    GUILayout.EndHorizontal();
                    index++;
                }
            } else { //Draw list of targets
                for (int i = 0; i < visibleTargetTypes.Length; i++) {
                    if (visibleTargetTypes[i]) {
                        TargetType t = (TargetType)Enum.GetValues(typeof(TargetType)).GetValue(i);
                        List<GUItargetsListItem> category = system.targetList.GetByTargetType(t);
                        if (category.Count > 0) {
                            GUILayout.Label(t.ToString().Replace('_', ' '), listItemHeader);
                            index++;
                            foreach (GUItargetsListItem tli in category) {
                                if (tli.c == null) {
                                    system.targetList.AddTask(tli, system.targetList.Remove);
                                } else {
                                    if (tli.c is Vessel && system.IsActiveVessel((Vessel)tli.c)) {
                                        listItem.normal.textColor = Color.green;
                                        if (GUILayout.Button(tli.name, listItem))
                                            focusedListItem = tli.id;
                                        listItem.normal.textColor = Color.white;
                                    } else {
                                        if (tli.c == system.Target && system.targetsTracked.Contains(tli.c)) {
                                            GUILayout.BeginHorizontal();
                                            GUILayout.Label("(X)", Util.StdLabelSmallCustom(Color.red));
                                            if (GUILayout.Button(tli.name, focusedListItem == tli.id ? listItemSelected : listItem))
                                                focusedListItem = tli.id;
                                            GUILayout.EndHorizontal();
                                        } else if (tli.c == system.Target) {
                                            GUILayout.BeginHorizontal();
                                            GUILayout.Label("X", Util.StdLabelSmallCustom(Color.red));
                                            if (GUILayout.Button(tli.name, focusedListItem == tli.id ? listItemSelected : listItem))
                                                focusedListItem = tli.id;
                                            GUILayout.EndHorizontal();
                                        } else if (system.targetsTracked.Contains(tli.c)) {
                                            GUILayout.BeginHorizontal();
                                            GUILayout.Label("()", Util.StdLabelSmallCustom(Color.red));
                                            if (GUILayout.Button(tli.name, focusedListItem == tli.id ? listItemSelected : listItem))
                                                focusedListItem = tli.id;
                                            GUILayout.EndHorizontal();
                                        } else {
                                            if (GUILayout.Button(tli.name, focusedListItem == tli.id ? listItemSelected : listItem))
                                                focusedListItem = tli.id;
                                        }
                                    }
                                }
                                index++;
                            }

                        }
                    }
                }
            }
            GUILayout.EndVertical();
            if (Event.current.type == EventType.Repaint)
                area = GUILayoutUtility.GetLastRect();
            GUILayout.EndScrollView();
            //END LIST

            GUILayout.Space(5);
            
            //Draw buttons
            GUILayout.BeginVertical(Util.Space_Formatted);

            if (targetMode == TargetModes.Targets) {
                //Target buttons
                foreach (GUItargetsListItem tli in system.targetList.list) {
                    if (focusedListItem == tli.id) {
                        GUILayout.BeginHorizontal();
                        foreach (GUItargetsList.Buttons b in standardButtons)
                            b(tli);
                        if (tli.buttons != null) //custom buttons
                            for (int i = 0; i < tli.buttons.Length; i++)
                                tli.buttons[i](tli);
                        GUILayout.EndHorizontal();
                        //EVA buttons........................................
                        if (tli.c is Vessel && ((Vessel)tli.c).isEVA) {
                            Util.GUIButton_BeamAboard(system, (Vessel)tli.c);
                        }
                    }
                }
                if (GUILayout.Button("Refresh Vessel List")) {
                    foreach (Vessel v in FlightGlobals.Vessels)
                        system.targetList.AddReplacement(new GUItargetsListItem("" + v.id, TargetType.Vessel, v.vesselName, v, 0, 0));
                }
            } else {
                //KPS location buttons
                GUILayout.BeginHorizontal();
                foreach (KPSmapLocation mapLoc in locationList.Get()) {
                    if (focusedListItem == mapLoc.id)
                        foreach (CFGI_KPS.Buttons b in locationList.standardButtons)
                            b(mapLoc);
                }
                GUILayout.EndHorizontal();

                GUILayout.Label("Add location: ", Util.StdLabelSmallCustom(Color.yellow));
                GUILayout.BeginHorizontal(Util.Space_Compact);
                toggleKSPmap = GUILayout.Toggle(toggleKSPmap, "Select on map", Util.Button_Toggle);
                if (GUILayout.Button("This location"))
                    locationList.StdButtonAdd(new KPSmapLocation("", "", system.LATITUDE, system.LONGITUDE, system.vessel.mainBody));
                GUILayout.EndHorizontal();
                GUILayout.BeginHorizontal();
                if (GUILayout.Button("Lazor hitpoint")) {
                    Vector3d surfacePos;
                    if (Vector3.Dot(system.Leader.GetAim(), system.vessel.upAxis) < 0 && PQS.LineSphereIntersection(system.Leader.GetOrigin() - system.vessel.mainBody.position, system.Leader.GetAim(), system.vessel.mainBody.Radius, out surfacePos)) {
                        surfacePos += system.vessel.mainBody.position;
                        locationList.StdButtonAdd(new KPSmapLocation("", "", system.vessel.mainBody.GetLatitude(surfacePos), system.vessel.mainBody.GetLongitude(surfacePos), system.vessel.mainBody));
                    }
                }
                if (GUILayout.Button("Manual input")) {
                    locationList.StdButtonAdd(new KPSmapLocation("", "", 0, 0, system.vessel.mainBody));
                }
                GUILayout.EndHorizontal();

                GUI.enabled = system.Target != null && system.vessel.mainBody.GetAltitude(system.Target.transform.position) < 10000;
                if (GUILayout.Button("Target location"))
                    locationList.StdButtonAdd(new KPSmapLocation("", system.Target.name, system.vessel.mainBody.GetLatitude(system.Target.transform.position), system.vessel.mainBody.GetLongitude(system.Target.transform.position), system.vessel.mainBody));
                GUI.enabled = true;
                system.guiOverlays.toggleKPSlocations = GUILayout.Toggle(system.guiOverlays.toggleKPSlocations, "Show locations on map");

            }
            GUILayout.EndVertical();

            //Execute queued tasks
            system.targetList.ExecuteTasks();
            

            Footer();
        }







        /////////////////////////////////////////////////////////////////////
        // BUTTONS
        /////////////////////////////////////////////////////////////////////

        public virtual void StdButtonTarget(GUItargetsListItem i) {
            if (!(i.c is Vessel && system.IsActiveVessel((Vessel)i.c)))
                if (system.Target == i.c) {
                    if (GUILayout.Button("UnTarget"))
                        system.Target = null;
                } else {
                    if (GUILayout.Button("Target")) {
                        system.Target = i.c;
                    }
                }
        }

        public void StdButtonTrack(GUItargetsListItem i) {
            if (!(i.c is Vessel && system.IsActiveVessel((Vessel)i.c))) {
                if (!system.targetsTracked.Contains(i.c)) {
                    if (GUILayout.Button("Track"))
                        system.TrackTarget(i.c);
                } else {
                    if (GUILayout.Button("UnTrack"))
                        system.UnTrackTarget(i.c);
                }
            }
        }

        virtual public void StdButtonFly(GUItargetsListItem i) {
            if (i.c is Vessel && !system.IsActiveVessel((Vessel)i.c)) {
                if (GUILayout.Button("Fly")) {
                    FlightGlobals.SetActiveVessel((Vessel)i.c); //change vessel
                }
            } else if (i.c is Vessel && FlightGlobals.ActiveVessel != system.vessel) {
                if (GUILayout.Button("Set Focus")) {
                    FlightGlobals.SetActiveVessel((Vessel)i.c); //change vessel
                }
            }
        }

        //button that opens flag in DrawGUI -> WindowGUI
        public void ButtonSefDestruct(GUItargetsListItem i) {
            if (i.c != null && i.c is Vessel) {
                if (GUILayout.Button("S-D")) {
                    deathVessel = (Vessel)i.c;
                    activeGUItargetsListItem = i;
                    visiblePopUp.set = true;
                }
            }
        }
        /*
        override public void DrawGUI() { // draw the popup window
            if (visiblePopUp) {
                GUI.skin = Util.LazorSkin;
                GUI.depth = -1;
                windowPos = GUILayout.Window(10, windowPos, SDPopup, "Warning", GUILayout.MinWidth(250));
            }
        }*/

        public void SDPopup(int windowID) {
            GUILayout.BeginVertical();
            GUILayout.Label("DESTROY this vessel?");
            GUILayout.BeginHorizontal();
            if (GUILayout.Button("Yes")) {
                visiblePopUp.set = false;
                if (system.Target = deathVessel)
                    system.Target = null;
                system.UnTrackTarget(deathVessel);
                if (deathVessel.rootPart == null) { //if it is far away.
                    deathVessel.Die();
                } else {
                    deathList = new List<Part>();
                    deathList.Add(DeptFirstParts(deathVessel.rootPart));
                    deathList.Reverse();
                    deathStack = new Stack<Part>(deathList);
                    while (deathStack.Count > 0) {
                        Part p = deathStack.Pop();
                        MonoBehaviour.print("+++++exploded++++++ " + p.name + " ++ " + p.GetType());
                        if (p != null && p is Parachutes) {
                            p.Die();
                        } else if (p != null) {
                            p.explode();
                        }
                    }
                }
                system.targetList.AddTask(activeGUItargetsListItem, system.targetList.Remove);
            }
            if (GUILayout.Button("NO"))
                visiblePopUp.set = false;
            GUILayout.EndHorizontal();
            GUILayout.EndVertical();
        }

        private Part DeptFirstParts(Part p) {
            if (p.children != null || p.children.Count != 0)
                foreach (Part q in p.children)
                    deathList.Add(DeptFirstParts(q));
            return p;
        }



        public void onFlightStateLoad(Dictionary<string, KSPParseable> parsedData) {
            onFlightStateLoad(parsedData, "GUItargets");
        }
        public void onFlightStateSave(Dictionary<string, KSPParseable> partDataCollection) {
            onFlightStateSave(partDataCollection, "GUItargets");
        }



    }


 



}
