/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
/*using System.Collections.Generic;
using System.Linq;
using System.Text;*/
using UnityEngine;

namespace Romfarer {

	
	//public void SetWaypoint(Vector3 tgtPos) in KerbalEVA to move kerbal

    class LazorEVA : PartModule {

        [KSPField(isPersistant=true)]
        public string myVessel = "-1";

        public LazorSystem system;
        public Guid vesselGuid = new Guid();
		private DoubleSwitch gui_expanded;
        private DoubleSwitch toggleFPS;
        private DoubleSwitch toggleJetpack;
        private DoubleSwitch toggleLamp;
        private Rect windowPos = new Rect(0, 0, 0, 0);
        private KerbalEVA kerbalEVA;

        //Teleport
        private Vector3 surfacePos; //teleport destination
        private GameObject teleportGO;
        private ParticleEmitter teleportEmitter;
        private bool teleOutShort;
        private bool teleOutShort2;
        private bool teleOutShortOngoing;
        private bool teleIn;
        private bool teleIn2;
        private bool teleOut;
        private bool teleOut2;
        private bool teleOut3;
        private bool teleOutOngoing;
        private float timerStart;
        private float timerTick;

        //Cameras
        private FlightCamera flightCamera;
        private GameObject pivotDummy;
        private float cam_minDistance;
        private float cam_maxDistance;
        private float cam_startDistance;
        private float cam_minPitch = -0.2f;
        private float cam_maxPitch = 0.2f;
        private float cam_pivotTranslateSharpness = 50;
        private float cam_minHeight = 0f;
        private float cam_minHeightAtMaxDist = 0f;
        private float cam_minHeightAtMinDist = 0f;

		//Lazor
		private DoubleSwitch toggleLazor;
		private bool lazorInitiated;
		private Lazor lazor;
		private GameObject lazorAttach;
		private DoubleSwitch toggleMouseLook_Debug; //debug lazor to display info of part
		private DoubleSwitch toggleMouseLook_Explode;
		private DoubleSwitch toggleMouseLook_Decouple;
		private DoubleSwitch toggleMouseLook_Weld;
		private bool toggleMouseLook_WeldSuccess;
		private bool toggleMouseLook_WeldFailure;
		private string toggleMouseLook_WeldFailure_msg;
		private Part weld1; //first part clicked when welding
		private Part weld2; //second part clicked when welding
		private Part hitPart; //last part selected (hover over) when welding
		private double weldDistance; //distance between weld1 and hitPart

		//fuel
		private float fuelReqForLazorEffects = 0.1f;



        [KSPEvent]
        public void Teleport(BaseEventData data) {
            surfacePos = data.Get<Vector3>("destination");
            teleOut = true;
            teleOutOngoing = true;
            TeleportAnimation<LazorEVA>(this);
        }

        [KSPEvent]
        public void TeleportShort(BaseEventData data) {
            surfacePos = data.Get<Vector3>("destination");
            teleOutShort = true;
            teleOutShortOngoing = true;
            TeleportAnimation<LazorEVA>(this);
        }

        private void TeleportInStart() {
            teleIn = true;
            TeleportAnimation<LazorEVA>(this);
            system.lazorSound.Play(SoundType.TELEPORT);
            MonoBehaviour.print("[Lazor EVA: Beaming " + FlightGlobals.ActiveVessel.name + " aboard]");
        }

        private void TeleportInEnd() {
            DisposeCams();
            if (system != null) {
                kerbalEVA.BoardPart(system.vessel.rootPart);
                system.vessel.SpawnCrew();
                system.vessel.ResumeStaging();
                system.vessel.MakeActive();
                MonoBehaviour.print("[Lazor EVA: Teleport aboard completed]");
            }
        }

		//[KSPAction("ToggleMouseLookTest", KSPActionGroup.Custom01)]
		/*public void ToggleMouseLookTest(KSPActionParam param) {
			if (toggleLazor && !MapView.MapIsEnabled) {
				toggleFreeAim = !toggleFreeAim;
				//leaderLocked = false;
			}
			MonoBehaviour.print("xxx");
		}*/

        public override void OnStart(PartModule.StartState state) {
            MonoBehaviour.print("[Lazor EVA: system started on " + vessel.vesselName + "]");
			Romfarer.Events.onVesselChange += OnVesselChange;
            Initialize();
        }

        /*public override void OnAwake() {
            MonoBehaviour.print("[Lazor EVA: system awoken]");
			Romfarer.Events.onVesselChange += OnVesselChange;
            Initialize();
        }*/



		private Transform anchor;
        private void Initialize() {
			isEnabled = true;
            if (vessel.rootPart != null) {
				gui_expanded.set = true;
				RenderingManager.AddToPostDrawQueue(0, new Callback(drawGUI));
                windowPos = new Rect(Screen.width - 120, 0, 120, 80);
                //kerbalEVA = (KerbalEVA)vessel.rootPart.Modules["KerbalEVA"];
				kerbalEVA = (KerbalEVA)part.Modules["KerbalEVA"];
				flightCamera = FlightCamera.fetch;
				pivotDummy = new GameObject("LazorEva: pivotDummy");
                lazor = new Lazor();
                
				/*foreach (Component c in vessel.transform.GetComponentsInChildren(typeof(Transform)))
					if(c.name == "helmetCollider")//"helmet01")
						anchor = c as Transform;*/
				foreach (Component c in part.transform.GetComponentsInChildren(typeof(Transform)))
					if(c.name == "helmetCollider")//"helmet01")
						anchor = c as Transform;
				pivotDummy.transform.parent = anchor;
				pivotDummy.transform.localPosition = Vector3.zero;
				pivotDummy.transform.rotation = anchor.rotation;

				/*part.Actions.Add(new BaseAction(Actions, "ToggleMouseLookTest", ToggleMouseLookTest, new KSPAction("ToggleMouseLookTest", KSPActionGroup.Custom01)));
				MonoBehaviour.print("XXXcountXXXX "+part.Actions.Count);
				foreach(BaseAction ba in part.Actions)
					MonoBehaviour.print("XXXXXXX "+ba.name + " " + ba.guiName);*/

				/*MonoBehaviour.print("XXXXXXX "+BaseAction.CreateActionList(vessel.parts, KSPActionGroup.Custom01, true).Count);
				foreach(BaseAction ba in BaseAction.CreateActionList(vessel.parts, KSPActionGroup.Custom01, true))
					MonoBehaviour.print("XXXXXXX "+ba.name + " " + ba.guiName);
				MonoBehaviour.print( new KSPAction("ToggleMouseLookTest", KSPActionGroup.Custom01));*/
				MonoBehaviour.print("[Lazor EVA: Initialized]");
            }else
                MonoBehaviour.print("[Lazor EVA: NOT Initialized] -> rootpart null"); //when would this happen? hmmm =p
        }

        private void drawGUI() {
			if(FlightGlobals.ActiveVessel != vessel || !FlightGlobals.ActiveVessel.isEVA || myVessel == "-1")
				return;
			if(system == null)
				system = FindSystem(myVessel);
			GUI.skin = Util.LazorSkin;
			if(gui_expanded.In()) {
				windowPos.width = 120;
			} else if(gui_expanded.Out()) {
				//windowPos.x += 30;
				windowPos.width = 90;
				windowPos.height = 42;
			}
			if(gui_expanded) {
				windowPos = Util.ClampToScreen(GUILayout.Window(part.GetInstanceID(), windowPos, WindowGUI, "Lazor EVA", Util.Window_eva));
				if(Event.current.type == EventType.Repaint){
					windowPos.height = 80;
					windowPos.width = 120;
				}
			} else
				windowPos = Util.ClampToScreen(GUILayout.Window(part.GetInstanceID(), windowPos, WindowGUIEVAbutton, "Lazor EVA  ", Util.Window_eva_mini));
        }

        private void WindowGUIEVAbutton(int windowID) {
            GUILayout.BeginVertical(Util.Space_Compact);
            gui_expanded.set = GUI.Toggle(new Rect(windowPos.width - 20, 0, 20, 20), gui_expanded, "");
            GUILayout.EndVertical();
            GUI.DragWindow(new Rect(0, 0, 200, 30));
        }

        private void WindowGUI(int windowID) {
			gui_expanded.set = GUI.Toggle(new Rect(windowPos.width - 20, 0, 20, 20), gui_expanded, "");
			GUILayout.BeginVertical(Util.ListItemBG_compact);
			if(system == null) {
				GUILayout.BeginHorizontal(Util.Space_Compact);
				GUILayout.Label("Vessel Lost! ", Util.StdLabelSmallCustom(Color.red));
				GUILayout.EndHorizontal();
				GUILayout.EndVertical();
			} else {
				GUILayout.Label("Vessel distance: ", Util.StdLabelSmallCustom(Color.green));
				GUILayout.BeginHorizontal(Util.Space_Compact);
				Util.DistanceFormatted((system.vessel.transform.position - vessel.transform.position).magnitude, Color.white, "");
				GUILayout.EndHorizontal();
				GUILayout.EndVertical();
				if(GUILayout.Button("Beam me aboard")) {
					TeleportInStart();
				}
				if(GUILayout.Button("Switch to vessel")) {
					DisposeCams();
					FlightGlobals.SetActiveVessel(system.vessel);
				}
			}

			GUILayout.Space(5);
			toggleFPS.set = GUILayout.Toggle(toggleFPS, "EVA view");
			if(toggleFPS.In()) {
				foreach(Component c in vessel.transform.GetComponentsInChildren(typeof(Transform)))
					if(c.name.Contains("headMesh") || c.name.Contains("eyeball") || c.name.Contains("upTeeth") || c.name.Contains("pupil"))
						c.gameObject.renderer.enabled = false;
				cam_minPitch = flightCamera.minPitch;
				cam_maxPitch = flightCamera.maxPitch;
				cam_pivotTranslateSharpness = flightCamera.pivotTranslateSharpness;
				cam_minHeight = flightCamera.minHeight;
				cam_minHeightAtMaxDist = flightCamera.minHeightAtMaxDist;
				cam_minHeightAtMinDist = flightCamera.minHeightAtMinDist;
				cam_minDistance = flightCamera.minDistance;
				cam_maxDistance = flightCamera.maxDistance;
				cam_startDistance = flightCamera.startDistance;
				flightCamera.minPitch = -0.2f;
				flightCamera.maxPitch = 0.2f;
				flightCamera.pivotTranslateSharpness = 50;
				flightCamera.minHeight = 0f;
				flightCamera.minHeightAtMaxDist = 0f;
				flightCamera.minHeightAtMinDist = 0f;
				flightCamera.minDistance = 0.2f;
				flightCamera.maxDistance = 0.2f;
				flightCamera.startDistance = 0.2f;
				flightCamera.SetDistance(0.2f);
				//pivotDummy.transform.parent = kerbalEVA.headLamp.transform;
				//pivotDummy.transform.parent = anchor;
				//pivotDummy.transform.localPosition = Vector3.zero - (Vector3.up * 0.4f) - Vector3.forward * 0.3f;
				flightCamera.transform.parent.parent = pivotDummy.transform;
				//flightCamera.transform.localPosition = Vector3.zero - (Vector3.up * 0.4f) - Vector3.forward * 0.3f;
			} else if(toggleFPS.Out()) {
				DisposeCams(true);
			}
			if(toggleFPS) {
				/*MonoBehaviour.print(Util.AngleSigned(Vector3.Cross(kerbalEVA.headLamp.transform.forward, vessel.transform.up), 
                                                    Vector3.Cross(flightCamera.transform.forward, vessel.transform.up), 
                                                    vessel.transform.up));*/
			}

			GUILayout.Space(5);
			GUILayout.BeginVertical(Util.ListItemBG_compact);
			GUILayout.BeginHorizontal();
			GUILayout.Label("Fuel: ", Util.StdLabelSmallCustom(Color.green));
			GUILayout.Label(kerbalEVA.Fuel.ToString("0.00") + "/" + kerbalEVA.FuelCapacity.ToString("0.00"), Util.StdLabelSmallCustom(Color.white));
			GUILayout.Label("U", Util.StdLabelSmallCustom(Color.green));
			GUILayout.EndHorizontal();
			GUILayout.EndVertical();

			//jetpack
			toggleJetpack.set = kerbalEVA.JetpackDeployed = GUILayout.Toggle(kerbalEVA.JetpackDeployed, "Jetpack");
			if(toggleJetpack.In()) {
				animation.Play("openJetpack");
				animation.Stop("idle");
				animation.PlayQueued("jp_idle");
				animation["jp_idle"].wrapMode = WrapMode.Loop;
			} else if(toggleJetpack.Out()) {
				animation.Play("closeJetpack");
				animation.Stop("jp_idle");
				animation.PlayQueued("idle");
				animation["idle"].wrapMode = WrapMode.Loop;
			}

			//lamp
			toggleLamp.set = kerbalEVA.lampOn = GUILayout.Toggle(kerbalEVA.lampOn, "Lamp");
			if(toggleLamp.In()) {
				kerbalEVA.headLamp.SetActive(true);
			} else if(toggleLamp.Out()) {
				kerbalEVA.headLamp.SetActive(false);
			}

			//lazor
			if(kerbalEVA.Fuel < fuelReqForLazorEffects)
				GUI.enabled = false;
			toggleLazor.set = GUILayout.Toggle(toggleLazor.set, "Lazor");
			if(!toggleLazor) {
				toggleMouseLook_Explode.set = false;
				toggleMouseLook_Decouple.set = false;
				toggleMouseLook_Weld.set = false;
				toggleMouseLook_Debug.set = false;
				GUI.enabled = false;
			}
			toggleMouseLook_Explode.set = GUILayout.Toggle(toggleMouseLook_Explode, "\"1\" Explode");
			if(toggleMouseLook_Explode) {
				GUILayout.BeginVertical(Util.ListItemBG_compact);
				if(hitPart != null)
					GUILayout.Label(hitPart.partInfo.title);
				else
					GUILayout.Label("Nothing hit");
				GUILayout.EndVertical();
			}
			toggleMouseLook_Decouple.set = GUILayout.Toggle(toggleMouseLook_Decouple, "\"2\" Decouple");
			if(toggleMouseLook_Decouple) {
				GUILayout.BeginVertical(Util.ListItemBG_compact);
				if(hitPart != null)
					GUILayout.Label(hitPart.partInfo.title);
				else
					GUILayout.Label("Nothing hit");
				GUILayout.EndVertical();
			}

			toggleMouseLook_Weld.set = GUILayout.Toggle(toggleMouseLook_Weld, "\"3\" Weld");

			if(toggleMouseLook_Weld || toggleMouseLook_WeldSuccess || toggleMouseLook_WeldFailure) {
				GUILayout.BeginVertical(Util.ListItemBG_compact);
				if(weld1 == null){
					ToggleGui("weld 1:");
					GUILayout.Label(hitPart != null ? hitPart.partInfo.title : "Nothing hit");
				}else{
					GUILayout.Label("weld 1:");
					GUILayout.Label(weld1.partInfo.title, Util.StdLabelSmallCustom(Color.green));
				}
				if(weld1 != null){
					if(weld2 == null){
						ToggleGui("weld 2:");
						GUILayout.Label(hitPart != null ? hitPart.partInfo.title : "Nothing hit");
						GUILayout.Label("Distance");
						GUILayout.Label(weldDistance.ToString("f4") + "m");
					}else{
						GUILayout.Label("weld 2:");
						GUILayout.Label(weld2.partInfo.title, Util.StdLabelSmallCustom(Color.green));
					}
					if(toggleMouseLook_WeldSuccess)
						GUILayout.Label("Welding success", Util.StdLabelSmallCustom(Color.green));
					else if(toggleMouseLook_WeldFailure){
						GUILayout.Label("Welding failure: ", Util.StdLabelSmallCustom(Color.red));
						GUILayout.Label(toggleMouseLook_WeldFailure_msg);
					}
				}
				GUILayout.EndVertical();
			}

			toggleMouseLook_Debug.set = GUILayout.Toggle(toggleMouseLook_Debug, "\"0\" Debug beam");
			if(toggleMouseLook_Debug) {
				GUILayout.BeginVertical(Util.ListItemBG_compact);
				if(hitPart != null)
					GUILayout.Label(hitPart.partInfo.title);
				else
					GUILayout.Label("Nothing hit");
				GUILayout.EndVertical();
			}

			GUI.enabled = true;
			GUILayout.Space(5);

			#region debug
			/*
			if (GUILayout.Button ("print"))
				foreach (Component c in vessel.transform.GetComponentsInChildren(typeof(Transform))) {
					if (c.name.Contains ("helmet")) {
						MonoBehaviour.print (c.name + " " + c.GetType ());
					}
					if (c.name.Contains ("visor")) {
						MonoBehaviour.print (c.name + " " + c.GetType ());
					}
				}

			if (GUILayout.Button ("print2"))
				MonoBehaviour.print (kerbalEVA.headLamp.transform.parent.name);


			if(GUILayout.Button("cam"))
				foreach(Component c in flightCamera.gameObject.GetComponentsInChildren(typeof(Component)))
					MonoBehaviour.print(c.name + " " + c.GetType());


			if (GUILayout.Button ("print")) {
				pivotDummy.AddComponent<Rigidbody> ();
				pivotDummy.rigidbody.mass = 0.1f;
				pivotDummy.rigidbody.isKinematic = true;
			}

			
			if(lazorAttach != null)
				lazorAttach.transform.localPosition  = Util.MovePoint(lazorAttach.transform.localPosition , 0.002f);
			*/
			/*
			if(GUILayout.Button("print"))
				foreach(Component c in kerbalEVA.GetComponentsInChildren(typeof(Component)))
					MonoBehaviour.print(c.name + " " + c.GetType());*/
			#endregion
            GUI.DragWindow(new Rect(0, 0, 200, 30));
        }

		#region toggle GUILayout anim
		private float timer = 0;
		private int counter = 0;
		private void ToggleGui(string n){
			if(Toggle())
				GUILayout.Label(n);
			else
				GUILayout.Label(n + " <-", Util.StdLabelSmallCustom(Color.green));
		}
		private bool Toggle() {
			if(counter > 1000)
				counter = 0;

			if(Time.time > timer + 0.25) {
				timer = Time.time;
				counter++;
			}
			return counter % 2 == 1;
		}
		#endregion

		private void FlipMouseLookMode(ref DoubleSwitch b, string key) {
			if(Input.GetKeyDown(key)) {
				if(!toggleLazor)
					toggleLazor.set = true;
				b.set = !b;
			}
		}

        public override void OnUpdate() {
			if(FlightGlobals.ActiveVessel != vessel || !FlightGlobals.ActiveVessel.isEVA || myVessel == "-1")
				return;
			if(lazorInitiated)
				lazor.onUpdate();
			if(!MapView.MapIsEnabled && !FlightDriver.Pause && kerbalEVA.Fuel > fuelReqForLazorEffects) {
				FlipMouseLookMode(ref toggleMouseLook_Explode, "1");
				FlipMouseLookMode(ref toggleMouseLook_Decouple, "2");
				FlipMouseLookMode(ref toggleMouseLook_Weld, "3");
				FlipMouseLookMode(ref toggleMouseLook_Debug, "0");
			} else {
				toggleMouseLook_Explode.set = false;
				toggleMouseLook_Decouple.set = false;
				toggleMouseLook_Weld.set = false;
				toggleMouseLook_Debug.set = false;
				toggleLazor.set = false;
			}
			if(toggleMouseLook_Explode.In()) {
				toggleMouseLook_Decouple.set = false;
				toggleMouseLook_Weld.set = false;
				toggleMouseLook_WeldSuccess = false;
				toggleMouseLook_WeldFailure = false;
				toggleMouseLook_Debug.set = false;
			} else if(toggleMouseLook_Explode.Out()) { }
			if(toggleMouseLook_Decouple.In()) {
				toggleMouseLook_Explode.set = false;
				toggleMouseLook_Weld.set = false;
				toggleMouseLook_WeldSuccess = false;
				toggleMouseLook_WeldFailure = false;
				toggleMouseLook_Debug.set = false;
			} else if(toggleMouseLook_Decouple.Out()) { }
			if(toggleMouseLook_Weld.In()) {
				weld1 = weld2 = null;
				toggleMouseLook_WeldSuccess = false;
				toggleMouseLook_WeldFailure = false;
				toggleMouseLook_Explode.set = false;
				toggleMouseLook_Decouple.set = false;
				toggleMouseLook_Debug.set = false;
			} else if(toggleMouseLook_Weld.Out()) { }
			if(toggleMouseLook_Debug.In()) {
				toggleMouseLook_Explode.set = false;
				toggleMouseLook_Decouple.set = false;
				toggleMouseLook_Weld.set = false;
				toggleMouseLook_WeldSuccess = false;
				toggleMouseLook_WeldFailure = false;
			} else if(toggleMouseLook_Debug.Out()) { }
			if(toggleLazor.In()) {
				if(!lazorInitiated) { //make the lazor when it is first requested
					lazorInitiated = true;
					lazorAttach = new GameObject();
					lazorAttach.transform.parent = pivotDummy.transform;
					lazorAttach.transform.rotation = pivotDummy.transform.rotation;
					lazorAttach.transform.localRotation *= Quaternion.Euler(180, 0, 20);
					lazorAttach.transform.localPosition = new Vector3(-0.25199f, -0.068f, -0.09599f);
					lazor.Initialize(system, lazorAttach.transform, new Vector3(0, 0, 0), Color.red);
					lazor.SetRange(10000);
				}
				lazor.On();
			} else if(toggleLazor.Out()) {
				lazor.Off();
			}

			//Handle mouse input in free aim mode
			if(toggleMouseLook_Explode || toggleMouseLook_Decouple || toggleMouseLook_Weld || toggleMouseLook_Debug) {
				lazor.Aim(Util.MouseLook(Event.current.mousePosition, 557059) - lazor.GetOrigin());
				if(lazor.LazorIsHittingVessel && !lazor.VesselHit.isEVA)
					hitPart = lazor.GetHitPart();
				else
					hitPart = null;

				//weld stuff
				if(weld1 != null && weld2 == null){
					if(hitPart != null){
						//rough estimation of how close weld1 and weld2 colliders are.
						Vector3 p1 = hitPart.collider.ClosestPointOnBounds(weld1.transform.position);
						weldDistance = (weld1.collider.ClosestPointOnBounds(p1) - p1).magnitude;;
					}else
						weldDistance = double.PositiveInfinity;
				}
			}
			//EXPLODE
			if(toggleMouseLook_Explode && Input.GetMouseButtonDown(0)) {
				toggleMouseLook_Explode.set = false;
				if(lazor.LazorIsHittingVessel && !lazor.VesselHit.isEVA) {
					//kerbalEVA.Fuel -= 10;
					kerbalEVA.part.Resources.Get(kerbalEVA.propellantResourceId).amount -= fuelReqForLazorEffects;


					Part p = lazor.GetHitPart();
					MonoBehaviour.print("Exploding " + p.name);
					p.explode();
				}
			}
			//DECOUPLE
			if(toggleMouseLook_Decouple && Input.GetMouseButtonDown(0)) {
				toggleMouseLook_Decouple.set = false;
				if(lazor.LazorIsHittingVessel && !lazor.VesselHit.isEVA) {
					Part p = lazor.GetHitPart();
					MonoBehaviour.print("Decoupling " + p.name);
					if(LazorRoboticArm.DecoupleHeldPart(p)){
						//kerbalEVA.Fuel -= 10;
						kerbalEVA.part.Resources.Get(kerbalEVA.propellantResourceId).amount -= fuelReqForLazorEffects;

						ScreenMessages.PostScreenMessage("Successfully decoupled " + p.partInfo.title, 3f, ScreenMessageStyle.UPPER_CENTER);
					}else
						ScreenMessages.PostScreenMessage("Failed to decouple " + p.partInfo.title, 3f, ScreenMessageStyle.UPPER_CENTER);
				}
			}
			//WELD
			if(toggleMouseLook_Weld && Input.GetMouseButtonDown(0)) {
				if(lazor.LazorIsHittingVessel && !lazor.VesselHit.isEVA) {
					if(weld1 == null){
						weld1 = lazor.GetHitPart();
					}else{
						weld2 = lazor.GetHitPart();
					}
					if(weld1 != null && weld2 != null){ //w1 and w2 exists ... w1 is not w2
						if(weld1 != weld2){
							if(weldDistance < 0.01d){
								if(weld1.vessel == weld2.vessel){
									FixedJoint joint = weld1.gameObject.AddComponent<FixedJoint>();
									joint.breakForce = float.PositiveInfinity;
									joint.breakTorque = float.PositiveInfinity;
									joint.connectedBody = weld2.Rigidbody;
									PartRoboticArm arm1 = LazorRoboticArm.FindMainArmInHierarchy(weld1, true);
									PartRoboticArm arm2 = LazorRoboticArm.FindMainArmInHierarchy(weld2, true);
									if(arm1 != null && arm1.vessel == weld2.vessel)
										arm1.robotArm.MakeStatic();
									if(arm2 != null && arm2.vessel == weld1.vessel)
										arm2.robotArm.MakeStatic();
								}else{
									Part dom = LazorRoboticArm.GetDominantPart(weld1, weld2);
									Part sub = dom != weld1 ? weld1 : weld2;
									dom.vessel.SetRotation(dom.vessel.transform.rotation);
									sub.vessel.SetRotation(sub.vessel.transform.rotation);
									sub.vessel.SetPosition(sub.vessel.transform.position, true);
									dom.attachMethod = AttachNodeMethod.FIXED_JOINT;
									sub.Couple(dom);
									/*weld1.vessel.SetRotation(weld1.vessel.transform.rotation);
									weld2.vessel.SetRotation(weld2.vessel.transform.rotation);
									weld2.vessel.SetPosition(weld2.vessel.transform.position, true);
									weld1.attachMethod = AttachNodeMethod.FIXED_JOINT;
									weld2.Couple(weld1);*/
								}
								//kerbalEVA.Fuel -= 10;
								kerbalEVA.part.Resources.Get(kerbalEVA.propellantResourceId).amount -= fuelReqForLazorEffects;

								toggleMouseLook_Weld.set = false;
								toggleMouseLook_WeldSuccess = true;
								ScreenMessages.PostScreenMessage("Welding SUCCESS!!", 3f, ScreenMessageStyle.UPPER_CENTER);
							}else{
								//failure: parts too far apart
								toggleMouseLook_Weld.set = false;
								toggleMouseLook_WeldFailure = true;
								toggleMouseLook_WeldFailure_msg = "The parts were too far apart.";
								ScreenMessages.PostScreenMessage("Welding FAILURE: The parts were too far apart.", 3f, ScreenMessageStyle.UPPER_CENTER);
							}
						}else{
							//failure: welding same part
							toggleMouseLook_Weld.set = false;
							toggleMouseLook_WeldFailure = true;
							toggleMouseLook_WeldFailure_msg = "Can't weld a part to itself.";
							ScreenMessages.PostScreenMessage("Welding FAILURE: Can't weld a part to itself.", 3f, ScreenMessageStyle.UPPER_CENTER);
						}
					}
				}
			}
			//ENGINEER
			if(toggleMouseLook_Debug && Input.GetMouseButtonDown(0)) {
				toggleMouseLook_Debug.set = false;
				if(lazor.LazorIsHittingVessel && !lazor.VesselHit.isEVA) {
					ScreenMessages.PostScreenMessage("Debug output printed to debug window (ALT-F2)", 3f, ScreenMessageStyle.UPPER_CENTER);
					Part p = lazor.GetHitPart();
					MonoBehaviour.print("Engineering beam found::::::::::::::::::::::::::::::::::::::");
					MonoBehaviour.print("Part: " + p.partInfo.title);
					MonoBehaviour.print("Part InstanceID: " + p.GetInstanceID());
					MonoBehaviour.print("Part flightID: " + p.flightID);
					MonoBehaviour.print("Part uid: " + p.uid);
					MonoBehaviour.print("Vessel: " + p.vessel.vesselName);
					MonoBehaviour.print("Vessel id: " + p.vessel.id.ToString());
					MonoBehaviour.print("Phys.sig.: " + p.physicalSignificance.ToString());
					MonoBehaviour.print("Packed: " + p.packed);
					MonoBehaviour.print("Rigidbody: " + (p.rigidbody != null));
					MonoBehaviour.print("Layer: " + p.gameObject.layer);
					if(p.rigidbody != null)
						MonoBehaviour.print("Kinematic: " + p.rigidbody.isKinematic);
					MonoBehaviour.print("Attach nodes: " + p.attachNodes.Count);
					foreach(AttachNode an in p.attachNodes)
						MonoBehaviour.print("Attach node: " + an.id);
					MonoBehaviour.print("AttachJoint: " + (p.attachJoint == null ? "null" : p.attachJoint.name));
					Joint[] js = part.rigidbody.GetComponents<Joint>();
					MonoBehaviour.print("Joints: " + js.Length);
					foreach(Joint j in js)
						MonoBehaviour.print("joint: " + j.name + j.connectedBody.name);
					PartRoboticArm arm = LazorRoboticArm.FindMainArmInHierarchy(p);
					MonoBehaviour.print("Arm in hierarchy (down): " + (arm == null ? "none" : arm.name));
					arm = LazorRoboticArm.FindMainArmInHierarchy(p, true);
					MonoBehaviour.print("Arm in hierarchy (up): " + (arm == null ? "none" : arm.name));
				}
			}


            //Teleport In
            if (teleIn2 && (timerStart + 3) < Time.time) {
                TeleportInEnd();
                teleIn2 = false;
            }else if (teleIn) {
                timerTick = Time.time;
                timerStart = Time.time;
                teleIn = false;
                teleIn2 = true;
            }
            if ((teleIn || teleIn2) && (Time.time - timerTick) > 0.25f) {
                TeleportAnimationUpdate();
                timerTick = Time.time;
            }
            //Teleport Out Short
            if (teleOutShort2 && (timerStart + 2) < Time.time) {
                vessel.transform.position = vessel.transform.position - ((vessel.transform.position - vessel.mainBody.position).normalized * (vessel.GetHeightFromTerrain() - 5f));
                vessel.rigidbody.velocity = Vector3.zero;
                vessel.rigidbody.angularVelocity = Vector3.zero;
                teleOutShort2 = false;
            }else if (teleOutShort) {
                vessel.transform.position = surfacePos + ((surfacePos - vessel.mainBody.position).normalized * 500);
                vessel.rigidbody.velocity = Vector3.zero;
                vessel.rigidbody.angularVelocity = Vector3.zero;
                teleOutShort = false;
                teleOutShort2 = true;
                timerStart = Time.time;
                timerTick = Time.time;
            } else if (teleOutShortOngoing) {
                if (vessel.situation != Vessel.Situations.LANDED) {
                    if (vessel.srf_velocity.magnitude > 3)
                        vessel.SetWorldVelocity(Vector3.ClampMagnitude(vessel.srf_velocity, 3));
                    vessel.transform.up = ((surfacePos - vessel.mainBody.position).normalized);
                } else {
                    teleOutShortOngoing = false;
                    MonoBehaviour.print("[Lazor EVA: Teleport (short range) completed]");
                }
            }
            if ((teleOutShort || teleOutShort2) && (Time.time - timerTick) > 0.25f) {
                TeleportAnimationUpdate();
                timerTick = Time.time;
            }
            //Teleport Out
            if (surfacePos != Vector3.zero) {
                if (teleOut3 && (timerStart + 2) < Time.time) {
                    vessel.transform.position = vessel.transform.position - ((vessel.transform.position - vessel.mainBody.position).normalized * (vessel.GetHeightFromTerrain() - 5f));
                    vessel.rigidbody.velocity = Vector3.zero;
                    vessel.rigidbody.angularVelocity = Vector3.zero;
                    teleOut3 = false;
                }else if (teleOut2) {
                    double h = vessel.mainBody.pqsController.GetClampedWorldSurfaceHeight(surfacePos) - vessel.mainBody.Radius;
                    vessel.transform.position = surfacePos + ((surfacePos - vessel.mainBody.position).normalized * (h + 5000f));
                    vessel.rigidbody.velocity = Vector3.zero;
                    vessel.rigidbody.angularVelocity = Vector3.zero;
                    teleOut2 = false;
                    teleOut3 = true;
                }else if (teleOut) {
                    double h = vessel.mainBody.pqsController.GetClampedWorldSurfaceHeight(surfacePos) - vessel.mainBody.Radius;
                    vessel.transform.position = surfacePos + ((surfacePos - vessel.mainBody.position).normalized * (h + Mathf.Clamp((float)vessel.orbit.altitude, 7500f, (float)vessel.orbit.altitude)));
                    vessel.rigidbody.velocity = Vector3.zero;
                    vessel.rigidbody.angularVelocity = Vector3.zero;
                    teleOut = false;
                    teleOut2 = true;
                    timerStart = Time.time;
                    timerTick = Time.time;
                }else if (teleOutOngoing) {
                    if (vessel.situation != Vessel.Situations.LANDED) {
                        if (vessel.srf_velocity.magnitude > 3)
                            vessel.SetWorldVelocity(Vector3.ClampMagnitude(vessel.srf_velocity, 3));
                        vessel.transform.up = ((surfacePos - vessel.mainBody.position).normalized);
                    } else {
                        teleOutOngoing = false;
                        MonoBehaviour.print("[Lazor EVA: Teleport (long range) completed]");
                    }
                }
                if ((teleOut || teleOut2 || teleOut3 ) && (Time.time - timerTick) > 0.25f) {
                    TeleportAnimationUpdate();
                    timerTick = Time.time;
                }
            }
        }

        private void TeleportAnimationUpdate() {
            teleportGO.transform.position = vessel.transform.position;
            teleportEmitter.Emit(50);
        }
        
        private GameObject TeleportAnimation<T>(T le) where T : MonoBehaviour {
            teleportGO = Part.Instantiate(UnityEngine.Resources.Load("Effects/fx_gasJet_white")) as GameObject;
            teleportGO.transform.parent = le.transform;
            teleportGO.transform.localEulerAngles = Vector3.zero;
            teleportGO.transform.localPosition = new Vector3(0, 0, 0);
            teleportGO.transform.position = le.transform.position;
            teleportEmitter = teleportGO.particleEmitter;
            teleportEmitter.emit = false;
            teleportEmitter.minSize = 100;
            teleportEmitter.maxSize = 150;
            teleportEmitter.minEnergy = 2f;
            teleportEmitter.maxEnergy = 3f;
            teleportEmitter.rndVelocity = Vector3.one * 0.7f;
            teleportEmitter.localVelocity = Vector3.zero;
            teleportEmitter.useWorldSpace = false;
            ParticleAnimator emitterAnimator = teleportEmitter.GetComponent<ParticleAnimator>();
            emitterAnimator.sizeGrow = -0.8F;
            Color[] emitterColors = emitterAnimator.colorAnimation;
            emitterColors[0] = new Color(1f, 1f, 1f, 0.5f);
            emitterColors[1] = new Color(0f, 0f, 1f, 0.8f);
            emitterColors[2] = new Color(0f, 0f, 0.7f, 0.7f);
            emitterColors[3] = new Color(0f, 0.5f, 0.7f, 0.6f);
            emitterColors[4] = new Color(0f, 0f, 1f, 0.6f);
            emitterAnimator.colorAnimation = emitterColors;
            teleportEmitter.Emit(50);
            return teleportGO;
        }

        private void TestButtons_PrintCamshelper(Camera c) {
            MonoBehaviour.print(c.name + " + " + c.depth + " + " + c.clearFlags + " + " + c.cullingMask + " + " + c.layerCullSpherical + " + " + c.orthographic + " + " + c.hdr + " + " + c.farClipPlane + " + " + c.fov);
            MonoBehaviour.print(c.transform.position + " + " + c.backgroundColor);
        }

        private LazorSystem FindSystem(string mySystem) {
            foreach (Vessel v in FlightGlobals.Vessels) {
                Guid guid = new Guid(myVessel);
                if (guid == v.id) {
                    foreach (Part p in v.parts) {
                        if (p is LazorSystem)
                            return (LazorSystem)p;
                    }
                }
            }
            return null;
        }

        private Camera FindCamera(string name) {
            foreach (Camera c in Camera.allCameras)
                if (c.name == name)
                    return c;
            return null;
        }

        //EVENTS

        public void OnVesselChange() {
            DisposeCams();
        }

        public void OnDestroy(){ //called whenever the kerbal is destroyed
			if(vessel != null) { ///XXX WARNING KEEP AN EYE ON THIS
				print("[LazorEVA: Disposing LazorEVA system on " + vessel.vesselName + "]");
				DisposeCams();
				UnityEngine.Object.Destroy(teleportGO);
				RenderingManager.RemoveFromPostDrawQueue(0, new Callback(drawGUI));
				Romfarer.Events.onVesselChange -= OnVesselChange;
			}
        }

        private void DisposeCams(bool force = false) {
            if (toggleFPS || force) {
                if(FlightGlobals.ActiveVessel != null)
                    flightCamera.setTarget(FlightGlobals.ActiveVessel.transform);
                flightCamera.minPitch = cam_minPitch;
                flightCamera.maxPitch = cam_maxPitch;
                flightCamera.pivotTranslateSharpness = cam_pivotTranslateSharpness;
                flightCamera.minHeight = cam_minHeight;
                flightCamera.minHeightAtMaxDist = cam_minHeightAtMaxDist;
                flightCamera.minHeightAtMinDist = cam_minHeightAtMinDist;
                flightCamera.minDistance = cam_minDistance;
                flightCamera.maxDistance = cam_maxDistance;
                flightCamera.startDistance = cam_startDistance;
                flightCamera.SetDistance(flightCamera.startDistance);
                if(vessel.transform != null)
                    foreach (Component c in vessel.transform.GetComponentsInChildren(typeof(Transform)))
                        if (c.name.Contains("headMesh") || c.name.Contains("eyeball") || c.name.Contains("upTeeth") || c.name.Contains("pupil"))
                            c.gameObject.renderer.enabled = true;
                toggleFPS.set = false;
            }
        }

		public override void OnLoad(ConfigNode node){
			MonoBehaviour.print("LazorEVA: LOADING EVA");
		}
		
		public override void OnSave(ConfigNode node){
			MonoBehaviour.print("LazorEVA: SAVING EVA");
			//PartmodulePersistenceFix.SaveEVA();


			//node.AddNode("vesselRef");
			//node.AddValue ("myVessel", myVessel);
		}
    }
}

