/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;


namespace Romfarer {
    static class LazorGlobals {

		//global config PERSISTENT
		public static bool GLOBALCONFIG_SyncTargets = true; //Sync targets with in-game targeting
		public static bool GLOBALCONFIG_ElectricGenerator = true; //if false: makes the system not generate power
		public static LoadDistance GLOBALCONFIG_CurrentLoadDistance = LoadDistance.Medium;
		public static PIDclamp GLOBALCONFIG_pid_plane_roll_manual = new PIDclamp("Roll manual", 0.25f, 0.001f, 0.085f, 90f);
		public static PIDclamp GLOBALCONFIG_pid_plane_roll_auto = new PIDclamp("Roll auto", 0.25f, 0.001f, 0.095f, 45f);
		public static PIDclamp GLOBALCONFIG_pid_plane_yaw_auto = new PIDclamp("Yaw auto", 0.102f, 0.055f, 0.06f, 1f);
		public static PIDclamp GLOBALCONFIG_pid_plane_pitch = new PIDclamp("Pitch auto/manual", 0.3f, 0.4f, 0.021f, 1f);
		public static PIDclamp GLOBALCONFIG_pid_rocket_yaw = new PIDclamp("Torque yaw", 8.3f, 0.3f, 0.5f, 1f);
		public static PIDclamp GLOBALCONFIG_pid_rocket_pit = new PIDclamp("Torque pitch", 8.3f, 0.3f, 0.5f, 1f);
		public static PIDclamp GLOBALCONFIG_pid_rocket_rol = new PIDclamp("Torque roll", 6.0f, 0.25f, 0.025f, 1f);
		public static PIDclamp GLOBALCONFIG_pid_rocket_lr = new PIDclamp("Translate left/right", 4.3f, 0.01f, 0.02f, 1f);
		public static PIDclamp GLOBALCONFIG_pid_rocket_ud = new PIDclamp("Translate up/down", 4.3f, 0.01f, 0.02f, 1f);
		public static PIDclamp GLOBALCONFIG_pid_rocket_fb = new PIDclamp("Translate fwd/back", 4.3f, 0.01f, 0.02f, 1f);
		//public static PIDclamp GLOBALCONFIG_pid_tractor_hover = new PIDclamp("Tractor hover", 10f, 0.06f, 20f, 30f);

		#region persistence
		
		public static void LoadConfig() {
			Dictionary<string, KSPParseable> data = WrapData();
			Util.Load("config.cfg", ref data);
			if(data.ContainsKey("SyncTargets"))
				LazorGlobals.GLOBALCONFIG_SyncTargets = data["SyncTargets"].value_bool;
			if(data.ContainsKey("ElectricGenerator"))
				LazorGlobals.GLOBALCONFIG_ElectricGenerator = data["ElectricGenerator"].value_bool;
			if(data.ContainsKey("CurrentLoadDistance"))
				LazorGlobals.GLOBALCONFIG_CurrentLoadDistance = (LazorGlobals.LoadDistance)data["CurrentLoadDistance"].value_int;
			
			//pid plane
			if(data.ContainsKey("pid_plane_roll_manual")) LazorGlobals.GLOBALCONFIG_pid_plane_roll_manual = PIDclamp.Parse(data["pid_plane_roll_manual"].value);
			if(data.ContainsKey("pid_plane_roll_auto")) LazorGlobals.GLOBALCONFIG_pid_plane_roll_auto = PIDclamp.Parse(data["pid_plane_roll_auto"].value);
			if(data.ContainsKey("pid_plane_yaw_auto")) LazorGlobals.GLOBALCONFIG_pid_plane_yaw_auto = PIDclamp.Parse(data["pid_plane_yaw_auto"].value);
			if(data.ContainsKey("pid_plane_pitch")) LazorGlobals.GLOBALCONFIG_pid_plane_pitch = PIDclamp.Parse(data["pid_plane_pitch"].value);
			//pid rocket
			if(data.ContainsKey("pid_rocket_yaw")) LazorGlobals.GLOBALCONFIG_pid_rocket_yaw = PIDclamp.Parse(data["pid_rocket_yaw"].value);
			if(data.ContainsKey("pid_rocket_pit")) LazorGlobals.GLOBALCONFIG_pid_rocket_pit = PIDclamp.Parse(data["pid_rocket_pit"].value);
			if(data.ContainsKey("pid_rocket_rol")) LazorGlobals.GLOBALCONFIG_pid_rocket_rol = PIDclamp.Parse(data["pid_rocket_rol"].value);
			if(data.ContainsKey("pid_rocket_lr")) LazorGlobals.GLOBALCONFIG_pid_rocket_lr = PIDclamp.Parse(data["pid_rocket_lr"].value);
			if(data.ContainsKey("pid_rocket_ud")) LazorGlobals.GLOBALCONFIG_pid_rocket_ud = PIDclamp.Parse(data["pid_rocket_ud"].value);
			if(data.ContainsKey("pid_rocket_fb")) LazorGlobals.GLOBALCONFIG_pid_rocket_fb = PIDclamp.Parse(data["pid_rocket_fb"].value);
			//tractor beam
			//if(data.ContainsKey("pid_tractor_hover")) LazorGlobals.GLOBALCONFIG_pid_tractor_hover = PIDclamp.Parse(data["pid_tractor_hover"].value);
		}

		public static void SaveConfig() {
			Util.Save("config.cfg", WrapData());
		}

		private static Dictionary<string, KSPParseable> WrapData() {
			Dictionary<string, KSPParseable> data = new Dictionary<string, KSPParseable>();
			data.Add("SyncTargets", new KSPParseable((bool)LazorGlobals.GLOBALCONFIG_SyncTargets, KSPParseable.Type.BOOL));
			data.Add("ElectricGenerator", new KSPParseable((bool)LazorGlobals.GLOBALCONFIG_ElectricGenerator, KSPParseable.Type.BOOL));
			data.Add("CurrentLoadDistance", new KSPParseable((int)LazorGlobals.GLOBALCONFIG_CurrentLoadDistance, KSPParseable.Type.INT));
			data.Add("pid_plane_roll_manual", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_plane_roll_manual.ToString(), KSPParseable.Type.STRING));
			data.Add("pid_plane_roll_auto", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_plane_roll_auto.ToString(), KSPParseable.Type.STRING));
			data.Add("pid_plane_yaw_auto", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_plane_yaw_auto.ToString(), KSPParseable.Type.STRING));
			data.Add("pid_plane_pitch", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_plane_pitch.ToString(), KSPParseable.Type.STRING));
			data.Add("pid_rocket_yaw", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_rocket_yaw.ToString(), KSPParseable.Type.STRING));
			data.Add("pid_rocket_pit", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_rocket_pit.ToString(), KSPParseable.Type.STRING));
			data.Add("pid_rocket_rol", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_rocket_rol.ToString(), KSPParseable.Type.STRING));
			data.Add("pid_rocket_lr", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_rocket_lr.ToString(), KSPParseable.Type.STRING));
			data.Add("pid_rocket_ud", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_rocket_ud.ToString(), KSPParseable.Type.STRING));
			data.Add("pid_rocket_fb", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_rocket_fb.ToString(), KSPParseable.Type.STRING));
			//data.Add("pid_tractor_hover", new KSPParseable(LazorGlobals.GLOBALCONFIG_pid_tractor_hover.ToString(), KSPParseable.Type.STRING));
			return data;
		}
		
		#endregion


		public static Vector4[] loadDistances = new Vector4[]{ new Vector4(1000, 900, 200, 1000), 
			new Vector4(2500, 2250, 350, 2250), 
			new Vector4(22500, 21500, 20000, 20000),
			new Vector4(99900, 97500, 95000, 95000) };
		public enum LoadDistance { Short = 0, Default = 1, Medium = 2, Long = 3};

		//global config 
		public const int TELEPORT_DISTANCE = 90000;
		public static int MAX_DISTANCE { get { return (int)LazorGlobals.loadDistances [(int)LazorGlobals.GLOBALCONFIG_CurrentLoadDistance].x; } }
		public static int VIEW_DISTANCE  { get { return (int)LazorGlobals.loadDistances[(int)LazorGlobals.GLOBALCONFIG_CurrentLoadDistance].y; } }
		//public static int LANDEDPARTPACK_DISTANCE { get { return (int)LazorGlobals.loadDistances[(int)LazorGlobals.GLOBALCONFIG_CurrentLoadDistance].z; } }
		//public static int PARTPACK_DISTANCE { get { return (int)LazorGlobals.loadDistances[(int)LazorGlobals.GLOBALCONFIG_CurrentLoadDistance].w; } }
		//public static OrbitPhysicsManager manager;

		public static int ElectricChargeID;
		public static int LiquidFuelID;

        //Docking camera
        public static ITargetable LastKSPtarget;
        private static LazorSystemDockingCamera activeDockingCamera;

        //Vessel change event tracking
		private static Vessel lastVessel;
		public static Vessel LastVessel {
			get { return lastVessel; }
			set {
				if (value != lastVessel) {
					Events.RegisterVesselChange(lastVessel == null ? "nothing" : lastVessel.vesselName, value == null ? "nothing" : value.vesselName);
					LazorSystem.TurnOffRemoteControlForVessel(lastVessel);
					LazorSystem.TurnOffRemoteControlForVessel(value);
					LazorSystemControllerGlobal.MergeLazorControllers(null, value);
					LazorGlobals.AddDestructionTrackerToAll();
				}
				lastVessel = value;
			}
		}

        //Celestial Bodies
        private static bool initiated;
		public static bool planetOverlaysSetup; //Are the overlays made?
        private static Dictionary<string, CelestialBodyType> bodyType = new Dictionary<string, CelestialBodyType>();
        private static Dictionary<string, CelestialBody> bodyByName = new Dictionary<string, CelestialBody>();
        private static Dictionary<string, Vector2> geosyncAltVel = new Dictionary<string, Vector2>();
        private static Dictionary<string, MeshRenderer> planetMapGrids = new Dictionary<string, MeshRenderer>();

        //objects
        public static Collider trackingStation;
		private static Dictionary<string,Transform> mapBodies = new Dictionary<string,Transform>();

		public static Color COLOR_GREENISH = new Color(0.282f, 1f, 0.803f, 1);


        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Docking Camera
        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


        public static LazorSystemDockingCamera ActiveDockingCamera {
            get { return activeDockingCamera; }
            set {
                if(activeDockingCamera != null)
                    activeDockingCamera.DeactivateLazorCamera();
                activeDockingCamera = value;
                activeDockingCamera.NinjaActivateLazorCamera();
            }
        }
	

        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Celestial Bodies 
        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


        public static void Initialize() {
			ElectricChargeID = "ElectricCharge".GetHashCode();
			LiquidFuelID = "LiquidFuel".GetHashCode();
			ScaledSpace space = GameObject.FindObjectsOfType(typeof(ScaledSpace)).OfType<ScaledSpace>().Single();
			mapBodies.Clear();
            bodyType.Clear();
            geosyncAltVel.Clear();
            bodyByName.Clear();
            foreach (CelestialBody cb in FlightGlobals.Bodies) {
                if (cb.name == "Sun")
                    bodyType.Add(cb.name, CelestialBodyType.STAR);
                else if (cb.referenceBody.name == "Sun")
                    bodyType.Add(cb.name, CelestialBodyType.PLANET);
                else if (cb.referenceBody.referenceBody.name == "Sun")
                    bodyType.Add(cb.name, CelestialBodyType.MOON);
                bodyByName.Add(cb.name, cb);
                //calculate geosync altitude and velocity
                double w = (2 * Math.PI) / cb.rotationPeriod;
                double r = Math.Pow(cb.gravParameter / (w * w), 1d/3d);
                double v = w * r;
                double a = r - cb.Radius;
                geosyncAltVel.Add(cb.name, new Vector2((float)a, (float)v));
				mapBodies.Add(cb.name, space.scaledSpaceTransforms.Find(x => x.name == cb.name).transform);
            }
            initiated = true;
        }

        public static Vector2 GeosyncAltVel(string body){
            if (!initiated)
                Initialize();
            return geosyncAltVel[body];
        }

		public static Transform GetMapBody(string name){
			return mapBodies[name];
		}

        public static CelestialBodyType BodyType(CelestialBody cb) {
            if (!bodyType.ContainsKey(cb.name))
                return CelestialBodyType.NULL;
            return bodyType[cb.name];
        }

        public static bool IsStar(CelestialBody cb) {
            return BodyType(cb) == CelestialBodyType.STAR;
        }

        public static bool IsPlanet(CelestialBody cb) {
            return BodyType(cb) == CelestialBodyType.PLANET;
        }

        public static bool IsMoon(CelestialBody cb) {
            return BodyType(cb) == CelestialBodyType.MOON;
        }

        public static CelestialBody GetBody(string name) {
            if(bodyByName.ContainsKey(name))
                return bodyByName[name];
            return null;
        }

        public enum CelestialBodyType {
            STAR = 0,
            PLANET = 1,
            MOON = 2,
            NULL = 3,
        }

		public static void ClearPlanetOverlay(){
			planetOverlaysSetup = false;
			foreach (MeshRenderer m in planetMapGrids.Values)
				GameObject.Destroy (m.gameObject);
			planetMapGrids.Clear();
		}

        public static void TogglePlanetOverlay(bool on) {
            if (!planetOverlaysSetup) {
				ClearPlanetOverlay ();
                foreach (CelestialBody cb in FlightGlobals.Bodies)
                    AddPlanetOverlay(cb.name);
                planetOverlaysSetup = true;
            }
            foreach (string s in planetMapGrids.Keys)
                planetMapGrids[s].enabled = on;
        }

        private static bool AddPlanetOverlay(string name) {
            ScaledSpace space = GameObject.FindObjectsOfType(typeof(ScaledSpace)).OfType<ScaledSpace>().Single();
            Transform planet = space.scaledSpaceTransforms.Find(x => x.name == name);
            if (planet != null) {
                GameObject planetGO = new GameObject("LazorPlanetOverlay");
                MeshFilter meshFilter = planetGO.AddComponent<MeshFilter>();
                meshFilter.mesh = (Mesh)UnityEngine.Object.Instantiate(planet.GetComponent<MeshFilter>().mesh);

                MeshRenderer meshRenderer = planetGO.AddComponent<MeshRenderer>();
                Material planetMat = new Material(Shader.Find("Transparent/Diffuse"));
                planetMat.mainTexture = Util.PlanetTexture;
                meshRenderer.material = planetMat;

				planetGO.layer = planet.gameObject.layer;
                planetGO.transform.parent = planet;

                planetGO.transform.localScale = new Vector3(1.005f, 1.005f, 1.005f);
                planetGO.transform.position = planet.position;

                planetGO.transform.rotation = planet.rotation;

                meshFilter.transform.rotation = planet.GetComponent<MeshFilter>().transform.rotation;
                meshRenderer.transform.rotation = planet.GetComponent<MeshRenderer>().transform.rotation;

                //Rotate the texture to align with global coordinates
                Vector3 v1 = ScaledSpace.LocalToScaledSpace(LazorGlobals.GetBody("Kerbin").GetWorldSurfacePosition(0, 0, 0)) - planet.position;
                Vector3 v2 = planet.forward;
                float angle = Util.Angle(v1, v2, planet.up);
                float offset = angle / 360;
                planetMat.mainTextureOffset = new Vector2(planetMat.mainTextureOffset.x + offset, planetMat.mainTextureOffset.y);
                planetMapGrids.Add(name, meshRenderer);
                return true;
            }
            return false;
        }

        public static CelestialBody GetMapFocusBody() {
			ScaledSpaceFader ssf = LazorSystemControllerGlobal.ActiveLazorSystem.planetariumCamera.target.GetComponent<ScaledSpaceFader>();
            if (ssf != null)
                return ssf.celestialBody;
            else
				return LazorSystemControllerGlobal.ActiveLazorSystem.vessel.mainBody;
        }


        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Vessels
        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


        //Add partmodule DestructionTracker to every command pod in the game.
		public static void AddDestructionTracker(Vessel v) {
			//MonoBehaviour.print("[LazorGlobals: Checking vessel " + v.vesselName + " parts " + v.parts.Count + "]");
			if (!v.isEVA) {
				foreach (Part p in v.parts)
					if (p is CommandPod && !p.Modules.Contains("DestructionTracker")) {
						p.AddModule("DestructionTracker");
						MonoBehaviour.print("[LazorGlobals: Added DestructionTracker to " + v.vesselName + "]");
					}
			} else if (v.rootPart != null && !v.rootPart.Modules.Contains("DestructionTrackerEVA")) {
				v.rootPart.AddModule("DestructionTrackerEVA");
				MonoBehaviour.print("[LazorGlobals: Added DestructionTrackerEVA to " + v.name + "]");
			}
		}

        public static void AddDestructionTrackerToAll() {
            foreach (Vessel v in FlightGlobals.Vessels) {
				AddDestructionTracker(v);
            }
        }

		static public void SetPackingDistance(Vessel vessel, 
		                             float distancePackThreshold = 5000, 
		                             float distanceUnpackThreshold = 200,
		                             float distanceLandedPackThreshold = 350, 
		                             float distanceLandedUnpackThreshold = 200){
			vessel.distancePackThreshold = distancePackThreshold;
			vessel.distanceUnpackThreshold = distanceUnpackThreshold;
			vessel.distanceLandedPackThreshold = distanceLandedPackThreshold;
			vessel.distanceLandedUnpackThreshold = distanceLandedUnpackThreshold;
		}

		static public void SetLoadDistance(float loadDistance = 2500, float unloadDistance = 2250){
			if(LazorGlobals.GLOBALCONFIG_CurrentLoadDistance != LoadDistance.Default) {
				Vessel.loadDistance = loadDistance;
				Vessel.unloadDistance = unloadDistance;
			}
		}

    }













}

