﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using UnityEngine;

namespace Romfarer {
    class LazorParameters {
        public delegate bool Parameter(bool highligh);
        public delegate bool Resource(bool highligh, string name);
        private LazorSystem system;
        private Vessel vessel;
        private VesselIterator vi;
        public VesselIterator VI {
            get { return vi; }
            set {
                vi = value;
                vessel = value.vessel;
            }
        }
        public CelestialBody cb;
		private Color ColorName = LazorGlobals.COLOR_GREENISH;//new Color(0.282f, 1f, 0.803f, 1);


        public class ParameterWrapper {
            public Parameter param;
            public bool highlight;
            public ParameterWrapper(Parameter param, bool highlight) {
                this.param = param;
                this.highlight = highlight;
            }
            public void X() {
                highlight = param(highlight);
            }
        }

        public class ResourceWrapper {
            public Resource resource;
            public bool highlight;
            public string name;
            public ResourceWrapper(Resource resource, bool highlight, string name) {
                this.resource = resource;
                this.highlight = highlight;
                this.name = name;
            }
            public void X() {
                highlight = resource(highlight, name);
            }
        }

        public LazorParameters(LazorSystem system, VesselIterator vi, CelestialBody cb) {
            this.system = system;
            if (vi != null) {
                VI = vi;
                /*foreach (Part p in vessel.parts)
                    if (p is AtmosphericEngine && ((AtmosphericEngine)p).engineEnabled)
                        atmosEngine = (AtmosphericEngine)p;*/
            }
            //Events.onStageChange += new Events.OnStageChange(OnStageChange);
            this.cb = cb;
        }

        // RESOURCES ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

        public bool ResourceItem(bool highlight, string name) {
            VesselIteratorOwn.Resource r = VI.resourceParts[name];
            GUILayout.BeginVertical(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.BeginHorizontal(Util.Space_Compact);
            GUILayout.Label(Util.CutString(name, 10, ".."), Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label((r.flow != 0 ? r.amount.ToString("f3") : r.amount.ToString("0.###")) + " / " + r.maxAmount.ToString("0.###"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("U", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.Label(highlight ? "▲" : "▼", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.EndHorizontal();
            if (highlight) {
                GUILayout.BeginHorizontal(Util.Space_Compact);
                GUILayout.Label("Flow", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
                GUILayout.Label((r.flow != 0 ? -r.flow : 0).ToString("0.###"), Util.StdLabelSmallCustom(Color.white));
                GUILayout.Label("U/s", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
                GUILayout.FlexibleSpace();
                GUILayout.EndHorizontal();

                GUILayout.BeginHorizontal(Util.Space_Compact);
                if (r.flow.value > 0) { //resource is depleting
                    GUILayout.Label("Depletion time: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
                    GUILayout.Label(r.amount > 0 ?  Util.SecondsAsTimeString(r.amount / r.flow.value) : Util.SecondsAsTimeString(0), Util.StdLabelSmallCustom(Color.white));
                } else if (r.flow.value < 0) {
                    GUILayout.Label("Restoration time: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
                    GUILayout.Label(r.amount < r.maxAmount ? Util.SecondsAsTimeString(r.capacity / -r.flow.value) : Util.SecondsAsTimeString(0), Util.StdLabelSmallCustom(Color.white)); //: "∞"
                } else {
                    GUILayout.Label("Resource is not flowing: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
                }
                GUILayout.FlexibleSpace();
                GUILayout.EndHorizontal();
            }
            GUILayout.EndVertical();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }


        // SPECIAL ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        public bool Distance(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            Util.DistanceFormatted((vessel.transform.position - system.vessel.transform.position).magnitude, highlight ? Color.yellow : ColorName, "Distance: ");
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        // GENERAL ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        public bool MissionTime(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Mission Time: T+", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(Util.SecondsAsTimeString(vessel.missionTime), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Crew(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Crew: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.GetCrewCount() + " / " + vessel.GetCrewCapacity(), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Situation(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Situation: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.situation.ToString(), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool SoI(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("SOI: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.mainBody.name, Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        // SURFACE ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        public bool AltitudeFT(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            Util.DistanceFormatted(system.HeightFromTerrain(), highlight ? Color.yellow : ColorName, "Altitude terrain: ");
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }























        public bool SurfaceSpeed(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Surface speed: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(system.rotatingFrameVelocity.magnitude.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("m/s", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            if (vessel == FlightGlobals.ActiveVessel) {
                if (!highlight)
                    GUILayout.Toggle(false, "");
                system.guiOverlays.toggleSpeedSrf = highlight && GUILayout.Toggle(system.guiOverlays.toggleSpeedSrf, "");
            }
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool VerticalSpeed(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Vertical speed: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.verticalSpeed.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("m/s", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Latitude(bool highlight) {
            GUILayout.BeginVertical(Util.Space_Compact);
            GUILayout.BeginVertical(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.BeginHorizontal();
            GUIContent ctnt = new GUIContent("Latitude: ");
            float min, max;
            Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName).CalcMinMaxWidth(ctnt, out min, out max);
            GUILayout.Label(ctnt, Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(system.LATITUDE.ToString("f5"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            if (vessel == FlightGlobals.ActiveVessel) {
                if (!highlight)
                    GUILayout.Toggle(false, "");
                system.guiOverlays.toggleMapGrid.set = highlight && GUILayout.Toggle(system.guiOverlays.toggleMapGrid, "");
            }
            GUILayout.EndHorizontal();
            GUILayout.BeginHorizontal();
            GUILayout.Space(max);
            Util.DegToSexFormatted(system.LATITUDE, highlight ? Color.yellow : ColorName);
            GUILayout.Label((system.LATITUDE > 0 ? " N" : " S"), Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.EndHorizontal();
            GUILayout.EndVertical();
            //return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        //}

        //public bool Longitude(bool highlight) {
            GUILayout.BeginVertical(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.BeginHorizontal();
            /*GUIContent*/ ctnt = new GUIContent("Longitude: ");
            /*float min, max;*/
            Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName).CalcMinMaxWidth(ctnt, out min, out max);
            GUILayout.Label(ctnt, Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(system.LONGITUDE.ToString("f5"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            if (vessel == FlightGlobals.ActiveVessel) {
                if (!highlight)
                    GUILayout.Toggle(false, "");
                system.guiOverlays.toggleMapGrid.set = highlight && GUILayout.Toggle(system.guiOverlays.toggleMapGrid, "");
            }
            GUILayout.EndHorizontal();
            GUILayout.BeginHorizontal();
            GUILayout.Space(max);
            Util.DegToSexFormatted(system.LONGITUDE, highlight ? Color.yellow : ColorName);
            GUILayout.Label((system.LONGITUDE > 0 ? " E" : " W"), Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.EndHorizontal();
            GUILayout.EndVertical();
            GUILayout.EndVertical();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        // ORBIT ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

        public bool Acceleration(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Acceleration: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.acceleration.magnitude.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("m/s²", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        //system vessel only
        public bool OrbitSpeed(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Orbit speed: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.orbit.GetVel().magnitude.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("m/s", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            if (vessel == FlightGlobals.ActiveVessel) {
                if (!highlight)
                    GUILayout.Toggle(false, "");
                system.guiOverlays.toggleSpeedObt = highlight && GUILayout.Toggle(system.guiOverlays.toggleSpeedObt, "");
            }
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Altitude(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            Util.DistanceFormatted(vessel.orbit.altitude, highlight ? Color.yellow : ColorName, "Altitude: ");
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Apoapsis(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            Util.DistanceFormatted(vessel.orbit.ApA, highlight ? Color.yellow : ColorName, "Apoapsis: ");
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool TimeToApoapsis(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Time to Ap: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(Util.SecondsAsTimeString(vessel.orbit.timeToAp), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Periapsis(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            Util.DistanceFormatted(vessel.orbit.PeA, highlight ? Color.yellow : ColorName, "Periapsis: ");
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool TimeToPeriapsis(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Time to Pe: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(Util.SecondsAsTimeString(vessel.orbit.timeToPe), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Period(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Period: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(Util.SecondsAsTimeString(vessel.orbit.period), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Eccentricity(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Eccentricity: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.orbit.eccentricity.ToString("f4"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Inclination(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Inclination: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.orbit.inclination.ToString("f4"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool MeanAnomaly(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Mean Anomaly: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.orbit.meanAnomaly.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool EccentricAnomaly(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Eccentric Anomaly: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.orbit.eccentricAnomaly.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool TrueAnomaly(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("True Anomaly: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.orbit.trueAnomaly.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }




        // PHYSICAL ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::




        public bool Mass(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Mass: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label((VI.mass).ToString("f1"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("t", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CurrentStage(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Current Stage: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.currentStage.ToString("f0"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        /*public bool Fuel(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Fuel: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(VI.fuel.ToString("f1") + " / " + VI.fuelCapacity.ToString("f1"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("U", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool FuelBurnrate(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Fuel burnrate: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(((double)VI.fuelBurnRate).ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("U/s", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        //system.vessel only
        public bool FuelDepletionTime(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Depletion: T-", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(VI.fuel > 0 ? VI.fuelBurnRate > 0 ? Util.SecondsAsTimeString(VI.fuel / VI.fuelBurnRate) : "∞" : Util.SecondsAsTimeString(0), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }
        //system.vessel only
        public bool RCSFuel(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("RCS Fuel: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(VI.rcs.ToString("f1") + " / " + VI.rcsCapacity.ToString("f1"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("U", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }
         */

        /*public bool AtmoshpericEngineEff(bool highlight) {
            if (VI.hasAtmosEngine && atmosEngine != null) {
                GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
                GUILayout.Label("Engine Efficiency: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
                GUILayout.Label((atmosEngine.totalEfficiency * 100).ToString("f2"), Util.StdLabelSmallCustom(Color.white));
                GUILayout.Label("%", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
                GUILayout.FlexibleSpace();
                GUILayout.EndHorizontal();
                return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
            }
            return false;
        }

        public bool AtmoshpericEngineAirflow(bool highlight) {
            if (VI.hasAtmosEngine && atmosEngine != null) {
                GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
                GUILayout.Label("Airflow Efficiency: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
                GUILayout.Label((atmosEngine.airflowEfficiency * 100).ToString("f2"), Util.StdLabelSmallCustom(Color.white));
                GUILayout.Label("%", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
                GUILayout.FlexibleSpace();
                GUILayout.EndHorizontal();
                return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
            }
            return false;
        }*/



        public bool AtmosphericDensity(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("atmDensity: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.atmDensity.ToString("f3"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("kg/m³", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Pressure(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Pressure: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.staticPressure.ToString("f3"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("Pa", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool DynamicPressure(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Dyn.Pressure: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label((0.5f * vessel.staticPressure * system.rotatingFrameVelocity.magnitude * system.rotatingFrameVelocity.magnitude).ToString("f1"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("Pa", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool GeeForce(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("GeeForce: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(vessel.geeForce.ToString("f3"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("g", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Gravity(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Gravity: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(FlightGlobals.getGeeForceAtPosition(vessel.transform.position).magnitude.ToString("f4"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("m/s²", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool Temperature(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Temperature: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(FlightGlobals.getExternalTemperature((float)vessel.orbit.altitude, vessel.mainBody).ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°C", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }






        // CELESTIAL BODY ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::









        public bool CB_Distance(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            Util.DistanceFormatted((cb.transform.position - system.vessel.transform.position).magnitude - cb.Radius, highlight ? Color.yellow : ColorName, "Distance: ");
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_Radius(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            Util.DistanceFormatted(cb.Radius, highlight ? Color.yellow : ColorName, "Radius: ");
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_Mass(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Mass: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            if (cb.theName == "the Sun") {
                GUILayout.Label((cb.Mass / 100000000000000000000f).ToString("f0"), Util.StdLabelSmallCustom(Color.white));
                GUILayout.Label("x10^20 kg ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            } else {
                GUILayout.Label((cb.Mass / 1000000000000000).ToString("f0"), Util.StdLabelSmallCustom(Color.white));
                GUILayout.Label("x10^15 kg ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            }
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_GravParam(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("μ: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.gravParameter.ToString("f0"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_ASLGee(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("ASL Gee: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.GeeASL.ToString("f3"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("m/s²", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_HasAtmosphere(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Atmosphere: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.atmosphere ? "true" : "false", Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_AtmosAlt(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            Util.DistanceFormatted(cb.maxAtmosphereAltitude, highlight ? Color.yellow : ColorName, "Atmos.Alt.: ");
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_ASLPressure(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("ASL Pressure: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            if (cb.theName == "the Sun") 
                GUILayout.Label(cb.staticPressureASL.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            else
                GUILayout.Label(FlightGlobals.getStaticPressure(0, cb).ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("Pa", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_ASLDensity(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("ASL Density: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(FlightGlobals.getAtmDensity(FlightGlobals.getStaticPressure(0, cb)).ToString("f4"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("kg/m³", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_ASLTemperature(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("ASL Temp: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(FlightGlobals.getExternalTemperature(0, cb).ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°C", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_HasOxygen(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Oxygen: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.atmosphereContainsOxygen ? "true" : "false", Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_HasOcean(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Ocean", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.ocean ? "true" : "false", Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_IsTidallyLocked(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("TidallyLocked: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.tidallyLocked ? "true" : "false", Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_IsRotating(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Rotates: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.rotates ? "true" : "false", Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_RotationPeriod(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("RotationPeriod: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(Util.SecondsAsTimeString(cb.rotationPeriod), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_IsRotatingInverse(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Rotates Inverse: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.inverseRotation ? "true" : "false", Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_OrbitPeriod(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("OrbitPeriod: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(Util.SecondsAsTimeString(cb.orbit.period), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_OrbitVelocity(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("OrbitVelocity: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label((cb.orbit.GetRelativeVel().magnitude).ToString("f1"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("m/s", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_OrbitEccentricity(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("OrbitEcc: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label((cb.orbit.eccentricity).ToString("f4"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_OrbitInclination(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("OrbitIncl: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label((cb.orbit.inclination).ToString("f4"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_MeanAnomaly(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Mean Anomaly: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.orbit.meanAnomaly.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_EccentricAnomaly(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("Eccentric Anomaly: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.orbit.eccentricAnomaly.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }

        public bool CB_TrueAnomaly(bool highlight) {
            GUILayout.BeginHorizontal(highlight ? Util.Parameters_highlight : Util.Parameters_normal);
            GUILayout.Label("True Anomaly: ", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.Label(cb.orbit.trueAnomaly.ToString("f2"), Util.StdLabelSmallCustom(Color.white));
            GUILayout.Label("°", Util.StdLabelSmallCustom(highlight ? Color.yellow : ColorName));
            GUILayout.FlexibleSpace();
            GUILayout.EndHorizontal();
            return GUI.Toggle(GUILayoutUtility.GetLastRect(), highlight, "", GUIStyle.none);
        }


        //private AtmosphericEngine atmosEngine;
        private AtmosphericEngine findActiveEngine() { //called from system.vesselIteratorTarget too so vessel will be null for it if no target.
            foreach (Part p in vessel.parts) {
                if (p is AtmosphericEngine && ((AtmosphericEngine)p).engineEnabled)
                    return (AtmosphericEngine)p;
            }
            return null;
        }

        
        /*private void OnStageChange(VesselIteratorOwn v) { //called from system.vesselIteratorTarget too so vessel will be null for it if no target.
            if(vessel != null){// && v.hasAtmosEngine){
                if (atmosEngine == null) {
                    atmosEngine = findActiveEngine();
                } else {
                    if (!atmosEngine.engineEnabled)
                        atmosEngine = findActiveEngine();
                }
            }
        }*/






    }
}
