﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {

    static class LazorSoundFiles {
        private static AudioClip[] clips;
        public static AudioClip[] Clips {get { return clips; }}
        //private static string rootPath;
        public static bool initialized;

        public static void Initialize() {
            if (!initialized) {
                //rootPath = "file://" + KSPUtil.ApplicationRootPath.Replace("\\", "/") + "PluginData/romfarer/sound/";
                clips = new AudioClip[7];
                clips[0] = LoadClip("lazor.wav");
                clips[1] = LoadClip("lazor_off.wav");
                clips[2] = LoadClip("switch.wav");
                clips[3] = LoadClip("teleport.wav");
                clips[4] = LoadClip("afterburner.wav");
                clips[5] = LoadClip("achievement.wav");
				clips[6] = LoadClip("servo.wav");
            }
            initialized = true;
        }

        public static AudioClip LoadClip(string file) {
            /*WWW www = new WWW(rootPath + clipName);
            return www.GetAudioClip(false);*/

			return GameDatabase.Instance.GetAudioClip (RuntimeLogic.ApplicationPath + "sound/" + file.Replace(".wav",""));
        }
    }

	public enum SoundType {
		STARTUP = 0,
		SHUTDOWN = 1,
		SWITCH = 2,
		TELEPORT = 3,
		AFTERBURNER = 4,
		ACHIEVEMENT = 5,
		SERVO = 6
	}

    class LazorSound : IDisposable{

        private Part system;
        private AudioSource loopSource;
        private float systemVolume = 0;
        private float loopVolume = 0;


        public LazorSound(Part system) {
            this.system = system;
            system.audio.playOnAwake = false;
            system.audio.loop = false;
            system.audio.volume = GameSettings.SHIP_VOLUME;
            loopSource = system.gameObject.AddComponent<AudioSource>();
            loopSource.playOnAwake = false;
            loopSource.loop = true;
            LazorSoundFiles.Initialize();
        }

        public void Dispose() {
            StopLoop();
            UnityEngine.Object.Destroy(loopSource);
        }


        public void Play(SoundType st) {
            system.audio.volume = GameSettings.SHIP_VOLUME;
            if ((int)st == (int)SoundType.STARTUP || (int)st == (int)SoundType.SHUTDOWN)
                system.audio.volume = GameSettings.UI_VOLUME;
            if (system.audio != null) {
                if (system.audio.isPlaying)
                    system.audio.Stop();
                system.audio.clip = LazorSoundFiles.Clips[(int)st];
                system.audio.Play();
            }
        }

        public void PlayLoop(SoundType st) {
            if (loopSource != null) {
                loopSource.volume = 0f;
                if (loopSource.isPlaying)
                    loopSource.Stop();
                loopSource.clip = LazorSoundFiles.Clips[(int)st];
                loopSource.Play();
            }

        }

        public void SetVolumeLoop(float volume) {
            loopSource.volume = volume;
        }

        public void StopLoop() {
            loopSource.Stop();
        }

        public void OnPause() {
            if (loopSource != null) {
                systemVolume = system.audio.volume;
                loopVolume = loopSource.volume;
                system.audio.volume = 0;
                loopSource.volume = 0;
            }
        }

        public void OnResume() {
            if (loopSource != null) {
                system.audio.volume = systemVolume;
                loopSource.volume = loopVolume;
            } 
        }


    }



}
