﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

/* WEIRD ISSUE LIST:::
 * attaching a joint to something, settings its parent to a part, then destroying something. Does NOT remove the parenting relation...
 * 
 * 
 * FIX LIST: NEW SYSTEMS::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * remote controlled robots/probes
 * EVA tool: strut two parts, decouple two parts, destroy part
 * lazor mirror, new part, requires los from main lazor, purpose: to extend the gimbal range of main lazor.
 * target cam: remote uplink to another target cam
 * Tractor beam: move perpendicular, kill horizontal speed
 * vessel to vessel and part to part teleporter
 * action group. abort. teleport all to ground
 * radial lazor module for planes
 * campaign system: guy at ksc gives you missions you have to complete in order.
 * ground target for kps
 * Target list: hide debris
 * Target list: hide unloaded (beyond max range) vessels
 * travel to target in lazor guided flight for planes (compliment travel to coordinate)
 * main lazor local bearing readouts
 * help tooltpis for all buttons, labels: alt click
 * DoubleSwitch hierarchy/tree structure, for better switch and lock handling. "Parent" field in DS, and "children" list.
 * Fire and forget for missiles
 * add delayed rearming option to partmodule
 * 
 * 
 * FIX LIST: CONSIDER:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * target cam, don't show lazors
 * add flow adjustmen slider to resource transfer / dumping
 * make lazor eva a stand alone
 * joystick control for red system
 * 
 * 
 * FIX LIST: LOW PRIORITY:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * dispersal mode, broken on non vertical surfaces
 * fix exception thrown when sunbeam mouselook is turned off, low priority
 * fix exception thrown when resource "dissapears" from ship by staging or explosion
 * when 'control from here' (multiple lazorsystems on vessel) axis overlays dont turn off, nightvision dont turn off.
 * Transfer system: adjust departure for craft dV
 * convert lazor model to .mu and add normal map
 * Docking port incompatible: http://kerbalspaceport.com/lowattractiondockingport/
 * Docking port incompatible: http://forum.kerbalspaceprogram.com/showthread.php/22876
 * selecting map location in kps locations and fly to coordinate inconsistencies.
 * the 'sidebar' with the config / subsystems / targets list on it isn't visible to me unless I collapse HydroTech.
 * target cam: gimballing
 * auto balance resource, by counterparts and vessel wide
 * set 45 degree deadzone for main lazor cam
 * DestructionCollider around KSC
 * problems loading textures when the path has non english characters
 * 
 * 
 * FIX LIST: HIGH PRIORITY::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::://●▼▲►◄◆▣◈◊○★⊕☉∪∞∈°
 * 
 * 
 * TO-DO::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * 
 * 
 * DONE::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * 0.24 Compatibility
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;
using KSP.IO;
using System.Linq;
using System.Collections;

namespace Romfarer {

    class LazorSystem: Part    {

        //Lazor System state
		public LazorSystemController lazorSystemController;
        private Lazor leader; //Lazor attached on top of LazorSystem
        public Lazor Leader { get { return leader; } }
        public VesselIteratorOwn VI; //Vessel iterator for own vessel
        public VesselIteratorTarget VIT; //Vessel iterator for target
        public bool SafeMode { get { return leader.safeMode; } }
        public bool leaderLocked; //Is leader locked on target?
        private bool initiated;
        private bool isHorizontal; //determines if the vessel is a rocket or a plane
        public bool IsHorizontal { get { return isHorizontal; } }
        public DoubleSwitch lazorOn; //Is the Lazor System main module powered on?
		public DoubleSwitch LazorMainWindowOn; //Is the main lazor window on?
        private Component lastTarget;
        private Component target;
        public Component Target { //Main target for the system
            get { return target; }
            set {
                if (value == null) {
                    target = null;
                    leaderLocked = false;
                } else {
                    target = value;
                    leaderLocked = true;
                    if (value is Vessel)
                        VIT = new VesselIteratorTarget(this, value as Vessel);
                }
                config.OnTargetSwitch();
				guiTargetCam.OnTargetSwitch();
            }
        }
        private Transform lazorBase;
        private Transform lazorCrystal;
        private Vessel.Situations lastSituation;
        private Dictionary<string, KSPParseable> parsedData; //load state into this, onLoad runs before onStart
        private CelestialBody currentBody;
        //private bool iconSet; //used to force icon to set in game
        //private bool iconSetEditor; //used to force icon to show in editor
        public LocalEvents localEvents;

		//Locks
		public bool sysLock_HOVER; //LOCK: So only one hover system can run
		public bool sysLock_LEADER_ignorehits; //LOCK: make leader ignore raycast hits
		public SystemLock sysLock_WaitDisconnect = new SystemLock("sysLock_WaitDisconnect");
		public SystemLockTimed sysLock_WaitCouple = new SystemLockTimed("sysLock_WaitCouple");
		//public bool sysLock_WaitCouple;
		//public int sysLock_WaitCouple_frame = int.MaxValue;

        //GUI
        public Rect windowPos;
        private string stringMark = ""; //'Mark' textField input
        private string stringHeading = ""; //'Heading' textField input
        private GUIContent[] upAxisButtons;
        public float lazorRange = 500; //How far the lazor travels
        public bool toggleFreeAim;
        private bool keyToggleFreeAimPrevious;
        public Vector2 mousePos;
        public bool useLocalCoordinates;
        public enum Axis { hUp = 0, vUp = 1 };
        public Axis upAxis;
        private bool flareTransition;

        //Ref's
        public DebugAssist debugAssist;
        public LazorSound lazorSound;
        public MovementControlExtended mc;
        public List<Component> targetsTracked = new List<Component>(); //Targets actively tracked by the system
        public GUItargetsList targetList = new GUItargetsList(); //targets for Target list
        public GUIcontroller guiController;
        public GUItargets guiTargets; //Target GUI
        public GUIsubsystems guiSubSystems; //Target GUI
        public GUItargetCam guiTargetCam; //Target GUI
        public GUIparameters guiParameters;
        public GUIoverlays guiOverlays;
        public GUIconfig guiConfig;
        public GUImodules guiModules;
        public GUIachievements guiAchievements;
        public Config config;
        public PlanetariumCamera planetariumCamera;

        //Orientation and heading
        public float actualBearing;
        public float actualMark;
        public float actualRoll;
        public Vector3d vectorNorth;
        public Vector3d vectorEast;
        public Quaternion orientationVessel;
        public Quaternion orientationLeader;
        public Vector3d nonRotatingFrameVelocity; //orbit velocity
        public Vector3d rotatingFrameVelocity; //velocity relative to SoI

        //From .cfg file
        public string systemType = "DEFAULT"; //DEFAULT, STACK, RADIAL
        public bool systemIsStacked;

        //Constants/parameters
        public double LONGITUDE = 0f;
        public double LATITUDE = 0f;
        public float CRAFT_MASS { get { return VI.mass; } }
        public bool GAMEPAUSED;
        private const string REGEX_surfaceCollider = "..[0-9][0-9][0-9][0-9][0-9][0-9]*";
	
        //Graphics
        private float flareTimeOn = 0f; //Timer from when flare was activated
        private ParticleEmitter flare_emitter;
        private ParticleAnimator flare_emitterAnimator;
        private Light flare_light;
        private Color[] flare_emitterColors;
        private Texture2D textureRed = Util.getTexture(new Color(0.6f, 0, 0, 0.9f));
        private Texture2D textureBlue = Util.getTexture(new Color(0, 0, 0.6f, 0.9f));
        private Texture2D textureGreen = Util.getTexture(new Color(0, 0.6f, 0, 0.9f));
        private Texture2D textureWhite = Util.getTexture(new Color(0.6f, 0.6f, 0.6f, 0.9f));
        private Material crystal_white; //material for main lazor crystal
        private Material crystal_red;
        private Material crystal_green;
        private Material crystal_blue;
        public static Color COLOR_list_identifier = new Color(0.282f, 1f, 0.803f, 1);
		private ParticleEmitter emitterSonar;

        //Remote control
        private List<Part> remote_stageQueue = new List<Part>();
        private Vector3 remote_dirSAS;
        private Vector3 remote_torque = Vector3.zero;
        private float remote_maxTorque;
        private float remote_maxTorqueF;
		private bool remote_useControlSurface; //if true, use control surfaces + torgue to control vessel
        private bool remote_yawPlus;
        private bool remote_yawMinus;
        private bool remote_pitPlus;
        private bool remote_pitMinus;
        private bool remote_rolPlus;
        private bool remote_rolMinus;
        private float remote_trottle;
        private Vessel remote_targetVessel;
        private bool remote_toggleRemoteControlLOCK;
        private DoubleSwitch remote_toggleRemoteControl;
        private DoubleSwitch remote_toggleSAS;
        private DoubleSwitch remote_togglePrograde;
        private DoubleSwitch remote_toggleRetrograde;
        private DoubleSwitch remote_toggleRetroGee;
        private DoubleSwitch remote_toggleProGee;
        private DoubleSwitch remote_toggleFollow;

        //camera
        private CameraModes[] cameraModes = new[] { CameraModes.Normal, CameraModes.NightVision, CameraModes.Invert };
        private string[] cameraModesStrings = new string[] { "Normal", "Night", "Invert" };
        private int cameraMode = 0;
        private CameraFilter cameraFilter;

		//teleport
		private bool toggleTeleport;
		private Vector3d surfacePos = Vector3.zero;
		private bool toggleTeleportShort;

		//scan
		private float scanTimer;
		private float sonarRange = 500;
		private bool toggleScanPulse;


        //Overrides
        //override protected void onJointDisable() { print("++++++ onJointDisable ++++++"); }
        //override protected void onJointReset() { print("++++++ onJointReset ++++++"); }
        //override protected void onPartAwake() { print("++++++ onPartAwake ++++++"); }
        //override protected bool onPartActivate() { print("++++++ onPartActivate ++++++"); return true; }
        //override protected void onPartDeactivate() { print("++++++ onPartDeactivate ++++++"); }
        //override protected void onPartDelete() { print("++++++ onPartDelete ++++++"); }
        //override protected void onPartDetach() { print("++++++ onPartDetach ++++++"); }
        //override protected void onPartExplode() { print("++++++ onPartExplode ++++++"); }
        //override protected void onPartLiftOff() { print("++++++ onPartLiftOff ++++++"); }
        //override protected void onPartSplashdown() { print("++++++ onPartSplashdown ++++++"); }
        //override protected void onPartTouchdown() { print("++++++ onPartTouchdown ++++++"); }
        //override protected void onUnpack() { print("++++++ onUnpack ++++++"); }
        //override protected void onPack() { print("++++++ onUnpack ++++++ "+vessel.vesselName); }
        //override protected void onDisconnect() { print("++++++ onDisconnect ++++++"); }
        //override public void OnDrawStats() { print("++++++ onDrawstats ++++++"); }


        [KSPAction("MouseLook Toggle", KSPActionGroup.Custom01)]
        public void ToggleMouseLook(KSPActionParam param) {
            if (lazorOn && !MapView.MapIsEnabled) {
                toggleFreeAim = !toggleFreeAim;
                leaderLocked = false;
            }
        }

        [KSPAction("Color Toggle", KSPActionGroup.Custom02)]
        public void ToggleColor(KSPActionParam param) {
            if (lazorOn && !MapView.MapIsEnabled) {
                if (leader.GetColor() == Color.white)
                    SetLeaderColor(Color.red);
                else if (leader.GetColor() == Color.red)
                    SetLeaderColor(Color.green);
                else if (leader.GetColor() == Color.green)
                    SetLeaderColor(Color.blue);
                else if (leader.GetColor() == Color.blue)
                    SetLeaderColor(Color.white);
                leaderLocked = false;
            }
        }

        override protected void onGamePause() {
            GAMEPAUSED = true;
            if (!IsActiveLazorSystem())
                return;
            //print("++++++ onGamePause ++++++");
            keyToggleFreeAimPrevious = Screen.showCursor;
            Screen.showCursor = true;
            if(lazorSound != null)
                lazorSound.OnPause();
        }

        override protected void onGameResume() {
            GAMEPAUSED = false;
			if (!IsActiveLazorSystem())
                return;
            //print("++++++ onGameResume ++++++");
            Screen.showCursor = keyToggleFreeAimPrevious;
            if (lazorSound != null)
                lazorSound.OnResume();
        }

        //called at the end of onFlightStart
        private void LateFlightStateLoad(Dictionary<string, KSPParseable> parsedData) {
			if(parsedData != null) {
				Util.LoadKSPParseableRect(parsedData, "LazorSystem_windowPos", ref windowPos);
				if(parsedData.ContainsKey("LazorSystem_lazorOn"))
					lazorOn.set = parsedData["LazorSystem_lazorOn"].value_bool;
				if(parsedData.ContainsKey("LazorSystem_LazorMainWindowOn"))
					LazorMainWindowOn.set = parsedData["LazorSystem_LazorMainWindowOn"].value_bool;
				if(parsedData.ContainsKey("LazorSystem_Color"))
					SetLeaderColor(Util.StringToColor(parsedData["LazorSystem_Color"].value));
				guiAchievements.onFlightStateLoad(parsedData);
				guiConfig.onFlightStateLoad(parsedData);
				guiModules.onFlightStateLoad(parsedData);
				guiParameters.onFlightStateLoad(parsedData);
				guiSubSystems.onFlightStateLoad(parsedData);
				guiTargetCam.onFlightStateLoad(parsedData);
				guiTargets.onFlightStateLoad(parsedData); 
				config.onFlightStateLoad(parsedData);
				guiController.onFlightStateLoad(parsedData);
			} else
				LazorMainWindowOn.set = true;
			LazorGlobals.ClearPlanetOverlay(); //make grid lines remake themselves when turned on again
        }

        public override void onFlightStateLoad(Dictionary<string, KSPParseable> parsedData) { 
            //print("++++++ onFlightStateLoad ++++++");
            base.onFlightStateLoad(parsedData);
            this.parsedData = parsedData; //store and load at the end of onFlightStart
        }

        public override void onFlightStateSave(Dictionary<string, KSPParseable> partDataCollection) { 
            //print("++++++ onFlightStateSave ++++++");
            if (partDataCollection != null) {
                base.onFlightStateSave(partDataCollection);
                if(windowPos!=null)
                    Util.SaveKSPParseableRect(partDataCollection, "LazorSystem_windowPos", windowPos);
                partDataCollection.Add("LazorSystem_lazorOn", new KSPParseable((bool)lazorOn, KSPParseable.Type.BOOL));
                partDataCollection.Add("LazorSystem_LazorMainWindowOn", new KSPParseable((bool)LazorMainWindowOn, KSPParseable.Type.BOOL));
                if(Leader != null)
                    partDataCollection.Add("LazorSystem_Color", new KSPParseable(Util.ColorToString(Leader.GetColor()), KSPParseable.Type.STRING));
                
                if (guiAchievements != null)
                    guiAchievements.onFlightStateSave(partDataCollection);
                if (guiConfig != null)
                    guiConfig.onFlightStateSave(partDataCollection);
                if (guiModules != null)
                    guiModules.onFlightStateSave(partDataCollection);
                if (guiParameters != null)
                    guiParameters.onFlightStateSave(partDataCollection);
                if (guiSubSystems != null)
                    guiSubSystems.onFlightStateSave(partDataCollection);
                if (guiTargetCam != null)
                    guiTargetCam.onFlightStateSave(partDataCollection);
                if (guiTargets != null)
                    guiTargets.onFlightStateSave(partDataCollection);

				if(config != null)
					config.onFlightStateSave(partDataCollection);
                if (guiController != null)
                    guiController.onFlightStateSave(partDataCollection);
            }
        }

        protected override void onPartAttach(Part part){
            base.onPartAttach(part);
            if (EditorLogic.fetch.editorType == EditorLogic.EditorMode.SPH)
                customPartData = "H";
            else
                customPartData = "V";
			SetIcon();
        }

        protected override void onPartDestroy() {
            Romfarer.Events.onVesselChange -= new Romfarer.Events.OnVesselChange(OnVesselChange);
			if(localEvents != null)
            	localEvents.onStageChange -= OnStageChange;
            if (vessel != null) {
                RenderingManager.RemoveFromPostDrawQueue(1, new Callback(drawGUI2));
                RenderingManager.RemoveFromPostDrawQueue(0, new Callback(drawGUI));
            }
			if(config != null)
				config.DisposeCFGs();
            if(guiOverlays!=null)
                guiOverlays.Dispose();
            Screen.showCursor = true;
			if(guiAchievements != null)
				guiAchievements.OnDestroy();
			Util.ResetGUI();
			TurnOffRemoteControl();

			if (cameraFilter != null)
				cameraFilter.SetMaterial(CameraModes.Normal);

			if(IsActiveLazorSystem())
				RuntimeLogic.Instance.AppLauncherButtonDisable();
        }
		
        public void OnVesselChange() {
            if (cameraFilter != null)
                cameraFilter.SetMaterial(CameraModes.Normal);
			TargetCheck_SystemSwitch();
        }

        public void OnStageChange(VesselIteratorOwn vi) {
            if (cameraFilter != null)
                cameraFilter.SetMaterial(CameraModes.Normal);
            //VI.OnVesselChange();
            //config.Reconfigure(VI);
        }

        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        // INITIALIZATION
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************

        protected override void onPartLoad() {
            base.onPartLoad();
            Save_VAB_SPH_data();
            //print("++++++ onPartLoad ++++++");
        }

        private void Save_VAB_SPH_data() {
            isHorizontal = customPartData == "H";
        }

        protected override void onPartStart() {
			if(HighLogic.LoadedScene == GameScenes.EDITOR || HighLogic.LoadedScene == GameScenes.SPH)
				StartCoroutine(SetIconDelayed());
			else {
				SetIcon();
			}
        }

		private IEnumerator SetIconDelayed(){
			yield return null;
			yield return null;
			SetIcon();
		}

		private void SetIcon(){
			try {
				base.stackIcon.SetIcon(this, RuntimeLogic.ApplicationPath + "textures/icons", 0, 0);
				base.stackIcon.CreateIcon();
			} catch (Exception) { }
		}

        protected override void onFlightStart(){
			//print("- -- --- FLIGHT START --- -- -" + vessel.vesselName);

			if(!Modules.Contains("LazorSystemMainModule")) {
				PartModule mem = AddModule("LazorSystemMainModule");
				mem.Fields.SetValue("toggleMouseLook", false);
			}

			Save_VAB_SPH_data();

			localEvents = new LocalEvents();
			LazorGlobals.AddDestructionTracker(vessel);

			if(isHorizontal)
				upAxis = Axis.hUp;
			else
				upAxis = Axis.vUp;
			upAxisButtons = new GUIContent[2] {
				new GUIContent("Plane", "Configure as Plane"),
				new GUIContent("Rocket", "Configure as Rocket")
			};

			stackIcon.SetBgColor(Color.black);
			debugAssist = new DebugAssist();  //CAUSE OF LAZOREGG BUG

			foreach(Part p in vessel.parts) {} //load all parts.

			if(lazorSystemController == null)
				lazorSystemController = new LazorSystemController(vessel, this);
			
			//initialize IModules, Lazors are IModules
			/*foreach(Part p in vessel.parts)
				if(p is IModule)
					((IModule)p).Initialize();*/
			leader = new Lazor();
			//leader.safeMode = true;
			systemIsStacked = systemType == "STACK";
			if(systemIsStacked)
				leader.Initialize(this, transform, new Vector3(0, 0, 0), Color.white);
			else
				leader.Initialize(this, transform, new Vector3(0, 0.368f, 0), Color.white);
			leader.GetTransform().gameObject.layer = 19; //shown by the main camera but not target cam
			lazorBase = transform.Find("model/LazorSystem/Lazor");
			lazorCrystal = transform.Find("model/LazorSystem/LazorCrystal");
			lazorCrystal.parent = lazorBase;
			crystal_white = Util.GetMaterial(Shader.Find("Specular"), Color.white);
			crystal_red = Util.GetMaterial(Shader.Find("Specular"), Color.red);
			crystal_green = Util.GetMaterial(Shader.Find("Specular"), Color.green);
			crystal_blue = Util.GetMaterial(Shader.Find("Specular"), Color.blue);
			Util.SetMaterial(lazorCrystal.gameObject, crystal_white);

			int lazorSystemCount = 0;
			// Find all lazors attached to the vessel and initialize them.
			foreach(Part p in vessel.parts) {
				if(p is LazorSystem)
					lazorSystemCount++;
			}

			//MonoBehaviour.print("[" + this + " " + vessel.vesselName + "  "+ uid.ToString() + ": Found " + lazors.Count + " attached lazors]");

			mc = new MovementControlExtended(this);
			guiController = new GUIcontroller(this);

			VI = new VesselIteratorOwn(this, vessel);
			VIT = new VesselIteratorTarget(this, vessel); //set to avoid VIT from throwing exception on celestial bodies as they use VIT to draw parameters.
			config = new Config(this);
			guiTargets = new GUItargets(this);
			guiSubSystems = new GUIsubsystems(this);
			guiTargetCam = new GUItargetCam(this);
			guiParameters = new GUIparameters(this);
			guiOverlays = new GUIoverlays(this);
			guiConfig = new GUIconfig(this);
			guiModules = new GUImodules(this);
			guiAchievements = new GUIachievements(this);
			guiController.CountTabs();

			//Activate safe mode
			//safemode should always be on...........
			/*leader.safeMode = true;
			foreach(LazorAimable l in lazors)
				l.Lazor.safeMode = true;*/

			//Flare.....
			GameObject o = Instantiate(UnityEngine.Resources.Load("Effects/fx_exhaustFlame_blue")) as GameObject;
			o.transform.parent = transform;
			o.transform.localEulerAngles = Vector3.zero;
			o.transform.localPosition = new Vector3(0, 0.11f, 0);
			if(systemIsStacked)
				o.transform.localPosition = Vector3.zero;
			flare_emitter = o.particleEmitter;
			flare_emitter.emit = true;
			flare_emitter.minSize = 0F;
			flare_emitter.maxSize = 0F;
			flare_emitter.rndVelocity = Vector3.one * 0.15f;
			flare_emitter.localVelocity = Vector3.zero;
			flare_emitter.useWorldSpace = false;
			(o.GetComponent<ParticleAnimator>() as ParticleAnimator).sizeGrow = -0.8F;
			flare_emitterAnimator = flare_emitter.GetComponent<ParticleAnimator>();
			flare_emitterColors = flare_emitterAnimator.colorAnimation;
			GameObject o2 = new GameObject();
			o2.transform.parent = transform;
			o2.transform.localEulerAngles = Vector3.zero;
			o2.transform.localPosition = new Vector3(0, 0.11f, 0);
			if(systemIsStacked)
				o2.transform.localPosition = Vector3.zero;
			o2.AddComponent<Light>();
			flare_light = o2.light;
			flare_light.color = new Color(1, 1, 1, 1);
			flare_light.renderMode = LightRenderMode.ForcePixel;
			flare_light.type = LightType.Point;
			flare_light.transform.parent = transform;
			flare_light.transform.position = o2.transform.position;
			flare_light.enabled = false;

			planetariumCamera = (PlanetariumCamera)GameObject.FindObjectOfType(typeof(PlanetariumCamera));
			lazorSound = new LazorSound(this);

			//GUI stuff
			if((windowPos.x == 0) && (windowPos.y == 0))
				windowPos = new Rect(Screen.width - 40, 100, 40, 80);

			RenderingManager.AddToPostDrawQueue(0, new Callback(drawGUI)); //start the GUI
			RenderingManager.AddToPostDrawQueue(1, new Callback(drawGUI2)); //start the GUI

			//onPartStart is called after onFlightStateLoad.
			LateFlightStateLoad(parsedData); //late loading

			Romfarer.Events.onVesselChange += new Romfarer.Events.OnVesselChange(OnVesselChange);
			localEvents.onStageChange += OnStageChange;

            //setup camera filters
			cameraFilter = FlightCamera.fetch.mainCamera.gameObject.GetComponent<CameraFilter>();
			if (cameraFilter == null)
				cameraFilter = FlightCamera.fetch.mainCamera.gameObject.AddComponent<CameraFilter>();

            //Scanner
            GameObject ooo = Part.Instantiate(UnityEngine.Resources.Load("Effects/fx_exhaustFlame_blue")) as GameObject;
            ooo.transform.parent = transform;
            ooo.transform.localEulerAngles = Vector3.zero;
            ooo.transform.localPosition = new Vector3(0, 0, 0);
            emitterSonar = ooo.particleEmitter;
            emitterSonar.emit = false;
            emitterSonar.minSize = 0.5F;
            emitterSonar.maxSize = 1F;
            emitterSonar.minEnergy = 9.5F;
            emitterSonar.maxEnergy = 10F;
            emitterSonar.rndVelocity = Vector3.one * 50;
            emitterSonar.localVelocity = Vector3.zero;
            emitterSonar.useWorldSpace = false;
            ParticleAnimator emitterAnimator = emitterSonar.GetComponent<ParticleAnimator>();
            Color[] emitterColors = emitterAnimator.colorAnimation;
            emitterColors[0] = new Color(0f, 0.9f, 0f, 0.6f);
            emitterColors[1] = new Color(0f, 0.8f, 0f, 0.8f);
            emitterColors[2] = new Color(0f, 0.6f, 0f, 0.6f);
            emitterColors[3] = new Color(0f, 0.4f, 0f, 0.4f);
            emitterColors[4] = new Color(0f, 0.2f, 0f, 0.2f);
            emitterAnimator.colorAnimation = emitterColors;
         
			initiated = true;
			//print("- -- --- FLIGHT START --- -- - (end)" + vessel.name);

			GameObject kspObj = GameObject.Find ("KSCSpacePlaneHangar/model_sph_exterior_holder");
			if(kspObj != null){
				Collider col = kspObj.GetComponentInChildren<MeshCollider>();
				if(col != null)
					targetList.AddReplacement(new GUItargetsListItem("" + col.GetInstanceID(), TargetType.KSC, "Space Plane Hangar", col, 0, 0));
			}

			kspObj = GameObject.Find ("KSCTrackingStation/model_tracking_exterior_main_v30");
			if(kspObj != null){
				LazorGlobals.trackingStation = kspObj.GetComponent<MeshCollider>();
				if(LazorGlobals.trackingStation != null)
					targetList.AddReplacement(new GUItargetsListItem("" + LazorGlobals.trackingStation.GetInstanceID(), TargetType.KSC, "Tracking Station", LazorGlobals.trackingStation, 0, 0));
			}

			kspObj = GameObject.Find ("KSCLaunchPad/model_launchpad_ground_v46n");
			if(kspObj != null){
				Collider col = kspObj.GetComponentInChildren<MeshCollider>();
				if(col != null)
					targetList.AddReplacement(new GUItargetsListItem("" + col.GetInstanceID(), TargetType.KSC, "Launch Pad", col, 0, 0));
			}

			kspObj = GameObject.Find ("KSCVehicleAssemblyBuilding/model_vab_exterior_main_v30");
			if(kspObj != null){
				Collider col = kspObj.GetComponent<MeshCollider>();
				if(col != null)
					targetList.AddReplacement(new GUItargetsListItem("" + col.GetInstanceID(), TargetType.KSC, "Vehicle Assembly Building", col, 0, 0));
			}

			kspObj = GameObject.Find ("KSCRunway/model_runway_new_v44_collider");
			if(kspObj != null){
				Collider col = kspObj.GetComponent<MeshCollider>();
				if(col != null)
					targetList.AddReplacement(new GUItargetsListItem("" + col.GetInstanceID(), TargetType.KSC, "Runway", col, 0, 0));
			}

			//Vessels
			foreach(Vessel v in FlightGlobals.Vessels)
				targetList.AddReplacement(new GUItargetsListItem("" + v.id, TargetType.Vessel, v.vesselName, v, 0, 0));
        }


        /// <summary>
        /// Resets the internal state of lazor system
        /// </summary>
        public void ResetState() {
			MonoBehaviour.print("[" + this + " : Resetting vessel state for vessel " + vessel.vesselName+"]");
            VI = new VesselIteratorOwn(this, vessel);
            Target = null;
            targetList.AddReplacement(new GUItargetsListItem("" + vessel.id, TargetType.Vessel, vessel.vesselName, vessel, 0, 0));  //add vessel to targetlist
            config = new Config(this);
            guiController.CountTabs();
        }


        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        // GUI
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************
		
        private void drawGUI2() {
            if(!IsActiveLazorSystem())
                return;
            GUI.skin = Util.LazorSkin;
            GUI.depth = 100;
            //draw other gui windows
            if(lazorOn)
                guiController.DrawGUIs();
        }

        private void drawGUI() {
			if (!IsActiveLazorSystem())
                return;

            stackIcon.SetIconColor(leader.GetColor()); //give the icon the leaders color

            GUI.skin = Util.LazorSkin;
            GUI.depth = 0;

            if (lazorOn && LazorMainWindowOn) {
				windowPos = Util.ClampToScreen(GUILayout.Window(GetInstanceID(), windowPos, WindowGUI, "LAZOR SYSTEM", Util.Window_border));
                if (Event.current.type == EventType.Repaint)
                    windowPos.height = 80;
                windowPos.width = 200;
			} else if(lazorOn && !LazorMainWindowOn){
				//windowPos.x = Screen.width - 40;
                windowPos.width = 40;
                windowPos.height = 80;
				windowPos = Util.ClampToScreen(GUILayout.Window(GetInstanceID(), windowPos, LazorButton, "", Util.Window_lazorButton));
            }
			mousePos = Event.current.mousePosition;
        }

        private void LazorButton(int windowID) {
            GUILayout.BeginVertical(Util.Space_Compact);
			//lazorOn.set = GUILayout.Toggle(lazorOn, "", GUILayout.Width(20), GUILayout.Height(20));
            if(GUILayout.Button("", Util.Button_lazorButton, GUILayout.Width(20), GUILayout.Height(28)))
                LazorMainWindowOn.set = lazorOn.set = true;
            GUILayout.EndVertical();
            GUI.DragWindow(new Rect(20, 0, 20, 180));
        }

        private void WindowGUI(int windowID){
            GUILayout.BeginVertical();
            if (GUI.Button(new Rect(windowPos.width - 20, 0, 20, 20), "", Util.Button_header_minimize))
                LazorMainWindowOn.set = false;

            if (lazorOn) {
                GUILayout.BeginVertical(Util.Space_Formatted); //Formatted area start..................................
                GUILayout.BeginHorizontal();
                Util.DistanceFormatted(lazorRange, Color.white, "Range: ");
                GUILayout.EndHorizontal();
				lazorRange = GUILayout.HorizontalSlider(lazorRange, 500.0f, LazorGlobals.VIEW_DISTANCE);

                toggleFreeAim = GUILayout.Toggle(toggleFreeAim, "Toggle Mouse Look"); //Mouse\nLook
                GUILayout.Space(5);

                //Target displaty
                GUILayout.BeginHorizontal(Util.ListItemBG_compact);
                GUILayout.Label("Target: ", Util.StdLabelSmallCustom_Compact(Color.yellow));
                GUILayout.Label(Util.CutString(TargetName(), 22, ".."), Util.StdLabelSmallCustom_Compact(Color.white));
                GUILayout.EndHorizontal();
                if (Target != null) {
                    GUILayout.BeginHorizontal(Util.ListItemBG_compact);
                    Util.DistanceFormatted(TargetDistance(), "Distance: ", Util.StdLabelSmallCustom_Compact, Color.yellow, Color.white);
                    GUILayout.EndHorizontal();
                } else {
                    GUILayout.BeginHorizontal(Util.ListItemBG_compact);
                    GUILayout.Label("Distance: ", Util.StdLabelSmallCustom_Compact(Color.yellow));
                    GUILayout.EndHorizontal();
                }

                //Main lazor hit display
                GUILayout.BeginHorizontal(Util.ListItemBG_compact);
                GUILayout.Label("Lazor hit: ", Util.StdLabelSmallCustom_Compact(Color.green));
                if (Leader.LazorIsHittingOwnVessel)
                    GUILayout.Label(Util.CutString("Own vessel", 22, ".."), Util.StdLabelSmallCustom_Compact(Color.white)); 
                else if (Leader.LazorIsHittingVessel)
                    GUILayout.Label(Util.CutString(Leader.VesselHit.vesselName, 22, ".."), Util.StdLabelSmallCustom_Compact(Color.white));
                else if (Leader.LazorIsHittingSomething)
					GUILayout.Label(Util.CutString(leader.GetHit().collider.name, 22, ".."), Util.StdLabelSmallCustom_Compact(Color.white));
                else
                    GUILayout.Label("nothing", Util.StdLabelSmallCustom_Compact(Color.white));
                GUILayout.EndHorizontal();

				if (Leader.LazorIsHittingSomething && !Leader.LazorIsHittingOwnVessel) {
                    GUILayout.BeginHorizontal(Util.ListItemBG_compact);
                    Util.DistanceFormatted(leader.HitDistance(), "Distance: ", Util.StdLabelSmallCustom_Compact, Color.green, Color.white);
                    GUILayout.EndHorizontal();
                } else {
                    GUILayout.BeginHorizontal(Util.ListItemBG_compact);
                    GUILayout.Label("Distance: ", Util.StdLabelSmallCustom_Compact(Color.green));
                    GUILayout.EndHorizontal();
                }
                GUILayout.EndVertical(); //Formatted area end..................................

                GUILayout.Space(3);

                //Color buttons ........................................................................................
                GUILayout.BeginVertical();

				DrawDivider("        " + "Support", Color.white, Util.Button_color_white, Content_White);
				DrawDivider("        " + "Attitude", Color.red, Util.Button_color_red, Content_Red);
				DrawDivider("        " + "Scan", Color.green, Util.Button_color_green, Content_Green);
				DrawDivider("        " + "Remote control", Color.blue, Util.Button_color_blue, Content_Blue);

                GUILayout.EndVertical();

                GUILayout.Space(5);

                GUILayout.BeginVertical(Util.Space_Formatted); //window split start ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

                // HELPERS........................................................................................
                GUILayout.Space(5);
                //debugAssist.TestButtons();

                GUILayout.EndVertical(); //window split end ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
            } //end lazorOn check
            GUILayout.EndVertical();
            GUI.DragWindow(new Rect(0, 0, 300, 30));
        }


		private void DrawDivider(string title, Color color, GUIStyle style, Axtion method){
			GUILayout.BeginHorizontal();
			GUILayout.BeginVertical(Util.Space_Compact);
			GUILayout.BeginHorizontal(Leader.GetColor() == color ? Util.Divider : Util.Divider_gray);
			GUILayout.BeginHorizontal(Util.Space_Compact);
			GUILayout.Label(title, Util.LabelBold);
			GUILayout.EndHorizontal();
			GUILayout.EndHorizontal();
			if (Leader.GetColor() == color) {
				GUILayout.Space(3);
				method();
				GUILayout.Space(7);
			}
			GUILayout.EndVertical();
			GUILayout.BeginHorizontal(Util.Space_Compact, GUILayout.Width(20));
			if (GUILayout.Button("", style, GUILayout.ExpandHeight(true), GUILayout.Width(20)))
				SetLeaderColor(color);
			GUILayout.EndHorizontal();
			GUILayout.EndHorizontal();
		}
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        // CONTENT 
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************


        private void Content_White() {
            if (vessel.GetCrewCount() > 0) {
                GUILayout.Label("Teleport Out: ", Util.StdLabelSmallCustom(Color.yellow));
                foreach (ProtoCrewMember pcm in vessel.GetVesselCrew()) {
                    GUILayout.BeginHorizontal();
                    GUILayout.Label(pcm.name);
                    GUILayout.FlexibleSpace();
                    if (!FlightGlobals.ActiveVessel.isEVA) {
                        if (GUILayout.Button("►", GUILayout.Width(18))) {
                            if (Leader.LazorIsHittingSomething && Leader.HitDistance() < 2500) {
                                surfacePos = Leader.GetHitPoint();
                                toggleTeleportShort = true;
                                lazorSound.Play(SoundType.TELEPORT);
                                FlightEVA.SpawnEVA(pcm.KerbalRef);
								MonoBehaviour.print("[" + this + " : Teleport starting from: " + Leader.HitDistance().ToString("f0") + "m] ");
                            } else {
                                double distance;
                                if (PQS.LineSphereIntersection(Leader.GetOrigin() - vessel.mainBody.position, Leader.GetAim(), vessel.mainBody.Radius, out surfacePos)) {
                                    surfacePos += vessel.mainBody.position;
                                    distance = (surfacePos - vessel.transform.position).magnitude;
									if (distance < LazorGlobals.TELEPORT_DISTANCE) {
                                        toggleTeleport = true;
                                        lazorSound.Play(SoundType.TELEPORT);
                                        FlightEVA.SpawnEVA(pcm.KerbalRef);
										MonoBehaviour.print("[" + this + " : Teleport starting from: " + distance.ToString("f0") + "m] ");
                                    } else
										MonoBehaviour.print("[" + this + " : Teleport is out of range: " + distance.ToString("f0") + "m. max range is: " + LazorGlobals.TELEPORT_DISTANCE + "m]");
                                } else
									MonoBehaviour.print("[" + this + " : Teleport could not lock on the surface]");
                            }
                        }

                    }
                    GUILayout.EndHorizontal();
                }
            }
            //Beam in buttons
            if (Target != null && Target is Vessel && ((Vessel)Target).isEVA) //beam aboard a kerbal
                Util.GUIButton_BeamAboard(this, (Vessel)Target);

            if (vessel.GetCrewCount() > 0)
                GUILayout.Space(5);

            //Nightvision etc.
            GUILayout.Label("Vision enhancer: ", Util.StdLabelSmallCustom(Color.yellow));
            cameraMode = GUILayout.SelectionGrid(cameraMode, cameraModesStrings, 3);

            cameraFilter.SetMaterial(cameraModes[cameraMode]);
        }


        private void Content_Red() {
            GUILayout.Label("Main lazor heading: ", Util.StdLabelSmallCustom(Color.yellow));

            GUILayout.BeginHorizontal();
            GUILayout.Label("Axis toggle");
            if (GUILayout.Button(useLocalCoordinates ? "Local" : "Global", GUILayout.Width(50)))
                useLocalCoordinates = !useLocalCoordinates;
            if (GUILayout.Button("Reset")) {
                leaderLocked = false;
                leader.ResetOrientation();
				config.ResetLazorOrientations();
            }
            GUILayout.EndHorizontal();
            // Leaders position as text input area........................................................................................
            GUILayout.BeginHorizontal();
            GUILayout.Label("Bearing ");
            GUI.SetNextControlName("bearing");
            float angleR = mc.GetRotation(leader);
            stringHeading = GUILayout.TextField(GUI.GetNameOfFocusedControl() == "bearing" ? stringHeading : angleR.ToString("f1"), 8, GUILayout.Width(44)).Replace("\n", "").Replace("\r", "");
            if (Event.current.keyCode == KeyCode.Return && Event.current.type == EventType.keyUp && GUI.GetNameOfFocusedControl() == "bearing") {
                mc.SetRotation(leader, float.Parse(stringHeading));
                GUI.UnfocusWindow();
                /*if (o1 || o1b) {
                    mvs.Activate();
                    print("[Set lazor Bearing] " + mvs.GetType() + " " + leader.LazorIsHittingSomething);
                }*/
            }

            GUILayout.Label("Mark ");
            GUI.SetNextControlName("mark");
            float angleA = mc.GetAngle(leader);
            stringMark = GUILayout.TextField(GUI.GetNameOfFocusedControl() == "mark" ? stringMark : angleA.ToString("f1"), 7, GUILayout.Width(44)).Replace("\n", "").Replace("\r", "");
            if (Event.current.keyCode == KeyCode.Return && Event.current.type == EventType.keyUp && GUI.GetNameOfFocusedControl() == "mark") {
                mc.SetAngle(leader, float.Parse(stringMark));
                GUI.UnfocusWindow();
                /*if (o1 || o1b) {
                    mvs.Activate();
                    print("[Set lazor Mark] " + mvs.GetType() + " " + leader.LazorIsHittingSomething);
                }*/
            }
            GUILayout.EndHorizontal();
            //guiOverlays.toggleHeading = GUILayout.Toggle(guiOverlays.toggleHeading, "Show vessel heading on map");

            GUILayout.Space(3);

            GUILayout.Label("Axis configured as: ", Util.StdLabelSmallCustom(Color.yellow));
            upAxis = (Axis)GUILayout.SelectionGrid((int)upAxis, upAxisButtons, 2); //Plane, Rocket
            isHorizontal = upAxis == Axis.hUp;

            GUILayout.Space(3);

            GUILayout.Label("Axis overlays: ", Util.StdLabelSmallCustom(Color.yellow));
            GUILayout.BeginHorizontal();
            guiOverlays.toggleAxisLocal.set = GUILayout.Toggle(guiOverlays.toggleAxisLocal.set, "Local", Util.Button_Toggle);
            guiOverlays.toggleAxisGlobal.set = GUILayout.Toggle(guiOverlays.toggleAxisGlobal.set, "Global", Util.Button_Toggle);
            GUILayout.EndHorizontal();
        }

        private void Content_Green() {
            GUILayout.BeginHorizontal(Util.Space_Compact);
            Util.DistanceFormatted(sonarRange, Color.yellow, "Emitter Scan:  ");
            GUILayout.EndHorizontal();
            sonarRange = GUILayout.HorizontalSlider(sonarRange, 100.0f, (float)VI.resourceParts["ElectricCharge"].amount * 10.0f -0.1f);
            GUI.enabled = VI.resourceParts["ElectricCharge"].amount > sonarRange / 10;
            if (GUILayout.Button("Send pulse"))
                SendPulse();
            GUI.enabled = true;
            toggleScanPulse = GUILayout.Toggle(toggleScanPulse, "Continous Scan");
            if (toggleScanPulse) {
                if (VI.resourceParts["ElectricCharge"].amount > sonarRange / 10 && Time.time > scanTimer + 3) {
                    scanTimer = Time.time;
                    SendPulse();
                }
            }
            GUILayout.Label("Requires " + (sonarRange / 10).ToString("f1") + " energy");
        }

        private void SendPulse() {
            emitterSonar.Emit(300);
            RequestEnergy((double)(sonarRange / 10.0f));
            Collider[] colliders = Physics.OverlapSphere(transform.position, sonarRange, 557059);
            foreach (Collider hit in colliders) {
                if (hit.attachedRigidbody != null) {
                    Part hitPart = hit.attachedRigidbody.GetComponent<Part>();
                    if (hitPart != null) {
                        Vessel hitVessel = hitPart.vessel;
                        if (hitVessel != null && hitVessel != vessel)
                            targetList.AddReplacement(new GUItargetsListItem("" + hitVessel.id, TargetType.Vessel, hitVessel.vesselName, hitVessel, 0, 0));
                    }
                } else if (!System.Text.RegularExpressions.Regex.IsMatch(hit.name, REGEX_surfaceCollider))
                    if (hit.name != "DestructionCollider")
                        targetList.AddExclusive(new GUItargetsListItem("" + hit.GetInstanceID(),
                                                                    TargetType.Object,
                                                                    hit.name,
                                                                    hit,
                                                                    0, 0));
            }
        }

        private void Content_Blue_Toggles() {
            GUILayout.BeginHorizontal(Util.Space_Compact);
            remote_toggleRemoteControl.set = GUILayout.Toggle(remote_toggleRemoteControl, remote_toggleRemoteControl ? "Connected" : "Connect");
            if (remote_targetVessel != null && remote_targetVessel.rootPart != null)
                remote_toggleRemoteControlLOCK = GUILayout.Toggle(remote_toggleRemoteControlLOCK, remote_toggleRemoteControlLOCK ? "Locked" : "Lock");
            GUILayout.EndHorizontal();
        }

        private void Content_Blue() {
            if (Target != null && Target is Vessel && Target != FlightGlobals.ActiveVessel) {
                if (remote_targetVessel != null && remote_targetVessel.rootPart == null) { //sysVessel target , locktarget is not loaded
                    GUI.enabled = false;
                    GUILayout.Toggle(false, "No target");
                    GUI.enabled = true;
                    GUILayout.Space(5);
                    GUILayout.Label("Note: Targeted craft is too far away to be remote controlled.");
                    remote_toggleRemoteControl.set = false;
                    remote_toggleRemoteControlLOCK = false;
                } else //all good ... sysVessel target , locktarget loaded
                    Content_Blue_Toggles();
            } else {
                if (!remote_toggleRemoteControlLOCK) { //no sysVessel target, and no lock
                    GUI.enabled = false;
                    GUILayout.Toggle(false, "No target");
                    GUI.enabled = true;
                    remote_toggleRemoteControl.set = false;
                } else //no sysVessel target, but lock
                    Content_Blue_Toggles();
            }

            if (remote_toggleRemoteControl && remote_targetVessel != null) {
				GUILayout.Space(5);
				remote_useControlSurface = GUILayout.Toggle(remote_useControlSurface, "Use control surfaces");
                GUILayout.Space(5);
                bool pm = FlightInputHandler.fetch.precisionMode;

				GUILayout.BeginHorizontal(Util.Space_Compact);
                GUILayout.FlexibleSpace();

				GUILayout.BeginVertical(Util.Space_Compact);
                GUILayout.FlexibleSpace();
				GUILayout.BeginHorizontal(Util.Space_Compact);
				remote_rolPlus = GUILayout.RepeatButton("Ͼ", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				GUILayout.Space(2);
				remote_yawPlus = GUILayout.RepeatButton("◄", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				GUILayout.EndHorizontal();
				GUILayout.EndVertical();

				GUILayout.BeginVertical(Util.Space_Compact);
				remote_pitPlus = GUILayout.RepeatButton("▲", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				remote_pitMinus = GUILayout.RepeatButton("▼", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
                GUILayout.EndVertical();

				GUILayout.BeginVertical(Util.Space_Compact);
				GUILayout.FlexibleSpace();
				GUILayout.BeginHorizontal(Util.Space_Compact);
				remote_yawMinus = GUILayout.RepeatButton("►", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				GUILayout.Space(2);
				remote_rolMinus = GUILayout.RepeatButton("Ͽ", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				GUILayout.EndHorizontal();
				GUILayout.EndVertical();

                GUILayout.FlexibleSpace();
                GUILayout.EndHorizontal();

                //Activate next stage
                if (GUILayout.Button("Next Stage (" + remote_targetVessel.currentStage + ")")) {
                    if (remote_targetVessel.currentStage > 0) {
                        foreach (Part p in remote_targetVessel.parts) {
                            if (p.inverseStage == remote_targetVessel.currentStage - 1) {
                                if (p is Decoupler || p is RadialDecoupler)  //if decoupler is activated collection changes...so q them up
                                    remote_stageQueue.Add(p);
                                else
                                    p.activate(p.inverseStage, p.vessel);
                            }
                        }
                        foreach (Part p in remote_stageQueue) //activate decouplers
							p.activate(p.inverseStage, p.vessel);
                        remote_stageQueue.Clear();
                        if (remote_targetVessel.currentStage > 1)
                            remote_targetVessel.currentStage -= 1;
                    } else {
                        foreach (Part p in remote_targetVessel.parts)
							p.activate(p.inverseStage, p.vessel);
                    }
                }
                GUILayout.Label("Orbital speed: " + remote_targetVessel.orbit.GetVel().magnitude.ToString("f2"));
                GUILayout.Label("Throttle: " + ((int)(remote_trottle * 100)).ToString("d3") + "%");
                remote_trottle = GUILayout.HorizontalSlider(remote_trottle, 0, 1);

                //p,r,y
                remote_toggleSAS.set = GUILayout.Toggle(remote_toggleSAS, "Stabilize (RSAS)");
                if (remote_toggleSAS.In()) {
                    remote_togglePrograde.set = false;
                    remote_toggleRetrograde.set = false;
                    remote_toggleProGee.set = false;
                    remote_toggleRetroGee.set = false;
                    remote_toggleFollow.set = false;
                } else if (remote_toggleSAS.Out()) { }
                if (remote_toggleSAS) {
                    remote_torque.x = -remote_targetVessel.angularMomentum.x;
                    remote_torque.y = -remote_targetVessel.angularMomentum.y;
                    remote_torque.z = -remote_targetVessel.angularMomentum.z;
                }
                GUILayout.BeginHorizontal();
                remote_togglePrograde.set = GUILayout.Toggle(remote_togglePrograde, "Prograde");
                if (remote_togglePrograde)
                    remote_dirSAS = remote_targetVessel.mainBody.getRFrmVel(remote_targetVessel.transform.position).normalized;
                if (remote_togglePrograde.In()) {
                    remote_toggleSAS.set = false;
                    remote_toggleRetrograde.set = false;
                    remote_toggleProGee.set = false;
                    remote_toggleRetroGee.set = false;
                    remote_toggleFollow.set = false;
                } else if (remote_togglePrograde.Out()) { }
                remote_toggleRetrograde.set = GUILayout.Toggle(remote_toggleRetrograde, "Retrograde");
                if (remote_toggleRetrograde)
                    remote_dirSAS = remote_targetVessel.mainBody.getRFrmVel(remote_targetVessel.transform.position).normalized * -1;
                if (remote_toggleRetrograde.In()) {
                    remote_toggleSAS.set = false;
                    remote_togglePrograde.set = false;
                    remote_toggleProGee.set = false;
                    remote_toggleRetroGee.set = false;
                    remote_toggleFollow.set = false;
                } else if (remote_toggleRetrograde.Out()) { }
                GUILayout.EndHorizontal();

                GUILayout.BeginHorizontal();
                remote_toggleProGee.set = GUILayout.Toggle(remote_toggleProGee, "ProGee");
                if (remote_toggleProGee)
                    remote_dirSAS = (remote_targetVessel.mainBody.position - remote_targetVessel.transform.position).normalized;
                if (remote_toggleProGee.In()) {
                    remote_toggleSAS.set = false;
                    remote_togglePrograde.set = false;
                    remote_toggleRetrograde.set = false;
                    remote_toggleRetroGee.set = false;
                    remote_toggleFollow.set = false;
                } else if (remote_toggleProGee.Out()) { }
                remote_toggleRetroGee.set = GUILayout.Toggle(remote_toggleRetroGee, "RetroGee");
                if (remote_toggleRetroGee)
                    remote_dirSAS = (remote_targetVessel.transform.position - remote_targetVessel.mainBody.position).normalized;
                if (remote_toggleRetroGee.In()) {
                    remote_toggleSAS.set = false;
                    remote_togglePrograde.set = false;
                    remote_toggleRetrograde.set = false;
                    remote_toggleProGee.set = false;
                    remote_toggleFollow.set = false;
                } else if (remote_toggleRetroGee.Out()) { }
                GUILayout.EndHorizontal();

                remote_toggleFollow.set = GUILayout.Toggle(remote_toggleFollow, "Follow");
                if (remote_toggleFollow)
                    remote_dirSAS = (vessel.transform.position - remote_targetVessel.transform.position).normalized;
                if (remote_toggleFollow.In()) {
                    remote_toggleSAS.set = false;
                    remote_togglePrograde.set = false;
                    remote_toggleRetrograde.set = false;
                    remote_toggleProGee.set = false;
                    remote_toggleRetroGee.set = false;
                } else if (remote_toggleFollow.Out()) { }
            } //end toggleRemoteControl
        }

        private void RemoteControl(FlightCtrlState s) {
			s.mainThrottle = remote_trottle;
			if(remote_useControlSurface) {
				s.pitch = remote_pitPlus ? -1 : remote_pitMinus ? 1 : 0;
				s.yaw = remote_yawPlus ? -1 : remote_yawMinus ? 1 : 0;
				s.roll = remote_rolPlus ? -0.5f : remote_rolMinus ? 0.5f : 0;
			}
        }

		private void RemoteControlFixedUpdate() {
			//Remote control
			if (remote_toggleRemoteControl && remote_targetVessel != null) {
				if(!remote_useControlSurface){
					if (remote_yawPlus)
						remote_targetVessel.rootPart.rigidbody.AddRelativeTorque(0, 0, remote_maxTorque);
					if (remote_yawMinus)
						remote_targetVessel.rootPart.rigidbody.AddRelativeTorque(0, 0, -remote_maxTorque);
					if (remote_pitPlus)
						remote_targetVessel.rootPart.rigidbody.AddRelativeTorque(remote_maxTorque, 0, 0);
					if (remote_pitMinus)
						remote_targetVessel.rootPart.rigidbody.AddRelativeTorque(-remote_maxTorque, 0, 0);
					if (remote_rolPlus)
						remote_targetVessel.rootPart.rigidbody.AddRelativeTorque(0, remote_maxTorque, 0);
					if (remote_rolMinus)
						remote_targetVessel.rootPart.rigidbody.AddRelativeTorque(0, -remote_maxTorque, 0);
				}
				if (remote_toggleSAS)
					remote_targetVessel.rootPart.rigidbody.AddRelativeTorque(remote_torque);
				else if (remote_togglePrograde || remote_toggleRetrograde || remote_toggleProGee || remote_toggleRetroGee || remote_toggleFollow){
					remote_targetVessel.ReferenceTransform.rigidbody.AddRelativeTorque(Controllers.StopSpin(remote_targetVessel.ReferenceTransform, remote_dirSAS, remote_targetVessel.angularVelocity, remote_targetVessel.angularMomentum, remote_targetVessel.MOI, remote_maxTorque, remote_maxTorqueF));
				
				
				}else
					remote_torque = Vector3.zero;
			}
			if (remote_toggleRemoteControl.In()) {
				remote_targetVessel = (Vessel)Target;
				LazorGlobals.SetPackingDistance(remote_targetVessel, float.MaxValue, float.MaxValue, float.MaxValue, float.MaxValue);
				remote_targetVessel.GoOffRails();
				remote_maxTorque = 4.5f;
				remote_maxTorqueF = 3.7f;
				/*if (remote_targetVessel.rootPart != null && remote_targetVessel.rootPart is CommandPod) {
					remote_maxTorque = ((CommandPod)remote_targetVessel.rootPart).rotPower;
					remote_maxTorqueF = ((CommandPod)vessel.rootPart).rotPower - ((CommandPod)vessel.rootPart).rotPower * (1 / 9);
				}*/
				remote_targetVessel.OnFlyByWire += new FlightInputCallback(RemoteControl);
			}
			if (remote_toggleRemoteControl.Out()) {
				TurnOffRemoteControl();
			}
		}

		public void TurnOffRemoteControl() {
			if(remote_targetVessel != null) {
				LazorGlobals.SetPackingDistance(remote_targetVessel);
				remote_targetVessel.OnFlyByWire -= new FlightInputCallback(RemoteControl);
			}
			remote_toggleRemoteControl.set = false;
			remote_toggleRemoteControlLOCK = false;
			remote_toggleSAS.set = false;
			remote_togglePrograde.set = false;
			remote_toggleRetrograde.set = false;
			remote_toggleRetroGee.set = false;
			remote_toggleProGee.set = false;
			remote_toggleFollow.set = false;
			remote_targetVessel = null;
		}

		public static void TurnOffRemoteControlForVessel(Vessel vessel){
			LazorSystemController controller = LazorSystemControllerGlobal.GetController(vessel);
			if(controller != null)
				controller.ActiveLazorSystem.TurnOffRemoteControl();
		}

        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        // UPDATE FUNCTIONS
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************

		#region update functions 

        /*protected override void onEditorUpdate() {
            if (!iconSetEditor) {
                try {
                    base.stackIcon.SetIcon(this, "icons.png", 0, 0);
                } catch (Exception) { }
                iconSetEditor = true;
            }
        }*/

        override protected void onPartFixedUpdate() {
            if (!initiated)
                return;
            if (lazorOn) {
				Leader.SetRange(lazorRange);
				Leader.onUpdate();
				ExecuteEnergyDrain();
				ExecuteFuelDrain();
                //Generate energy if LazorPowered is enabled for this.
				if(LazorGlobals.GLOBALCONFIG_ElectricGenerator){
	                double energyGen = 1.0d * Time.deltaTime * TimeWarp.CurrentRate;
	                double diff = Resources["ElectricCharge"].maxAmount  - Resources["ElectricCharge"].amount;
	                if(diff < energyGen)
	                    vessel.rootPart.RequestResource("ElectricCharge", -energyGen);
	                else 
	                    TransferResource(Resources["ElectricCharge"].info.id, energyGen);
				}

				if (IsActiveLazorSystem()) {
                    /*if (!iconSet) {
                        try {
                            base.stackIcon.SetIcon(this, "icons.png", 0, 0);
                        } catch (Exception) { }
                        foreach (Part p in vessel.parts) {
                            if (p is IModule)
                                (p as IModule).ActivateIcon();
                        }
                        iconSet = true;
                    }*/
                    guiOverlays.OnFixedUpdate();

                    //Mouse Look
                    if (!MapView.MapIsEnabled) { //mouse cursors are not available on the map
						if (toggleFreeAim && !FlightDriver.Pause) { //Handle mouse input in free aim mode
							Screen.showCursor = false;
                            Target = null;
                            if ((leader.GetColor() == Color.white || leader.GetColor() == Color.blue) && Input.GetMouseButtonDown(0)) {
                                toggleFreeAim = false;
                                Target = leader.GetHitVesselOrCollider();
                                AddToTargetList(leader);
                            }else if (leader.GetColor() == Color.red)
                                guiModules.ModuleTriggerOn(sunbeamLock && Input.GetMouseButton(0)); //activate sunbeam
                            if (leader.GetColor() == Color.green && Input.GetMouseButton(0)) {
                                AddToTargetList(leader);
								MouseLook(25); //when scanning, produce a scattering effect on lazor
                            } else
                                MouseLook();
                        }else
							Screen.showCursor = true;
					}else
						toggleFreeAim = false;

                    //Make lazor follow KSP targets
					TargetCheck_TargetSwitch();
					/*if (LazorGlobals.GLOBALCONFIG_SyncTargets) {
                        if (LazorGlobals.LastKSPtarget != FlightGlobals.fetch.VesselTarget) {
                            if (FlightGlobals.fetch.VesselTarget is Vessel)
                                Target = FlightGlobals.fetch.VesselTarget as Vessel;
                            else if (FlightGlobals.fetch.VesselTarget is CelestialBody)
                                Target = FlightGlobals.fetch.VesselTarget as CelestialBody;
                            else if (FlightGlobals.fetch.VesselTarget is ModuleDockingNode)
                                Target = FlightGlobals.fetch.VesselTarget as ModuleDockingNode;
                            LazorGlobals.LastKSPtarget = FlightGlobals.fetch.VesselTarget;
                        }
                        if (lastTarget != Target) {
                            if (Target is Vessel)
                                FlightGlobals.fetch.SetVesselTarget(Target as Vessel);
							else if (Target is CelestialBody && (Target as CelestialBody) != vessel.mainBody)
                                FlightGlobals.fetch.SetVesselTarget(Target as CelestialBody);
                            else if (Target == null)
                                FlightGlobals.fetch.SetVesselTarget(null);
                            lastTarget = Target;
                        }
                    }*/
                    VI.Update();
                    if (Target != null && VIT != null && Target is Vessel && ((Vessel)Target).rootPart != null)
                        VIT.Update();

					guiAchievements.OnFixedUpdate();
                    guiModules.OnFixedUpdate();
					RemoteControlFixedUpdate();
                } else { 
                    //iconSet = false;
                    //Teleport
                    if (toggleTeleport && FlightGlobals.ActiveVessel.rootPart.Modules.Contains("LazorEVA")) {
                        BaseEventData data = new BaseEventData(BaseEventData.Sender.USER);
                        data.Set<Vector3>("destination", surfacePos);
                        FlightGlobals.ActiveVessel.rootPart.Modules.Send("Teleport", data);
                        toggleTeleport = false;
                    }
                    //Teleport short
                    if (toggleTeleportShort && FlightGlobals.ActiveVessel.rootPart.Modules.Contains("LazorEVA")) {
                        BaseEventData data = new BaseEventData(BaseEventData.Sender.USER);
                        data.Set<Vector3>("destination", surfacePos);
                        FlightGlobals.ActiveVessel.rootPart.Modules.Send("TeleportShort", data);
                        toggleTeleportShort = false;
                    }
                }

				//Set leader aim in non-mouse-look mode.....
				LeaderAim();
            
                //Bearing, Mark, Roll.........................................................................................................................................................
                vectorEast = vessel.mainBody.getRFrmVel(vessel.transform.position).normalized; //uses the rotation of the body's frame to determine "east"
                vectorNorth = Vector3d.Cross(vectorEast, vessel.upAxis); //north = east cross up
                orientationVessel = Quaternion.Inverse(Quaternion.Euler(90, 0, 0) * Quaternion.Inverse(vessel.transform.rotation) * Quaternion.LookRotation(vectorNorth, vessel.upAxis));
				orientationLeader = Quaternion.Inverse(Quaternion.Euler(90, 0, 0) * Quaternion.Inverse(leader.GetTransform().rotation) * Quaternion.LookRotation(vectorNorth, vessel.upAxis));
				actualBearing = orientationVessel.eulerAngles.y;
                actualMark = orientationVessel.eulerAngles.x > 180 ? 360 - orientationVessel.eulerAngles.x : -orientationVessel.eulerAngles.x;
                actualRoll = -(orientationVessel.eulerAngles.z > 180 ? orientationVessel.eulerAngles.z - 360 : orientationVessel.eulerAngles.z);
                //other stuff
                LONGITUDE = vessel.mainBody.GetLongitude(vessel.transform.position);
                LATITUDE = vessel.mainBody.GetLatitude(vessel.transform.position);
                nonRotatingFrameVelocity = vessel.orbit.GetVel(); //orbit velocity
                rotatingFrameVelocity = nonRotatingFrameVelocity - vessel.mainBody.getRFrmVel(vessel.transform.position); //velocity relative to SoI
                
				config.OnFixedUpdate();

                //FLARE
                if (flareTransition)
                    FlareStartup();
            }
        }

        protected override void onPartUpdate() {
			//Lock down system when no controller is present: happends when lazors are disconnected awaiting reset
			if(sysLock_WaitDisconnect) {
				RuntimeLogic.Instance.Log("(f" + Time.frameCount + ")" + this + " : onPartUpdate() : unlocking sysLock_WaitDisconnect", true);
				lazorSystemController.DisableSystemLock((sys)=>{return sys.sysLock_WaitDisconnect;}, (sys)=>{sys.OnStageChange(sys.VI); sys.config.Reconfigure();});
				return;
			}
			if (sysLock_WaitCouple.Locked_bool) {
				if (sysLock_WaitCouple.Locked_frame){
					RuntimeLogic.Instance.Log ("(f" + Time.frameCount + ")" + this + " : onPartUpdate() : unlocking sysLock_WaitCouple", true);
					lazorSystemController.DisableSystemLock((sys)=>{return sys.sysLock_WaitCouple;}, (sys)=>{sys.OnStageChange(sys.VI); sys.ResetState();});
				}
				return;
			}
	
            if (lazorOn.In()) {
				if (IsActiveLazorSystem())
                    lazorSound.Play(SoundType.STARTUP);
                FlareStartup();
                leader.On();
            } else if (lazorOn.Out()) {
				if (IsActiveLazorSystem())
                    lazorSound.Play(SoundType.SHUTDOWN);
                FlareOff();
                leader.Off();
				config.ShutDown();
                guiOverlays.Off();
				TurnOffRemoteControl();
            }

			if (IsActiveLazorSystem()) {
                LazorSystemControllerGlobal.LastVesselWithLazor = FlightGlobals.ActiveVessel;
                //determine SoI change
                if (vessel.mainBody != null && currentBody != vessel.mainBody) {
					localEvents.RegisterSoIChange(currentBody == null ? "none" : currentBody.name, vessel.mainBody == null ? "none" : vessel.mainBody.name);
                    currentBody = vessel.mainBody;
                }
				//determine situation change
				if (vessel.situation != lastSituation){
					localEvents.RegisterSituationChange(lastSituation.ToString(), vessel.situation.ToString());
					lastSituation = vessel.situation;
				}
            }
        }
		
		#endregion

        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        // METHODS
        //*******************************************************************************************************************
        //*******************************************************************************************************************
        //*******************************************************************************************************************

		#region methods

        // PRIVATE
        private void FlareStartup() {
            if (!flareTransition) {
                flareTimeOn = Time.time;
                flare_emitter.minEnergy = 0.1F;
                flare_emitter.maxEnergy = 0.5F;
                flare_light.enabled = true;
            }
            flareTransition = true;
            flare_emitter.minSize = 0.1f * ((Time.time - flareTimeOn) / 3);
            flare_emitter.maxSize = 0.3f * ((Time.time - flareTimeOn) / 3);
            flare_light.range = 3f * ((Time.time - flareTimeOn) / 3);
            if (Time.time - flareTimeOn > 3) {
                flareTransition = false;
                flare_emitter.minSize = 0.1f;
                flare_emitter.maxSize = 0.3f;
                flare_light.range = 3f;
            }
        }

        private void FlareOff() {
            flare_emitter.minEnergy = 0F;
            flare_emitter.maxEnergy = 0F;
            flare_emitter.minSize = 0F;
            flare_emitter.maxSize = 0F;
            flare_light.enabled = false;
            flareTransition = false;
        }
        
        //LEADER ONLY
        public void MouseLook(float offset = 0) {
			Vector3 endPos = Util.MouseLook(mousePos, sysLock_LEADER_ignorehits ? 0 : 557059);
			Vector3 dir = endPos - Leader.GetOrigin();
			//MonoBehaviour.print (dir);
			if (offset != 0) {
				dir = dir.normalized;
				dir = dir + UnityEngine.Random.insideUnitSphere * UnityEngine.Random.Range (-offset/360, offset/360);
			}
			LeaderRevolver(endPos);
			Leader.Aim(dir);
        }

		private void LeaderAim() {
			if(!toggleFreeAim) {
				if(leaderLocked) {
					if(Target != null) {
						LeaderRevolver(Target.transform.position);
						leader.Aim(Target.transform.position - leader.GetOrigin());
					}
				} else {
					target = null; //TO AVOID THE Target.Set stuff form triggering.....
					LeaderRevolver(Leader.GetAim() * 1000 + Leader.GetOrigin());
				}
			}
		}

		//Rotate/revolve the lazor turret
		private void LeaderRevolver(Vector3 aimPos){
			if(systemIsStacked) {
				Vector3 dir = aimPos - transform.position;
				Util.Revolve2(dir, lazorBase, transform); //rotate lazor turret
				leader.SetOrigin(lazorCrystal.transform.position);
			} else
				Util.RotateAroundAxis(Leader.GetAim(), Vector3.up, lazorBase, transform);
		}

        public Texture2D GetLeaderTextureColor() {
            if (leader.GetColor() == Color.red)
                return textureRed;
            else if (leader.GetColor() == Color.blue)
                return textureBlue;
            else if (leader.GetColor() == Color.green)
                return textureGreen;
            else if (leader.GetColor() == Color.white)
                return textureWhite;
            return textureWhite;
        }


        //PUBLIC


		public void TargetCheck_TargetSwitch() {
			//Make lazor follow KSP targets
			if (LazorGlobals.GLOBALCONFIG_SyncTargets) {
				if (LazorGlobals.LastKSPtarget != FlightGlobals.fetch.VesselTarget)
					TargetEnforce_KSPToSystem();
				if (lastTarget != Target)
					TargetEnforce_SystemToKSP();
			}
		}

		public void TargetCheck_SystemSwitch() {
			//There is likely to be a conflict here, prioritize lazor target if it has a target, if not, prioritize global target
			if (LazorGlobals.GLOBALCONFIG_SyncTargets) {
				if(Target == null){
					if(FlightGlobals.fetch.VesselTarget != null)
						TargetEnforce_KSPToSystem();
				}else
					TargetEnforce_SystemToKSP();
			}
		}

		private void TargetEnforce_SystemToKSP() {
			if (Target is Vessel)
				FlightGlobals.fetch.SetVesselTarget(Target as Vessel);
			else if (Target is CelestialBody && (Target as CelestialBody) != vessel.mainBody)
				FlightGlobals.fetch.SetVesselTarget(Target as CelestialBody);
			else if (Target == null)
				FlightGlobals.fetch.SetVesselTarget(null);
			lastTarget = Target;
		}

		private void TargetEnforce_KSPToSystem() {
			if(FlightGlobals.fetch.VesselTarget is Vessel)
				Target = FlightGlobals.fetch.VesselTarget as Vessel;
			else if(FlightGlobals.fetch.VesselTarget is CelestialBody)
				Target = FlightGlobals.fetch.VesselTarget as CelestialBody;
			else if(FlightGlobals.fetch.VesselTarget is ModuleDockingNode)
				Target = FlightGlobals.fetch.VesselTarget as ModuleDockingNode;
			LazorGlobals.LastKSPtarget = FlightGlobals.fetch.VesselTarget;
		}

		#region resource drain

		private double fuelDrain;
		public double RequestFuelDelta(double amount) {
			if(!CheatOptions.InfiniteFuel)
				fuelDrain += amount * Time.deltaTime * TimeWarp.CurrentRate;
			return amount;
		}
		private void ExecuteFuelDrain() {
			RequestResource(LazorGlobals.LiquidFuelID, fuelDrain, ResourceFlowMode.STAGE_PRIORITY_FLOW);
			fuelDrain = 0;
		}

        //systems requiring energy, amount adjusted by deltatime
		private double energyDrain;
        public double RequestEnergyDelta(double amount) {
			energyDrain += amount * Time.deltaTime * TimeWarp.CurrentRate;
			return amount;
            //return RequestResource("ElectricCharge", amount * Time.deltaTime * TimeWarp.CurrentRate);
        }
		private void ExecuteEnergyDrain() {
			RequestResource(LazorGlobals.ElectricChargeID, energyDrain, ResourceFlowMode.ALL_VESSEL);
			energyDrain = 0;
		}

        //get a flat non-adjusted amount of energy
        public void RequestEnergy(double amount) {
			print(this + ": Scanner drained : "+VI.DrainResource("ElectricCharge", (float)amount)+ " units of electric charge");
        }

		#endregion

        public void AddToTargetList(Lazor lazor) {
            if (lazor.LazorIsHittingSomething && !lazor.LazorIsHittingOwnVessel) {
                if (lazor.LazorIsHittingVessel) {
                    Vessel v = lazor.GetHitVesselOrCollider() as Vessel;
                    targetList.AddReplacement(new GUItargetsListItem("" + v.id, TargetType.Vessel, v.vesselName, v, 0, 0));
                } else {
                    if (System.Text.RegularExpressions.Regex.IsMatch(lazor.GetHit().collider.name, REGEX_surfaceCollider))
                        return;
                    targetList.AddExclusive(new GUItargetsListItem("" + lazor.GetHit().collider.GetInstanceID(),
                                                                   TargetType.Object,
                                                                   lazor.GetHit().collider.name,
                                                                   lazor.GetHit().collider,
                                                                   0, 0));
                }
            }
        }

        public void TrackTarget(Component c) {
            targetsTracked.Add(c);
        }

        public void UnTrackTarget(Component c) {
            targetsTracked.Remove(c);
        }
		
        public string TargetName() {
            if (Target == null) {
                return "nothing";
            } else {
                if (Target is Vessel)
                    return (Target as Vessel).vesselName;
                else
                    return Target.name;
            }
        }

        public float TargetDistance() {
            if (Target != null)
                return (Target.transform.position - transform.position).magnitude;
            return float.NaN;
        }

		//FIX: Local is missleading, it is really world space
        public Vector3 LocalUp {
            get {
                if (upAxis == Axis.hUp)
                    return -vessel.transform.forward;
                else
                    return vessel.transform.up;
            }
        }

        public Vector3 LocalForward {
            get {
                if (upAxis == Axis.hUp)
                    return vessel.transform.up;
                else
                    return vessel.transform.forward;
            }
        }

        public bool IsActiveVessel(Vessel v) {
            if (!FlightGlobals.ActiveVessel.isEVA) {
                if (v != FlightGlobals.ActiveVessel)
                    return false;
            } else {
                if (FlightGlobals.ActiveVessel.rootPart.Modules.Contains("LazorEVA")) {
                    string id = FlightGlobals.ActiveVessel.rootPart.Modules["LazorEVA"].Fields.GetValue<string>("myVessel");
                    Guid guid = new Guid(id);
					return guid == v.id && (FlightGlobals.ActiveVessel.transform.position - v.transform.position).magnitude < LazorGlobals.TELEPORT_DISTANCE;
                } else
                    return false;
            }
            return true;
        }

		public bool IsActiveLazorSystem() {
			if(lazorSystemController == null) {
				MonoBehaviour.print("(f" + Time.frameCount + ")" + this + " : IsActiveLazorSystem() : system locked down");
				return false;
			}
			return lazorSystemController.IsActiveLazorSystem(this);
		}

        public void SetLeaderColor(Color c, bool forceFreeAimOff = true) {
            if (lazorOn) {
                leader.SetColor(c);
                flare_emitterColors[0] = c;
                flare_emitterColors[1] = c;
                flare_emitterAnimator.colorAnimation = flare_emitterColors;
                flare_light.color = c;
                lazorSound.Play(SoundType.SWITCH);
                if (c == Color.red)
                    Util.SetMaterial(lazorCrystal.gameObject, crystal_red);
                else if (c == Color.green)
                    Util.SetMaterial(lazorCrystal.gameObject, crystal_green);
                else if (c == Color.blue)
                    Util.SetMaterial(lazorCrystal.gameObject, crystal_blue);
                else if (c == Color.white)
                    Util.SetMaterial(lazorCrystal.gameObject, crystal_white);
                if (forceFreeAimOff)
                    sunbeamLock.set = false;
            }
        }

        //fix
        private DoubleSwitch sunbeamLock;
        public bool SunbeamLock{
            set {
                sunbeamLock.set = value;
                if (sunbeamLock.In() || sunbeamLock.Out()) {
                    toggleFreeAim = value;
                    SetLeaderColor(Color.red, false);
                }
            }
            get{return sunbeamLock;}
        }
		
        public double HeightFromTerrain(){
            return vessel.GetHeightFromTerrain() == -1 ? vessel.orbit.altitude : vessel.GetHeightFromTerrain();
        }

		override public string ToString(){
			return "LazorSystem(" + uid.ToString() + ")";
		}

		#endregion

    }











}

