/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {
    abstract class MVS { //multi vector support

        protected Lazor leader = null;
        protected MovementControlExtended mc;
		public List<LazorAimable> parts = null;
        public LazorSystem system;


        public MVS(LazorSystem system, List<LazorAimable> parts) {
            this.system = system;
            this.leader = system.Leader;
            this.parts = parts;
            this.mc = system.mc;
        }

        abstract public void Activate(); //called when this scheme is activated
        abstract public void MoveUp();
        abstract public void MoveDown();
        abstract public void MoveLeft();
        abstract public void MoveRight();
    }

	class MVSPanel{
		private MVS mvs; //movement controller for claimed lazors
		public bool mode_chase, mode_lock, mode_disperse;
		private DoubleSwitch safeMode = new DoubleSwitch(true, true);

		public MVSPanel(MVS mvs) {
			this.mvs = mvs;
		}

		public void MoveWithLeader(){
			if(mvs.system.toggleFreeAim && !FlightDriver.Pause && (mode_chase || mode_lock))
				mvs.Activate();
		}

		public void GUIPanel(){
			#region movement buttons
			
			GUILayout.BeginHorizontal(Util.Divider_directions_BG);
			GUILayout.BeginHorizontal(Util.Space_Compact);
			GUILayout.BeginVertical(Util.Divider_directions_arrowBG, GUILayout.Width(76));
			GUILayout.BeginHorizontal();
			GUILayout.FlexibleSpace();
			if (GUILayout.RepeatButton("▲", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22)))
				mvs.MoveUp();
			GUILayout.FlexibleSpace();
			GUILayout.EndHorizontal();
			
			GUILayout.BeginHorizontal();
			GUILayout.FlexibleSpace();
			if (GUILayout.RepeatButton("◄", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22)))
				mvs.MoveLeft();
			mvs.system.toggleFreeAim = GUILayout.Toggle(mvs.system.toggleFreeAim, "⊙", Util.Buttons_arrows_mainPanel, GUILayout.Width(22)); //Mouse\nLook
			if (GUILayout.RepeatButton("►", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22)))
				mvs.MoveRight();
			GUILayout.FlexibleSpace();
			GUILayout.EndHorizontal();
			
			GUILayout.BeginHorizontal();
			GUILayout.FlexibleSpace();
			if (GUILayout.RepeatButton("▼", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22)))
				mvs.MoveDown();
			GUILayout.FlexibleSpace();
			GUILayout.EndHorizontal();
			GUILayout.EndVertical();
			GUILayout.EndHorizontal();
			
			GUILayout.BeginHorizontal(Util.Space_Compact);
			GUILayout.Space(1);
			GUILayout.BeginVertical();
			GUILayout.Space(10);
			
			#region l o a safe mode d c l reset

			GUILayout.BeginHorizontal();
			//safe mode
			safeMode.set = GUILayout.Toggle(safeMode, "Safe Mode", Util.Button_Toggle, GUILayout.Width(64));
			if(safeMode.In())
				foreach (LazorAimable l in mvs.parts)
					l.Lazor.safeMode = true;
			else if(safeMode.Out())
				foreach (LazorAimable l in mvs.parts)
					l.Lazor.safeMode = false;
			/*if (GUILayout.Toggle(mvs.system.Leader.safeMode, "Safe Mode", Util.Button_Toggle, GUILayout.Width(64))) { //, GUILayout.Width(70
				mvs.system.Leader.safeMode = true;
				foreach (LazorAimable l in mvs.parts)
					l.Lazor.safeMode = true;
			} else {
				mvs.system.Leader.safeMode = false;
				foreach (LazorAimable l in mvs.parts)
					l.Lazor.safeMode = false;
			}*/
			if (GUILayout.Button("Reset", GUILayout.Width(44))) {
				mvs.system.leaderLocked = false;
				foreach (LazorAimable l in mvs.parts)
					l.Lazor.ResetOrientation();
			}
			GUILayout.EndHorizontal();
			
			GUILayout.BeginHorizontal();
			// MVS ........................................................................................
			mode_disperse = GUILayout.Toggle(mode_disperse, "Dispersal mode", Util.Button_Toggle, GUILayout.Width(108));
			if (mode_disperse && mvs.GetType().ToString() != typeof(MVSdisperse).ToString()) {     //Major pain in the behind! throws MethodNotFoundException on System.Type.op_Equality  if i just '==' compare the types.
				mvs = new MVSdisperse(mvs.system, mvs.parts);
				mvs.Activate();
				mode_chase = false;
				mode_lock = false;
			}
			GUILayout.EndHorizontal();
		
			GUILayout.BeginHorizontal();
			mode_chase = GUILayout.Toggle(mode_chase, "Chase", Util.Button_Toggle, GUILayout.Width(54));
			if (mode_chase && mvs.GetType().ToString() != typeof(MVSchase).ToString()) {
				mvs = new MVSchase(mvs.system, mvs.parts);
				mvs.Activate();
				mode_disperse = false;
				mode_lock = false;
			}
			mode_lock = GUILayout.Toggle(mode_lock, "Lock", Util.Button_Toggle, GUILayout.Width(54));
			if (mode_lock && mvs.GetType().ToString() != typeof(MVSlock).ToString()) {
				mvs = new MVSlock(mvs.system, mvs.parts);
				mvs.Activate();
				mode_disperse = false;
				mode_chase = false;
			}
			/*else
				mode_lock = false;*/
			if (!mode_chase && !mode_lock && !mode_disperse && mvs.GetType().ToString() != typeof(MVSnormal).ToString())
				mvs = new MVSnormal(mvs.system, mvs.parts);
			GUILayout.EndHorizontal();
			
			#endregion
		
			GUILayout.EndVertical();
			GUILayout.FlexibleSpace();
			GUILayout.EndHorizontal();
			
			/*Rect a = GUILayoutUtility.GetLastRect();
			Rect b = new Rect(a.x + 8, a.y + a.height - 25, 50, 20);
			Matrix4x4 matrix = GUI.matrix;
			GUIUtility.RotateAroundPivot(90, new Vector2(a.x + 70, a.y + a.height)); //clipping problem on edge of window. solution, rotate it into position.
			if (GUI.Button(b, "Reset")) {
				mvs.system.leaderLocked = false;
				//mvs.system.Leader.ResetOrientation();
				foreach (LazorAimable l in mvs.parts)
					l.Lazor.ResetOrientation();
			} GUI.matrix = matrix;*/
			
			GUILayout.EndHorizontal();
			
			#endregion
		}
	}
}

