﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {

    /* Lazor dispersal pattern: 
     * 1. Makes the lazors fire towards the direction of their anchoring point but preserve the "height"
     * 2. Synced lazor movement keeps the lazors firing in a dispersed pattern.
     * 3. Avoid mad-pole syndrom by gradually moving the lazors toward anchor as well as pole when lazor is close to pole.
     */
    class MVSdisperse : MVS {

        public MVSdisperse(LazorSystem system, List<LazorAimable> parts) : base(system, parts) { }

        override public void Activate() {
            /*if(system.moveBoth || system.moveLeaderOnly)
                if (leader != null)
                    leader.Aim(leader.GetTransformBase().up);*/
            //if (moveBoth || moveOthersOnly)
                foreach (LazorAimable l in parts) {
                    float f = Vector3.Angle(l.Lazor.GetAim(), system.LocalUp);
                    l.Lazor.SetEulerAngles(l.transform.eulerAngles);
                    l.Lazor.SetLocalEulerAngles(new Vector3((l.Lazor.GetTransform().localEulerAngles.x - 90 + f) % 360, l.Lazor.GetTransform().localEulerAngles.y, l.Lazor.GetTransform().localEulerAngles.z));
                }
        }
        
        override public void MoveUp() {
            /*if (system.moveBoth || system.moveLeaderOnly)
                if (leader != null) mc.RotateUp(leader);*/
            //if (moveBoth || moveOthersOnly)
                foreach (LazorAimable l in parts) {
                    float f = Vector3.Angle(l.Lazor.GetAim(), system.LocalUp);
                    if (f > 170) {
                        Vector3 original = l.Lazor.GetTransform().eulerAngles;
                        l.Lazor.SetEulerAngles(l.transform.eulerAngles);
                        l.Lazor.SetLocalEulerAngles(new Vector3((l.Lazor.GetTransform().localEulerAngles.x - 100 + f) % 360, l.Lazor.GetTransform().localEulerAngles.y, l.Lazor.GetTransform().localEulerAngles.z));
                        Quaternion e = Quaternion.Euler(l.Lazor.GetTransform().eulerAngles);
                        l.Lazor.SetEulerAngles(original);
                        Vector3 v = e * Vector3.up; //quaternion as vector
                        mc.RotateTowards(l.Lazor, v + system.LocalUp);
                    } else if (f < 25 || f > 130) {
                        Vector3 original = l.Lazor.GetTransform().eulerAngles;
                        l.Lazor.SetEulerAngles(l.transform.eulerAngles);
                        l.Lazor.SetLocalEulerAngles(new Vector3((l.Lazor.GetTransform().localEulerAngles.x - 90 + f) % 360, l.Lazor.GetTransform().localEulerAngles.y, l.Lazor.GetTransform().localEulerAngles.z));
                        Quaternion e = Quaternion.Euler(l.Lazor.GetTransform().eulerAngles);
                        l.Lazor.SetEulerAngles(original);
                        Vector3 v = e * Vector3.up; //quaternion as vector
                        mc.RotateTowards(l.Lazor, v + system.LocalUp);
                    }  else {
                        mc.RotateUp(l.Lazor);
                    }
                }
        }

        override public void MoveDown() {
            /*if (system.moveBoth || system.moveLeaderOnly)
                if (leader != null) mc.RotateDown(leader);*/
            //if (moveBoth || moveOthersOnly)
                foreach (LazorAimable l in parts) {
                    float f = Vector3.Angle(l.Lazor.GetAim(), system.LocalUp);
                    if (f < 5) {
                        Vector3 original = l.Lazor.GetTransform().eulerAngles;
                        l.Lazor.SetEulerAngles(l.transform.eulerAngles);
                        l.Lazor.SetLocalEulerAngles(new Vector3((l.Lazor.GetTransform().localEulerAngles.x - 80 + f) % 360, l.Lazor.GetTransform().localEulerAngles.y, l.Lazor.GetTransform().localEulerAngles.z));
                        Quaternion e = Quaternion.Euler(l.Lazor.GetTransform().eulerAngles);
                        l.Lazor.SetEulerAngles(original);
                        Vector3 v = e * Vector3.up; //quaternion as vector
                        mc.RotateTowards(l.Lazor, v - system.LocalUp);
                    } else if (f < 25 || f > 130) {
                        Vector3 original = l.Lazor.GetTransform().eulerAngles;
                        l.Lazor.SetEulerAngles(l.transform.eulerAngles);
                        l.Lazor.SetLocalEulerAngles(new Vector3((l.Lazor.GetTransform().localEulerAngles.x - 90 + f) % 360, l.Lazor.GetTransform().localEulerAngles.y, l.Lazor.GetTransform().localEulerAngles.z));
                        Quaternion e = Quaternion.Euler(l.Lazor.GetTransform().eulerAngles);
                        l.Lazor.SetEulerAngles(original);
                        Vector3 v = e * Vector3.up; //quaternion as vector
                        mc.RotateTowards(l.Lazor, v - system.LocalUp);
                    } else {
                        mc.RotateDown(l.Lazor);
                    }
                }
        }
	
        override public void MoveLeft() {
            /*if (system.moveBoth || system.moveLeaderOnly)
                if (leader != null) mc.RotateLeft(leader);*/
            //if (moveBoth || moveOthersOnly)
                foreach (LazorAimable l in parts)
                    mc.RotateLeft(l.Lazor);
        }

        override public void MoveRight() {
            /*if (system.moveBoth || system.moveLeaderOnly)
                if (leader != null) mc.RotateRight(leader);*/
            //if (moveBoth || moveOthersOnly)
                foreach (LazorAimable l in parts)
                    mc.RotateRight(l.Lazor);
        }
     


    }
}

