﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {
    class MVSlock : MVS {

        public MVSlock(LazorSystem system, List<LazorAimable> parts) : base(system, parts) { }

        override public void Activate() {
            ChaseOrLock();
        }

        override public void MoveUp() {
            mc.RotateUp(leader);
            ChaseOrLock();
        }

        override public void MoveDown() {
            mc.RotateDown(leader);
            ChaseOrLock();
        }

        override public void MoveLeft() {
            mc.RotateLeft(leader);
            ChaseOrLock();
        }

        override public void MoveRight() {
            mc.RotateRight(leader);
            ChaseOrLock();
        }

        private void ChaseOrLock() {
            if(leader != null)
                if (leader.LazorIsHittingSomething) {
                    LockOnLeader();
                } else {
                    ChaseLeader();
                }
        }

        private void ChaseLeader() {
            foreach (LazorAimable l in parts)
                if(l != null)
                    l.Lazor.Aim(leader.GetAim());
        }

        private void LockOnLeader() {
            foreach (LazorAimable l in parts) { //make all lasers hit what the leader is hitting
                if (l != null) {
                    Vector3 a = leader.GetHitPoint();
                    Vector3 b = l.Lazor.GetOrigin();
                    l.Lazor.Aim((a - b).normalized);
                }
            }
        }
    }
}

