﻿/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace Romfarer {
	class MovementControl {

		//FIX: define speed of movement
		public MovementControl() {}
		
		private static float Td() {
			return FlightInputHandler.fetch.precisionMode ? 0.001f : 0.01f;
		}
		
		public void RotateRight(IAimable t, Vector3 normal) {
			t.Aim(RotateAround(t.GetAim(), Vector3.zero, normal, Td()));
		}
		
		public void RotateLeft(IAimable t, Vector3 normal) {
			t.Aim(RotateAround(t.GetAim(), Vector3.zero, normal, -Td()));
		}
		
		public void RotateUp(IAimable t, Vector3 normal) {
			t.Aim(Vector3.RotateTowards(t.GetAim(), normal, Td(), Td()));
		}
		
		public void RotateDown(IAimable t, Vector3 normal) {
			t.Aim(Vector3.RotateTowards(t.GetAim(), -normal, Td(), Td()));
		}

		public void RotateTowards(IAimable t, Vector3 v) {
			t.Aim(Vector3.RotateTowards(t.GetAim(), v, Td(), Td()));
		}

		public static Vector3 RotateAround(Vector3 point, Vector3 origin, Vector3 normal, float angle) {
			return new Vector3((origin.x * (Mathf.Pow(normal.y, 2) + Mathf.Pow(normal.z, 2)) - normal.x * (origin.y * normal.y + origin.z * normal.z - normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.x * Mathf.Cos(angle) + (-origin.z * normal.y + origin.y * normal.z - normal.z * point.y + normal.y * point.z) * Mathf.Sin(angle),
			                   (origin.y * (Mathf.Pow(normal.x, 2) + Mathf.Pow(normal.z, 2)) - normal.y * (origin.x * normal.x + origin.z * normal.z - normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.y * Mathf.Cos(angle) + (origin.z * normal.x - origin.x * normal.z + normal.z * point.x - normal.x * point.z) * Mathf.Sin(angle),
			                   (origin.z * (Mathf.Pow(normal.x, 2) + Mathf.Pow(normal.y, 2)) - normal.z * (origin.x * normal.x + origin.y * normal.y - normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.z * Mathf.Cos(angle) + (-origin.y * normal.x + origin.x * normal.y - normal.y * point.x + normal.x * point.y) * Mathf.Sin(angle));
		}

		public static Vector3 RotateAroundNormal(Vector3 point, Vector3 normal, bool right) {
			if(right)
				return RotateAroundNormal(point, normal, Td());
			else
				return RotateAroundNormal(point, normal, -Td());
		}

		public static Vector3 RotateAroundNormal(Vector3 point, Vector3 normal, float angle) {
			return new Vector3(( - normal.x * (- normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.x * Mathf.Cos(angle) + (- normal.z * point.y + normal.y * point.z) * Mathf.Sin(angle),
			                   (- normal.y * (- normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.y * Mathf.Cos(angle) + (+ normal.z * point.x - normal.x * point.z) * Mathf.Sin(angle),
			                   (- normal.z * (- normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.z * Mathf.Cos(angle) + (- normal.y * point.x + normal.x * point.y) * Mathf.Sin(angle));
		}
	}




	class MovementControlExtended : MovementControl{

        private LazorSystem system;

        public MovementControlExtended(LazorSystem system) {
            this.system = system;
        }
		/*
        private float Td() {
            return FlightInputHandler.fetch.precisionMode ? 0.001f : 0.01f;
        }*/

        public void RotateRight(IAimable t) {
            system.leaderLocked = false;
			base.RotateRight(t, system.LocalUp);
            //t.Aim(RotateAround(t.GetAim(), Vector3.zero, system.LocalUp, Td()));
        }

		public void RotateLeft(IAimable t) {
            system.leaderLocked = false;
			base.RotateLeft(t, system.LocalUp);
            //t.Aim(RotateAround(t.GetAim(), Vector3.zero, system.LocalUp, -Td()));
        }

		public void RotateUp(IAimable t) {
            system.leaderLocked = false;
			base.RotateTowards(t, system.LocalUp);
            //t.Aim(Vector3.RotateTowards(t.GetAim(), system.LocalUp, Td(), Td()));
        }

		public void RotateDown(IAimable t) {
            system.leaderLocked = false;
			base.RotateTowards(t, -system.LocalUp);
            //t.Aim(Vector3.RotateTowards(t.GetAim(), -system.LocalUp, Td(), Td()));
        }

        //Angle is angle between lazor and vessel.up
        public void SetAngle(IAimableLazor l, float angle){
            system.leaderLocked = false;
            if (system.useLocalCoordinates) {
                if (system.IsHorizontal) {
                    l.GetTransform().rotation = Quaternion.LookRotation(system.vessel.transform.up, system.LocalUp) * Quaternion.Euler(90, 0, 0) *
                                                Quaternion.Inverse(Quaternion.Euler(angle, 0, GetRotation(l)));
                } else {
                    float rot = GetRotation(l);
                    Vector3 original = l.GetTransform().eulerAngles;
                    l.SetEulerAngles(l.GetTransformBase().eulerAngles);
                    l.SetLocalEulerAngles(new Vector3(l.GetTransform().localEulerAngles.x, rot % 360, l.GetTransform().localEulerAngles.z));
                    l.SetLocalEulerAngles(new Vector3(angle % 360, l.GetTransform().localEulerAngles.y, l.GetTransform().localEulerAngles.z));
                }
            } else {
                l.GetTransform().rotation = Quaternion.LookRotation(system.vectorNorth, (system.vessel.rootPart.transform.position - system.vessel.mainBody.position).normalized) * Quaternion.Euler(90, 0, 0) *
                                            Quaternion.Inverse(Quaternion.Euler(angle,0, GetRotation(l)));
            }
        }

        public void SetRotation(IAimableLazor l, float rot) {
            system.leaderLocked = false;
            if (system.useLocalCoordinates) {
                if (system.IsHorizontal) {
                    l.GetTransform().rotation = Quaternion.LookRotation(system.vessel.transform.up, system.LocalUp) * Quaternion.Euler(90, 0, 0) *
                                                Quaternion.Inverse(Quaternion.Euler(GetAngle(l), 0, rot));
                } else {
                    float angle = Vector3.Angle(l.GetAim(), system.LocalUp);
                    l.SetEulerAngles(l.GetTransformBase().eulerAngles);
                    l.SetLocalEulerAngles(new Vector3(l.GetTransform().localEulerAngles.x, rot % 360, l.GetTransform().localEulerAngles.z));
                    l.SetLocalEulerAngles(new Vector3(angle % 360, l.GetTransform().localEulerAngles.y, l.GetTransform().localEulerAngles.z));
                }
            } else {
                l.GetTransform().rotation = Quaternion.LookRotation(system.vectorNorth, (system.vessel.rootPart.transform.position - system.vessel.mainBody.position).normalized) * Quaternion.Euler(90, 0, 0) *
                                            Quaternion.Inverse(Quaternion.Euler(GetAngle(l), 0, rot));
            }
        }

        public float GetAngle(IAimableLazor l) {
            if (system.useLocalCoordinates) {
                if (system.IsHorizontal)
                    return 90 - Vector3.Angle(system.Leader.GetAim(), system.LocalUp);
                else
                    return Vector3.Angle(system.Leader.GetAim(), system.LocalUp);
                
            }
            return system.orientationLeader.eulerAngles.x > 180 ? 360 - system.orientationLeader.eulerAngles.x : -system.orientationLeader.eulerAngles.x;
        }
        
        public float GetRotation(IAimableLazor l) {
            if (system.useLocalCoordinates) { //LOCAL  Vessel.forward is 0 rotation. vessel.right is 90 degree and vessel.left is 270
                /*if (system.IsHorizontal) {
                    return Quaternion.Inverse(Quaternion.Euler(90, 0, 0) * Quaternion.Inverse(system.GetLeader().GetTransform().rotation) * Quaternion.LookRotation(system.vessel.transform.up, system.LocalUp)).eulerAngles.y;
                } else {
                    Vector3 a = system.vessel.transform.InverseTransformDirection(l.GetAim());
                    float rot = Mathf.Atan2(a.y, a.z) * Mathf.Rad2Deg;
                    return rot < 0 ? 360 + rot : rot;
                }*/
                return Util.Angle(Vector3.Cross(system.LocalForward, system.LocalUp),
                                  Vector3.Cross(l.GetAim(), system.LocalUp),
                                  system.LocalUp);
                //return Util.Angle(system.LocalForward, system.vessel.transform.InverseTransformDirection(l.GetAim()), system.LocalUp);
            }
            //global
            return system.orientationLeader.eulerAngles.y;
        }

        /*public void RotateTowards(IAimable t, Vector3 v) {
            t.Aim(Vector3.RotateTowards(t.GetAim(), v, Td(), Td()));
        }*/

        //Matrix rotation from a textbook.
        /*public static Vector3 RotateAround(Vector3 point, Vector3 origin, Vector3 normal, float angle) {
            return new Vector3((origin.x * (Mathf.Pow(normal.y, 2) + Mathf.Pow(normal.z, 2)) - normal.x * (origin.y * normal.y + origin.z * normal.z - normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.x * Mathf.Cos(angle) + (-origin.z * normal.y + origin.y * normal.z - normal.z * point.y + normal.y * point.z) * Mathf.Sin(angle),
                               (origin.y * (Mathf.Pow(normal.x, 2) + Mathf.Pow(normal.z, 2)) - normal.y * (origin.x * normal.x + origin.z * normal.z - normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.y * Mathf.Cos(angle) + (origin.z * normal.x - origin.x * normal.z + normal.z * point.x - normal.x * point.z) * Mathf.Sin(angle),
                               (origin.z * (Mathf.Pow(normal.x, 2) + Mathf.Pow(normal.y, 2)) - normal.z * (origin.x * normal.x + origin.y * normal.y - normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.z * Mathf.Cos(angle) + (-origin.y * normal.x + origin.x * normal.y - normal.y * point.x + normal.x * point.y) * Mathf.Sin(angle));
        }*/


        /*public static Vector3 RotateAroundNormal(Vector3 point, Vector3 normal, float angle) {
            return new Vector3(( - normal.x * (- normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.x * Mathf.Cos(angle) + (- normal.z * point.y + normal.y * point.z) * Mathf.Sin(angle),
                               (- normal.y * (- normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.y * Mathf.Cos(angle) + (+ normal.z * point.x - normal.x * point.z) * Mathf.Sin(angle),
                               (- normal.z * (- normal.x * point.x - normal.y * point.y - normal.z * point.z)) * (1 - Mathf.Cos(angle)) + point.z * Mathf.Cos(angle) + (- normal.y * point.x + normal.x * point.y) * Mathf.Sin(angle));
        }*/


    }
}

