/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using UnityEngine;

namespace Romfarer {
	class LazorCam{
	
		/* The Old:::::::::::::::::::::::::::::::::::::::
		 * "SkySphere Cam" -4
         * "Space Camera" -3   kerbin in map view?
         * "Far Camera" -2  hills far away      far clipping pane
         * "Mid Camera" -1                      middle clipping pane
         * "Main Camera" 0  own vessel +        close clipping pane
         * "UICamera" 100
         * "UI camera" 2
         * "UI mask camera" 3
         * 
         * The New:::::::::::::::::::::::::::::::::::::::
         * Camera ScaledSpace : planets and skybox
         * Camera 00 : same as the old Main Camera
         * Camera 01 : same as the old Mid and Far Camera in one
         * */

		private GameObject camGameObject_Main;
		//private GameObject camGameObject_Mid;
		private GameObject camGameObject_Far;
		private GameObject camGameObject_Space;
		//private GameObject camGameObject_SkySphere;
		public Camera cam_Main;
		//public Camera cam_Mid;
		public Camera cam_Far;
		public Camera cam_Space;
		//public Camera cam_SkySphere;

		protected RenderTexture renTex;
		protected DoubleSwitch toggle_normalCam;
		protected Texture bg = Util.getTexture(new Color(0.45f,0.45f,0.45f,1));
		//private Material material;
		private float fov;
		private int cullingMask;

		public Vector3 forward{ get { return cam_Main.transform.forward; } }
		public Vector3 up{ get { return cam_Main.transform.up; } }
		public Vector3 right{ get { return cam_Main.transform.right; } }
		public Vector3 position{ get { return cam_Main.transform.position; } }

		//image effects
		private Material grayscaleMaterial;
		private Material sepiaMaterial;

		public LazorCam(int width, int height, int depth, RenderTextureFormat format, RenderTextureReadWrite rw, float fov, int cullingMask){
			this.fov = fov;
			this.cullingMask = cullingMask;

			renTex = new RenderTexture(width, height, depth, format, rw);
			renTex.isPowerOfTwo = (width != 0) && ((width & (width - 1)) == 0) && (height != 0) && ((height & (height - 1)) == 0);
			renTex.Create();
			SetupCamera("TargetCam Space", "Camera ScaledSpace", ref camGameObject_Space, ref cam_Space);
			SetupCamera("TargetCam Far", "Camera 01", ref camGameObject_Far, ref cam_Far);
			SetupCamera("TargetCam Main", "Camera 00", ref camGameObject_Main, ref cam_Main);

					//cam_Far.transform.parent = cam_Main.transform; DON'T!!!!!!!!!!!!!!!!!!!!
			cam_Main.cullingMask = cullingMask; //32771; //only render layer 0, 1 and 15 ... layer 19 excluded, lazors are in layer 19
			cam_Far.cullingMask = cullingMask;
					//cam_Space.cullingMask = -1602222592;
			cam_Main.fov = fov;
			cam_Far.fov = fov;
			cam_Space.fov = fov;



			sepiaMaterial = new Material(Shader.Find("Hidden/Sepiatone Effect"));

			grayscaleMaterial = new Material (Shader.Find("Hidden/Grayscale Effect"));
			grayscaleMaterial.SetTexture("_RampTex", Util.LoadTexture(256, 1, "ramp_grayscale.png"));
			grayscaleMaterial.SetFloat("_RampOffset", 0);

		}

		private void SetupCamera(string name, string findName, ref GameObject go, ref Camera cam) {
			go = new GameObject();
			go.name = name +" "+ go.GetInstanceID();
			cam = go.AddComponent<Camera>();
			cam.CopyFrom(FindCamera(findName));
			cam.targetTexture = renTex;
			cam.enabled = false;
		}

		public void DestroyCams(){
			GameObject.Destroy(cam_Space.gameObject);
			GameObject.Destroy(cam_Far.gameObject);
			GameObject.Destroy(cam_Main.gameObject);
		}

		public void SetNewRenderTexture(int width, int height, int depth, RenderTextureFormat format, RenderTextureReadWrite rw, float fov, int cullingMask){
			renTex = new RenderTexture(width, height, depth, format, rw);
			renTex.isPowerOfTwo = (width != 0) && ((width & (width - 1)) == 0) && (height != 0) && ((height & (height - 1)) == 0);
			//RenderTexture.active = renTex;
			renTex.Create();
			SetupCamera("TargetCam Space", "Camera ScaledSpace", ref camGameObject_Space, ref cam_Space);
			SetupCamera("TargetCam Far", "Camera 01", ref camGameObject_Far, ref cam_Far);
			SetupCamera("TargetCam Main", "Camera 00", ref camGameObject_Main, ref cam_Main);

			cam_Main.cullingMask = cullingMask;
			cam_Far.cullingMask = cullingMask;
			/*cam_Main.targetTexture = renTex;
			cam_Far.targetTexture = renTex;
			cam_Space.targetTexture = renTex;*/

			cam_Main.fov = fov;
			cam_Far.fov = fov;
			cam_Space.fov = fov;
		}

		public static Camera FindCamera(string name) {
			foreach (Camera c in Camera.allCameras)
				if (c.name == name)
					return c;
			return null;
		}
		
		public Vector3 ViewportToWorldPoint(Vector3 v, bool planetary){
			if(planetary)
				return cam_Space.ViewportToWorldPoint(v);
			else
				return cam_Main.ViewportToWorldPoint(v);
		}

		/*public void RenderDockingCam(Rect rect, Material mat, float fov, Vector3 pos, Vector3 aim, Quaternion rot, Vector3 normal, Vector3 normal2, bool planetary){
			RenderTexture currentRT = RenderTexture.active;
			RenderTexture.active = renTex;

			renTex.DiscardContents ();
			if(planetary) {
				toggle_normalCam.set = false;
				cam_Space.transform.position = pos;
				cam_Space.transform.forward = aim;
				cam_Space.transform.localRotation = rot;
				cam_Space.Render();
			} else {
				toggle_normalCam.set = true;
				cam_Main.transform.position = pos;
				cam_Main.transform.forward = aim;
				cam_Main.transform.localRotation = rot;

				cam_Far.transform.position = pos;
				cam_Far.transform.forward = aim;
				cam_Far.transform.localRotation = rot;

				cam_Space.transform.rotation = cam_Main.transform.rotation;
				//cam_Space.transform.rotation = rot;

				cam_Space.Render();
				cam_Far.Render();
				cam_Main.Render();

				Graphics.Blit(renTex, renTex, sepiaMaterial);
				Graphics.Blit(renTex, renTex, grayscaleMaterial);
			}
			RenderTexture.active = currentRT;

			Graphics.DrawTexture(rect, bg);
			Graphics.DrawTexture(rect, renTex, mat);
		}*/


		virtual public void RenderCam(Rect rect, Material mat, float fov, Vector3 pos, Vector3 aim, Quaternion rot, Vector3 normal, Vector3 normal2, bool planetary, params Material[] postProcessMats){
			RenderTexture currentRT = RenderTexture.active;
			RenderTexture.active = renTex;

			cam_Main.fov = fov;
			cam_Far.fov = fov;
			cam_Space.fov = fov;

			if(planetary) {
				toggle_normalCam.set = false;
				cam_Space.transform.position = pos;
				cam_Space.transform.forward = aim;
				cam_Space.transform.localRotation = rot;
				cam_Space.Render();
			} else {
				toggle_normalCam.set = true;
				cam_Main.transform.position = pos;
				cam_Main.transform.forward = aim;
				cam_Main.transform.localRotation = rot;

				cam_Far.transform.position = pos;
				cam_Far.transform.forward = aim;
				cam_Far.transform.localRotation = rot;

				cam_Space.transform.rotation = cam_Main.transform.rotation;
				cam_Space.Render();
				cam_Far.Render();
				cam_Main.Render();

				for (int i = 0; i < postProcessMats.Length; i++) {
					Graphics.Blit(renTex, renTex, postProcessMats[i]);
				}
			}
			RenderTexture.active = currentRT;

			Graphics.DrawTexture(rect, bg);
			Graphics.DrawTexture(rect, renTex, mat);
		}


	}









	class LazorCamGui {

		private LazorCam lazorCam;
		public float sliderZoom = 0;
		private float cameraDistance = 15;
		private Vector3 posOffset = Vector3.zero;
		private bool big;
		public Rect vBig = new Rect(5, 30, 266, 266);
		private Rect vSmall = new Rect(5, 30, 200, 200);
		private Rect vCurrent = new Rect(5, 30, 200, 200);
		public Rect RectBounds{ get { return vCurrent; } }

		public LazorCamGui(LazorCam lazorCam){
			this.lazorCam = lazorCam;
		}

		public void ResetOffset() {
			posOffset = Vector3.zero;
		}

		public void CameraGui(CameraModes cameraMode, 
		                      out int width, 
		                      out int height, 
		                      Vector3 ref_aim, 
		                      Vector3 ref_origin,
		                      Vector3 normal,
		                      Vector3 normal2,
		                      Component target){
			Rect expandButton = new Rect(vCurrent.width - 25, vCurrent.height, 20, 20);
			
			float fov = 60;
			Vector3 aim = ref_aim;
			Vector3 pos = Vector3.zero;
			Quaternion rot = Quaternion.identity;
			//cam positioning
			if (target != null) {
				//aim camera after target
				if (target is Collider) {
					cameraDistance = ((Collider)target).bounds.extents.magnitude / 5 * 12;
					//aim = (((Collider)target).bounds.max - ref_origin).normalized;
					aim = (((Collider)target).transform.position - ref_origin).normalized;
					rot = Quaternion.LookRotation(aim, normal);
					pos = target.transform.position - posOffset - aim * cameraDistance;
				} else if (target is Vessel) {
					cameraDistance = 12;
					aim = (target.transform.position - ref_origin).normalized;
					rot = Quaternion.LookRotation(aim, normal);
					pos = target.transform.position - posOffset - aim * cameraDistance;
				} else if (target is CelestialBody) {
					cameraDistance = (float)((CelestialBody)target).Radius * 0.0008f;
					aim = (LazorGlobals.GetMapBody(target.name).position - Planetarium.fetch.Sun.position).normalized;
					rot = Quaternion.LookRotation(aim, Planetarium.up);
					pos = LazorGlobals.GetMapBody(target.name).position - (posOffset) - (aim * cameraDistance); 
				}
				//mouse input on camera window
				if (!expandButton.Contains(Event.current.mousePosition) && Event.current.mousePosition.x > vCurrent.x && Event.current.mousePosition.x < vCurrent.x + vCurrent.width && Event.current.mousePosition.y > vCurrent.y && Event.current.mousePosition.y < vCurrent.y + vCurrent.height) {
					if (Input.GetAxis("Mouse ScrollWheel") < 0) {
						if (sliderZoom < 12f)
							sliderZoom += 0.5f;
					} else if (Input.GetAxis("Mouse ScrollWheel") > 0){
						if (sliderZoom > 0f)
							sliderZoom -= 0.5f;
					} else if (Event.current.type == EventType.mouseDown && Event.current.clickCount > 1) {
						posOffset = Vector3.zero;
					} else if (Event.current.type == EventType.mouseDown) {
						float xO = (Event.current.mousePosition.x - 5) / (vCurrent.width-10);
						float yO = ((vCurrent.width - 10) - (Event.current.mousePosition.y - 30)) / (vCurrent.width - 10);
						if(target is CelestialBody)
							posOffset = LazorGlobals.GetMapBody(target.name).position - lazorCam.ViewportToWorldPoint(new Vector3(xO, yO, cameraDistance), target is CelestialBody);
						else
							posOffset = target.transform.position - lazorCam.ViewportToWorldPoint(new Vector3(xO, yO, cameraDistance), target is CelestialBody);
					}
				}
				//Make cam point in lazor direction
			} else if (target == null || target != null && (target.transform.position - ref_origin).magnitude > LazorGlobals.VIEW_DISTANCE) { 
				if (Vector3.Angle(aim, normal) < 0.5f)
					rot = Quaternion.LookRotation(aim, normal2);
				else
					rot = Quaternion.LookRotation(aim, normal);
				pos = ref_origin;
			}
			fov = 60 - (sliderZoom) * 4.99f;
			
			//draw camera square
			if (big)
				lazorCam.RenderCam(new Rect(5, 30, 256, 256), CameraFilter.Material(cameraMode), fov, pos, ref_aim, rot, normal, normal2, target is CelestialBody);
			else
				lazorCam.RenderCam( new Rect(5, 30, 190, 190), CameraFilter.Material(cameraMode), fov, pos, ref_aim, rot, normal, normal2, target is CelestialBody);
			GUILayout.Space(vCurrent.height);
			
			Rect a = new Rect(expandButton.x, expandButton.y, 20, 20);
			Matrix4x4 matrix = GUI.matrix;
			if (!big){
				GUIUtility.RotateAroundPivot(-90, new Vector2(a.x + 10, a.y + 10)); 
				if (GUI.Button(a, "", Util.Divider_gray_button)) {
					big = true;
					vCurrent = vBig;
				}
			} else if (big){
				GUIUtility.RotateAroundPivot(90, new Vector2(a.x + 10, a.y + 10)); 
				if (GUI.Button(a, "", Util.Divider_gray_button)) {
					big = false;
					vCurrent = vSmall;
				}
			}
			GUI.matrix = matrix;
			width = (int)vCurrent.width;
			height = (int)vCurrent.height;
		}

		public void CameraGuiDocking(Material material, 
		                      int width, 
		                      int height, 
		                      Vector3 ref_aim, 
		                      Vector3 ref_origin,
		                      Vector3 normal,
		                      Vector3 normal2){
			Quaternion rot = Quaternion.identity;
			if (Vector3.Angle(ref_aim, normal) < 0.5f)
				rot = Quaternion.LookRotation(ref_aim, normal2);
			else
				rot = Quaternion.LookRotation(ref_aim, normal);
			Rect expandButton = new Rect(width - 25, height, 20, 20);
			lazorCam.RenderCam(new Rect(5, 30, width, height), material, 40, ref_origin, ref_aim, rot, normal, normal2, false, MultiMaterial.Sepiatone, MultiMaterial.Grayscale);
			GUILayout.Space(height);
		}
	}




	class PartLazorCam : Part{
		private Vector3 magic = new Vector3(0.005f,0,0);
		private Transform partChassy;
		public Transform partSphere;
		public Transform partRotator;
		public Transform partLense;
		private LazorSystemCamera cameraModule;
		public Vector3 Origin{ get { return partSphere.transform.position; } }


		protected override void onPartFixedUpdate() {

			if (cameraModule.on) {
				//print("onPartFixedUpdate 1");
				partChassy.transform.Rotate(Vector3.up, Util.AngleSigned(partChassy.right, Vector3.Cross(cameraModule.aimer.forward + magic, partChassy.up), partChassy.up));
				//print("onPartFixedUpdate 2");
				partSphere.transform.rotation = Quaternion.LookRotation(-cameraModule.aimer.forward - magic, transform.up);
				//print("onPartFixedUpdate 3");
				partRotator.transform.Rotate(Vector3.forward, Util.AngleSigned(partRotator.right, Vector3.Cross(cameraModule.aimer.up, partRotator.forward), partRotator.forward));
				//print("onPartFixedUpdate 4");
				partLense.transform.Rotate(Vector3.forward,  Util.AngleSigned(partLense.right, partSphere.right, partLense.forward) + cameraModule.zoom * 10);
				//print("onPartFixedUpdate 5");
				partLense.localPosition = new Vector3(0, 0, 0.0055644055828f - cameraModule.zoom / 1200);
				//print("onPartFixedUpdate 6");
				RequestResource("ElectricCharge", 0.1d * Time.deltaTime * TimeWarp.CurrentRate);
				//print("onPartFixedUpdate 7");
			}
		}

		protected override void onPartStart() {
			if(HighLogic.LoadedScene == GameScenes.FLIGHT) {
				partChassy = transform.Find("model/CameraChassy");
				partSphere = transform.Find("model/CameraSphere");
				partRotator = transform.Find("model/CameraRotator");
				partLense = transform.Find("model/CameraLense");
				partSphere.parent = partChassy;
				partRotator.parent = partSphere;
				partLense.parent = partSphere;
				if(!Modules.Contains("LazorSystemCamera"))
					cameraModule = AddModule("LazorSystemCamera") as LazorSystemCamera;
				else
					cameraModule = Modules["LazorSystemCamera"] as LazorSystemCamera;
				cameraModule.Initialize();
			}
		}

	}




	class LazorSystemCamera : PartModule{

		private int camHeight = 200;
		private int camWidth = 200;
		private Rect windowPos;
		public bool on;
		private LazorCam lazorCam;
		private LazorCamGui lazorGui;
		public float zoom{ get { return lazorGui.sliderZoom; } }
		private MovementControl mc = new MovementControl();
		public AimingDevice aimer;
		Vector3 originAtLense = Vector3.zero;
		private bool remote_yawPlus;
		private bool remote_yawMinus;
		private bool remote_pitPlus;
		private bool remote_pitMinus;
		private bool remote_rolPlus;
		private bool remote_rolMinus;
		private DoubleSwitch toggleFix;
		private DoubleSwitch toggleLock;
		private Transform ref_origin;
		private PartLazorCam partLazorCam;
		private Transform activeTarget;


		[KSPEvent(guiActive = true, guiName = "Activate Camera")]
		public void ActivateLazorCamera() {
			Events["DeactivateLazorCamera"].active = true;
			Events["ActivateLazorCamera"].active = false;
			on = true;
		}
		
		[KSPEvent(guiActive = true, guiName = "Deactivate Camera")]
		public void DeactivateLazorCamera() {
			Events["DeactivateLazorCamera"].active = false;
			Events["ActivateLazorCamera"].active = true;
			on = false;
		}

		public override void OnStart(PartModule.StartState state) {
			if (vessel != null) {
				//MonoBehaviour.print("LazorSystemCamera START");
				Events["DeactivateLazorCamera"].active = false;
				Events["ActivateLazorCamera"].active = true;
				//Initialize();
			}
		}

		public override void OnAwake() {
			if (vessel != null) {
				//MonoBehaviour.print("LazorSystemCamera AWAKE");
				Events["DeactivateLazorCamera"].active = false;
				//Initialize();
			}
		}

		public void Initialize(){
			//print("Initialize 1");
			lazorCam = new LazorCam(256, 256, 24, RenderTextureFormat.Default, RenderTextureReadWrite.sRGB, 60, 557059); //32771
			//print("Initialize 2");
			lazorGui = new LazorCamGui(lazorCam);
			//print("Initialize 3");
			partLazorCam = (part as PartLazorCam);
			//print("Initialize 4");
			ref_origin = partLazorCam.partSphere;
			//print("Initialize 5");
			/*if(partLazorCam == null)
				print("partlazorcam == null");
			if(partLazorCam.transform == null)
				print("partLazorCam.transform == null");
			if(ref_origin == null)
				print("ref_origin == null");*/
			aimer = new AimingDevice(partLazorCam.transform, 
			                         ref_origin,
			                         -partLazorCam.transform.forward, 
			                         ref_origin.position, 
			                         partLazorCam.transform.up, 
			                         0.1f, 1.0f);
			//print("Initialize 6");
			toggleFix.set = true;
			//print("Initialize 7");
			RenderingManager.AddToPostDrawQueue(0, new Callback(CamGUI));
			//print("Initialize 8");
			vessel.OnJustAboutToBeDestroyed += Dispose;
			//print("Initialize 9");
		}

		private void Dispose() {
			RenderingManager.RemoveFromPostDrawQueue(0, new Callback(CamGUI));
		}

		public override void OnUpdate(){
			if(on) {
				if(toggleLock.In()){
					if(LazorSystemControllerGlobal.ActiveLazorSystemTarget != null){
						activeTarget = LazorSystemControllerGlobal.ActiveLazorSystemTarget.transform;
					} else if(FlightGlobals.fetch.VesselTarget != null){
						activeTarget = FlightGlobals.fetch.VesselTarget.GetTransform();
					}
				}else if(toggleLock.Out()){
				}

				if(toggleLock){
					toggleFix.set = false;
				}
				Vector3 normal = -FlightGlobals.getGeeForceAtPosition(vessel.transform.position);
				if(toggleLock && activeTarget != null)
					aimer.SetLockDirection ((activeTarget.position - transform.position).normalized, aimer.up);

				if(!toggleLock){
					if(remote_yawPlus)
						aimer.Right();
					else if(remote_yawMinus)
						aimer.Left();
					else if(remote_pitPlus)
						aimer.Up();
					else if(remote_pitMinus)
						aimer.Down();
				}
				if(remote_rolPlus)
					aimer.RollRight();
				else if(remote_rolMinus)
					aimer.RollLeft();

				if(toggleFix.In())
					aimer.Fix(true);
				else if(toggleFix.Out())
					aimer.Fix(false);
				aimer.CorrectDeadzone();
				originAtLense = aimer.position + aimer.forward * 0.1f;
			}
		}

		public void CamGUI() {
			UnityEngine.GUI.skin = Util.LazorSkin;
			if (!FlightDriver.Pause && on){ // && part.vessel == FlightGlobals.ActiveVessel) {
				windowPos = Util.ClampToScreen(GUILayout.Window(GetInstanceID(), windowPos, WindowGUI, "Lazor Cam"));
				windowPos.width = camWidth;
				if (Event.current.type == EventType.Repaint)
					windowPos.height = camHeight;
			}
		}

		 public void WindowGUI(int windowID) {
			if(GUI.Button(new Rect(windowPos.width - 20, 0, 20, 20), "", Util.Button_header_close)) { //red close button
				on = false;
				Events["ActivateLazorCamera"].active = true;
				Events["DeactivateLazorCamera"].active = false;
			}
			try {
				lazorGui.CameraGui(CameraModes.Normal, out camWidth, out camHeight, aimer.forward, originAtLense, aimer.up, aimer.up, null);

				GUILayout.BeginHorizontal();
				GUILayout.Label("Zoom");
				lazorGui.sliderZoom = GUILayout.HorizontalSlider(lazorGui.sliderZoom, 0, 12);
				GUILayout.EndHorizontal();

				GUILayout.BeginHorizontal(Util.Space_Compact);
				GUILayout.FlexibleSpace();
			
				GUILayout.BeginVertical(Util.Space_Compact);
				toggleFix.set = GUILayout.Toggle(toggleFix, "Fix");
				GUILayout.FlexibleSpace();
				GUILayout.BeginHorizontal(Util.Space_Compact);
				remote_rolPlus = GUILayout.RepeatButton("Ͼ", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				GUILayout.Space(2);
				remote_yawPlus = GUILayout.RepeatButton("◄", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				GUILayout.EndHorizontal();
				GUILayout.EndVertical();
			
				GUILayout.BeginVertical(Util.Space_Compact);
				remote_pitPlus = GUILayout.RepeatButton("▲", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				remote_pitMinus = GUILayout.RepeatButton("▼", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				GUILayout.EndVertical();
			
				GUILayout.BeginVertical(Util.Space_Compact);
				toggleLock.set = GUILayout.Toggle(toggleLock, "Lock");
				GUILayout.FlexibleSpace();
				GUILayout.BeginHorizontal(Util.Space_Compact);
				remote_yawMinus = GUILayout.RepeatButton("►", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				GUILayout.Space(2);
				remote_rolMinus = GUILayout.RepeatButton("Ͽ", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(22));
				GUILayout.EndHorizontal();
				GUILayout.EndVertical();
			
				GUILayout.FlexibleSpace();
				GUILayout.EndHorizontal();
			} catch (Exception) { 
				on = false;
			}

			GUI.DragWindow(new Rect(0, 0, windowPos.width, 30));
		}

	}









}



