/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using UnityEngine;

namespace Romfarer{

	public class LazorControlT2 : PartModule {
		/*
        MODULE
        {
            name = LazorControlT2
            maxTorque = 1
            maxThrust = 35
            ejectionForce = 5
            burnTime = 45
        }
         */
		[KSPField]
		public Vector3 lastAimpoint = Vector3.zero;
		
		[KSPField]
		public Vector3 lastVelocity = Vector3.zero;






		
		[KSPField]
		public bool receiveControl;
		
		[KSPField]
		public float maxTorque;
		
		[KSPField]
		public float maxThrust;
		
		[KSPField]
		public float ejectionForce;
		
		[KSPField]
		public float burnTime;
		
		[KSPEvent]
		public void AddTorque(BaseEventData data) {//+v.x = yaw right, +v.y = pitch up , v.z = roll
			Vector3 v = data.Get<Vector3>("vector");
			rigidbody.AddRelativeTorque(ClampTorque(v.x), ClampTorque(v.y), ClampTorque(v.z));
		} 
		
		private float ClampTorque(float t) {
			return Mathf.Clamp(t, -maxTorque, maxTorque);
		}
		
		[KSPEvent(guiActive = true, guiName = "Deploy")]
		public void Activate() {
			part.force_activate();
		}
	}
	
	
	public class LazorWarhead : PartModule {
		[KSPField]
		public float blastRadius;

		[KSPField]
		public float destructionRadius;
		
		[KSPEvent]
		public void Detonate() {
			part.explode();
		}
	}



	public class LazorControlT1 : PartModule {
		
		//MODULE
		//{
		// name = LazorControlT1
		//}
		
		
		// in, set by LazorSystem. Should be set to "false" 
		// before LazorSystem goes off-line, decoupled or destroyed.
		[KSPField]
		public bool underControl;
		
		// out; may be "false" when turret is processing "Aim" 
		// command (or if it cannot process it at all).
		// out; in part's Local Space.
		[KSPField]
		public Vector3 aimingOrigin;

		[KSPField]
		public Vector3 aimingDirection;

		// "data" contains Vector3 "direction" (in World Space), 
		// otherwise aim's dropped.
		[KSPEvent]
		public void Aim(BaseEventData data) {
			aimingDirection = data.Get<Vector3>("vector");
		

			//fix
		}
		
		
	}



	public class LazorTriggerContinous : PartModule {
		
		[KSPField] public string triggerName; // out constant, the name of the trigger button in LazorSystem (Fire, Trigger, On, Active etc.)
		
		[KSPField] public bool triggerOn;    // in, true = continous fire
		
	}
	
	
	
	
	
	
	
	
	

	//Put on each vessel OnFlightStart. Tracks if the vessel is destroyed and registers the event.
	public class DestructionTracker : PartModule {
		private int crewCount;

		public void LateUpdate() {
			crewCount = vessel.GetCrewCount();
			//add DestructionTrackerEVA to eva's
			if (FlightGlobals.ActiveVessel.isEVA && !FlightGlobals.ActiveVessel.rootPart.Modules.Contains("DestructionTrackerEVA")) {
				MonoBehaviour.print("[DestructionTracker: Added DestructionTrackerEVA to " + FlightGlobals.ActiveVessel.name + "]");
				FlightGlobals.ActiveVessel.rootPart.AddModule("DestructionTrackerEVA");
			}
		}

		private void OnJustAboutToBeDestroyed(){
			print("[DestructionTracker: OnJustAboutToBeDestroyed]");
			Romfarer.Events.RegisterVesselDestroyed(vessel.vesselName);
			//register kerbals killed
			for (int i = 0; i < crewCount; i++)
				Romfarer.Events.RegisterKerbalKilled("Crewmember in " + vessel.vesselName);
		}

		public override void OnAwake() {
			MonoBehaviour.print("[DestructionTracker: OnAwake " + vessel.vesselName + "]");
			part.OnJustAboutToBeDestroyed += OnJustAboutToBeDestroyed;
		}
	}

	public class DestructionTrackerEVA : PartModule {
		public override void OnAwake() {
			MonoBehaviour.print("[DestructionTrackerEVA: OnAwake " + vessel.name + "]");
			part.OnJustAboutToBeDestroyed += OnJustAboutToBeDestroyed;
		}

		private void OnJustAboutToBeDestroyed() {
			print("[DestructionTrackerEVA: OnJustAboutToBeDestroyed]");
			Romfarer.Events.RegisterKerbalKilled(vessel.name);
		}
	}



	class LazorSystemDockingCamera : PartModule {

		public bool on;
		public DockingCamera dockingCamera;


		[KSPEvent(guiActive = true, guiName = "Activate Docking Camera")]
		public void ActivateLazorCamera() {
			Events["DeactivateLazorCamera"].active = true;
			Events["ActivateLazorCamera"].active = false;
			//on = true;
			LazorGlobals.ActiveDockingCamera = this;
		}

		public void NinjaActivateLazorCamera() {
			Events["DeactivateLazorCamera"].active = true;
			Events["ActivateLazorCamera"].active = false;
			on = true;
		}

		[KSPEvent(guiActive = true, guiName = "Deactivate Docking Camera")]
		public void DeactivateLazorCamera() {
			Events["DeactivateLazorCamera"].active = false;
			Events["ActivateLazorCamera"].active = true;
			on = false;
		}

		public void CamGUI() {
			UnityEngine.GUI.skin = Util.LazorSkin;
			if (!FlightDriver.Pause && on && part.vessel == FlightGlobals.ActiveVessel){// && (!LazorSystemControllerGlobal.ActiveVesselHasLazorSystem() || LazorSystemControllerGlobal.ActiveVesselHasLazorSystemAndItIsOn())) {
				dockingCamera.Update();
				dockingCamera.windowPos = Util.ClampToScreen(GUILayout.Window(0, dockingCamera.windowPos, dockingCamera.WindowGUI, name, GUILayout.MinWidth(266)));
			}
		}

		public override void OnStart(PartModule.StartState state) {
			if (vessel != null) {
				MonoBehaviour.print("LazorSystemDockingCamera START");
				Events["DeactivateLazorCamera"].active = false;
				Events["ActivateLazorCamera"].active = true;
				dockingCamera = new DockingCamera(this, part.Modules["ModuleDockingNode"]);
				RenderingManager.AddToPostDrawQueue(0, new Callback(CamGUI));
				part.OnJustAboutToBeDestroyed += Dispose;
			}
		}

		public override void OnAwake() {
			if (vessel != null) {
				MonoBehaviour.print("LazorSystemDockingCamera AWAKE");
				Events["DeactivateLazorCamera"].active = false;
				dockingCamera = new DockingCamera(this, part.Modules["ModuleDockingNode"]);
				RenderingManager.AddToPostDrawQueue(0, new Callback(CamGUI));
				part.OnJustAboutToBeDestroyed += Dispose;
			}
		}

		private void Dispose() {
			RenderingManager.RemoveFromPostDrawQueue(0, new Callback(CamGUI));
		}
	}

	public class LazorSystemMainModule : PartModule {

		[KSPEvent(guiActive = true, guiName = "Activate")]
		public void Activate() {
			Events["Deactivate"].active = true;
			Events["Activate"].active = false;
			((LazorSystem)part).lazorOn.set = true;
			RuntimeLogic.Instance.AppLauncherButtonSetState();
		}

		[KSPEvent(guiActive = true, guiName = "Deactivate")]
		public void Deactivate() {
			Events["Deactivate"].active = false;
			Events["Activate"].active = true;
			((LazorSystem)part).lazorOn.set = false;
			RuntimeLogic.Instance.AppLauncherButtonSetState();
		}

		[KSPEvent(guiActive = true, guiName = "Control From Here")]
		public void ControlFromHere() {
			(part as LazorSystem).lazorSystemController.SetActive(part as LazorSystem);
			MonoBehaviour.print("[LazorSystemMainModule: Setting active lazor system to " + ((LazorSystem)part).uid.ToString() + "]");
			RuntimeLogic.Instance.AppLauncherButtonSetState();
		}

		[KSPEvent(guiActive = true, guiName = "Toggle Color")]
		public void ToggleColor() {
			LazorSystem system = (LazorSystem)part;
			if (system.Leader.GetColor() == Color.white)
				system.SetLeaderColor(Color.red);
			else if (system.Leader.GetColor() == Color.red)
				system.SetLeaderColor(Color.green);
			else if (system.Leader.GetColor() == Color.green)
				system.SetLeaderColor(Color.blue);
			else if (system.Leader.GetColor() == Color.blue)
				system.SetLeaderColor(Color.white);
			system.leaderLocked = false;
		}

		public override void OnUpdate() {
			//if (LazorGlobals.IsActiveLazorSystem((LazorSystem)part)) {
			if ((part as LazorSystem).IsActiveLazorSystem()) {
				if (((LazorSystem)part).lazorOn) {
					Events["Deactivate"].active = true;
					Events["Activate"].active = false;
					Events["ToggleColor"].active = true;
				} else {
					Events["Deactivate"].active = false;
					Events["Activate"].active = true;
					Events["ToggleColor"].active = false;
				}
				Events["ControlFromHere"].active = false;
			} else {
				Events["Deactivate"].active = true;
				Events["Activate"].active = false;
				Events["ControlFromHere"].active = true;
				Events["ToggleColor"].active = false;
			}
		}

	}
	




}

