/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/

using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace Romfarer
{
	interface IRoboticArmControlSuper{
		string moduleName{ get; }
	}
	interface IRoboticArmControl : IRoboticArmControlSuper{
		void Initialize();
		bool Fixed();
		void Gui();
		string title{ get; }
		Vector3 cameraPosition{ get; }
		Vector3 cameraForward{ get; }
		Vector3 cameraNormal{ get; }
		float magnetDistance{ get; }
		//string moduleName{ get; }
	}

	class PartRoboticArm : Part, IRoboticArmControlSuper {
		protected string ModuleNameXXX = "CRAP";
		public string moduleName { get { return ModuleNameXXX; } }

		public LazorRoboticArm robotArm;
		public LazorSound lazorSound;
		virtual public Transform armTip{ get{return null;} }


		protected override void onFlightStart(){
			state = PartStates.ACTIVE;
			lazorSound = new LazorSound(this);
		}

		override protected void onGamePause(){
			if(lazorSound != null)
				lazorSound.OnPause();
		}

		override protected void onGameResume() {
			if (lazorSound != null)
				lazorSound.OnResume();
		}

		override protected void onPack(){
			robotArm.onPack();
		}
		override protected void onUnpack(){
			robotArm.onUnpack();
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////////////////
	/// BURAN
	///////////////////////////////////////////////////////////////////////////////////////////////////

	class PartRoboticArmBuran : PartRoboticArm, IRoboticArmControl{
		public Transform tBase;
		public Transform tShoulder;
		public Transform tShoulderLong;
		public Transform tElbowLong;
		public Transform tArmWrist;
		public Transform tArmWristLong;
		public Transform tCamera;
		override public Transform armTip{ get{return tCamera;} }
		private bool rep_basePos;
		private bool rep_baseNeg;
		private bool rep_baseRotLeft;
		private bool rep_baseRotRight;
		private bool rep_baseLongFwd;
		private bool rep_baseLongBwd;
		private bool rep_baseArmFwd;
		private bool rep_baseArmBwd;
		private bool rep_ArmWhistFwd;
		private bool rep_ArmWhistBwd;
		private bool rep_ArmWhistRotateFwd;
		private bool rep_ArmWhistRotateBwd;
		public string title { get { return "Buran Cam"; } }
		public Vector3 cameraPosition{ get { return armTip.position + armTip.up.normalized * 0.4f; } }
		public Vector3 cameraForward{ get { return armTip.up; } }
		public Vector3 cameraNormal{ get { return armTip.forward; } }
		public float magnetDistance{ get { return 0.5f; } }
		private float armSpeed = 0.3f;

		override protected void onPartAwake() {
			ModuleNameXXX = "LazorRoboticArmBuran";
		}

		public void Initialize(){
			LazorRoboticArmBuran robotArm = this.robotArm as LazorRoboticArmBuran;
			if(robotArm.tBase != Vector3.zero)
				tBase.localEulerAngles = robotArm.tBase;
			if(robotArm.tShoulder != Vector3.zero)
				tShoulder.localEulerAngles = robotArm.tShoulder;
			if(robotArm.tShoulderLong != Vector3.zero)
				tShoulderLong.localEulerAngles = robotArm.tShoulderLong;
			if(robotArm.tElbowLong != Vector3.zero)
				tElbowLong.localEulerAngles = robotArm.tElbowLong;
			if(robotArm.tArmWrist != Vector3.zero)
				tArmWrist.localEulerAngles = robotArm.tArmWrist;
			if(robotArm.tArmWristLong != Vector3.zero)
				tArmWristLong.localEulerAngles = robotArm.tArmWristLong;
		}

		public bool Fixed(){
			LazorRoboticArmBuran robotArm = this.robotArm as LazorRoboticArmBuran;
			robotArm.tBase = tBase.localEulerAngles;
			robotArm.tShoulder = tShoulder.localEulerAngles;
			robotArm.tShoulderLong = tShoulderLong.localEulerAngles;
			robotArm.tElbowLong = tElbowLong.localEulerAngles;
			robotArm.tArmWrist = tArmWrist.localEulerAngles;
			robotArm.tArmWristLong = tArmWristLong.localEulerAngles;

			armSpeed = 0.15f;
			if(FlightInputHandler.fetch.precisionMode)
				armSpeed = 0.02f;

			if (rep_basePos && (robotArm.tBase.y < 90 || robotArm.tBase.y > 260)) { //90*2 dof
				robotArm.Movement (delegate() { tBase.Rotate (Vector3.up, armSpeed); });
				return true;
			} else if (rep_baseNeg && (robotArm.tBase.y > 270 || robotArm.tBase.y < 100)) {
				robotArm.Movement (delegate() { tBase.Rotate (Vector3.up, -armSpeed); });
				return true;
			} else if (rep_baseRotLeft) { //360 dof
				robotArm.Movement (delegate() { tShoulder.Rotate (Vector3.forward, -armSpeed); });
				return true;
			} else if (rep_baseRotRight) {
				robotArm.Movement (delegate() { tShoulder.Rotate (Vector3.forward, armSpeed); }); 
				return true;
			} else if (rep_baseLongFwd && !((robotArm.tShoulderLong.x < 320 && robotArm.tShoulderLong.y < 10 || robotArm.tShoulderLong.x < 320 && robotArm.tShoulderLong.y > 170) && robotArm.tShoulderLong.x > 260)) {
				robotArm.Movement (delegate() { tShoulderLong.Rotate (Vector3.right, -armSpeed); });
				return true;
			} else if (rep_baseLongBwd && !((robotArm.tShoulderLong.x < 310 && robotArm.tShoulderLong.y < 10 || robotArm.tShoulderLong.x < 330 && robotArm.tShoulderLong.y > 170) && robotArm.tShoulderLong.x > 260)) {
				robotArm.Movement (delegate() { tShoulderLong.Rotate (Vector3.right, armSpeed); });
				return true;
			} else if (rep_baseArmFwd && !(robotArm.tElbowLong.x > 340 && robotArm.tElbowLong.y > 170 || robotArm.tElbowLong.x < 65 && robotArm.tElbowLong.y > 170)) {
				robotArm.Movement (delegate() { tElbowLong.Rotate (Vector3.right, -armSpeed); });
				return true;
			} else if (rep_baseArmBwd && !(robotArm.tElbowLong.x > 350 && robotArm.tElbowLong.y > 170 || robotArm.tElbowLong.x < 75 && robotArm.tElbowLong.y > 170)) {
				robotArm.Movement (delegate() { tElbowLong.Rotate (Vector3.right, armSpeed); });
				return true;
			} else if (rep_ArmWhistFwd && (robotArm.tArmWrist.x > 270 || robotArm.tArmWrist.x < 100)) { //90*2 dof
				robotArm.Movement (delegate() { tArmWrist.Rotate (Vector3.right, -armSpeed); });
				return true;
			} else if (rep_ArmWhistBwd && (robotArm.tArmWrist.x > 260 || robotArm.tArmWrist.x < 90)) {
				robotArm.Movement (delegate() { tArmWrist.Rotate (Vector3.right, armSpeed); });
				return true;
			} else if (rep_ArmWhistRotateFwd && (robotArm.tArmWristLong.z > 270 || robotArm.tArmWristLong.z < 100)) { //90*2 dof
				robotArm.Movement (delegate() { tArmWristLong.Rotate (Vector3.forward, -armSpeed); });
				return true;
			} else if (rep_ArmWhistRotateBwd && (robotArm.tArmWristLong.z > 260 || robotArm.tArmWristLong.z < 90)) {
				robotArm.Movement (delegate() { tArmWristLong.Rotate (Vector3.forward, armSpeed); });
				return true;
			} else
				return false;
		}

		public void Gui(){
			GUILayout.BeginHorizontal(Util.Space_Compact);
			GUILayout.FlexibleSpace();

			GUILayout.BeginHorizontal(Util.Space_Compact);
			GUILayout.BeginVertical(Util.Space_Compact);
			rep_baseRotLeft = GUILayout.RepeatButton("Ͼ", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_basePos = GUILayout.RepeatButton("◄", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_baseLongFwd = GUILayout.RepeatButton("▲", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_baseLongBwd = GUILayout.RepeatButton("▼", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_baseRotRight = GUILayout.RepeatButton("Ͽ", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_baseNeg = GUILayout.RepeatButton("►", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();
			GUILayout.EndHorizontal();

			GUILayout.FlexibleSpace();

			GUILayout.BeginHorizontal(Util.Space_Compact);
			GUILayout.BeginVertical(Util.Space_Compact);
			rep_baseArmFwd = GUILayout.RepeatButton("▲", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_baseArmBwd = GUILayout.RepeatButton("▼", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_ArmWhistFwd = GUILayout.RepeatButton("▲", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_ArmWhistBwd = GUILayout.RepeatButton("▼", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_ArmWhistRotateFwd = GUILayout.RepeatButton("◄", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_ArmWhistRotateBwd = GUILayout.RepeatButton("►", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();
			GUILayout.EndHorizontal();

			GUILayout.FlexibleSpace();
			GUILayout.EndHorizontal();
		}

		protected override void onFlightStart(){
			base.onFlightStart ();
			robotArm = Modules["LazorRoboticArmBuran"] as LazorRoboticArmBuran;
			tBase = transform.Find("model/Buran/arm_base/Arm_base_rotate");
			tShoulder = tBase.Find("arm_shoulder");
			tShoulderLong = tShoulder.Find("Arm_shoulder_long");
			tElbowLong = tShoulderLong.Find("arm_elbow_long");
			tArmWrist = tElbowLong.Find("Arm_whist");
			tArmWristLong = tArmWrist.Find("Arm_whist_rotate");
			tCamera = tArmWristLong.Find("Arm_camera");
			robotArm.Initialize(this);
		}
	}

	class LazorRoboticArmBuran : LazorRoboticArm{
		[KSPField (isPersistant=true)] public Vector3 tBase = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tShoulder = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tShoulderLong = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tElbowLong = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tArmWrist = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tArmWristLong = Vector3.zero;
	}




	///////////////////////////////////////////////////////////////////////////////////////////////////
	/// CANADARM 2
	///////////////////////////////////////////////////////////////////////////////////////////////////




	class PartRoboticArmCanadarm : PartRoboticArm, IRoboticArmControl{
		override public Transform armTip{ get{return tHand;} }
		public Transform tBase;
		public Transform tShoulder;
		public Transform tLong1;
		public Transform tLong2;
		public Transform tWrist;
		public Transform tWristRot;
		public Transform tHand;
		private float armSpeed = 0.3f;
		private bool rep_basePos;
		private bool rep_baseNeg;
		private bool rep_RotLeft;
		private bool rep_RotRight;
		private bool rep_LongFwd;
		private bool rep_LongBwd;
		private bool rep_ArmFwd;
		private bool rep_ArmBwd;
		private bool rep_ArmWristFwd;
		private bool rep_ArmWristBwd;
		private bool rep_ArmWristRotateFwd;
		private bool rep_ArmWristRotateBwd;
		private bool rep_ArmHandFwd;
		private bool rep_ArmHandBwd;
		public string title { get{return "Canadarm2 Cam";}}
		public Vector3 cameraPosition{ get { return armTip.position + armTip.forward.normalized * 0.4f; } }
		public Vector3 cameraForward{ get { return armTip.forward; } }
		public Vector3 cameraNormal{ get { return armTip.up; } }
		public float magnetDistance{ get { return 0.9f; } }


		override protected void onPartAwake() {
			ModuleNameXXX = "LazorRoboticArmCanadarm";
		}

		public void Initialize(){
			LazorRoboticArmCanadarm robotArm = this.robotArm as LazorRoboticArmCanadarm;
			if(robotArm.tBase != Vector3.zero)
				tBase.localEulerAngles = robotArm.tBase;
			if(robotArm.tShoulder != Vector3.zero)
				tShoulder.localEulerAngles = robotArm.tShoulder;
			if(robotArm.tLong1 != Vector3.zero)
				tLong1.localEulerAngles = robotArm.tLong1;
			if(robotArm.tLong2 != Vector3.zero)
				tLong2.localEulerAngles = robotArm.tLong2;
			if(robotArm.tWrist != Vector3.zero)
				tWrist.localEulerAngles = robotArm.tWrist;
			if(robotArm.tWristRot != Vector3.zero)
				tWristRot.localEulerAngles = robotArm.tWristRot;
			if(robotArm.tHand != Vector3.zero)
				tHand.localEulerAngles = robotArm.tHand;
		}

		public bool Fixed(){
			LazorRoboticArmCanadarm robotArm = this.robotArm as LazorRoboticArmCanadarm;
			robotArm.tBase = tBase.localEulerAngles;
			robotArm.tShoulder = tShoulder.localEulerAngles;
			robotArm.tLong1 = tLong1.localEulerAngles;
			robotArm.tLong2 = tLong2.localEulerAngles;
			robotArm.tWrist = tWrist.localEulerAngles;
			robotArm.tWristRot = tWristRot.localEulerAngles;
			robotArm.tHand = tHand.localEulerAngles;

			armSpeed = 0.15f;
			if(FlightInputHandler.fetch.precisionMode)
				armSpeed = 0.02f;

			if (rep_basePos) { //90*2 dof
				robotArm.Movement (delegate() { tBase.Rotate (Vector3.forward, armSpeed); });
				return true;
			} else if (rep_baseNeg) {
				robotArm.Movement (delegate() { tBase.Rotate (Vector3.forward, -armSpeed); });
				return true;
			} else if (rep_RotLeft) { //360 dof
				robotArm.Movement (delegate() { tShoulder.Rotate (Vector3.up, -armSpeed); });
				return true;
			} else if (rep_RotRight) {
				robotArm.Movement (delegate() { tShoulder.Rotate (Vector3.up, armSpeed); }); 
				return true;
			} else if (rep_LongFwd) {
				robotArm.Movement (delegate() { tLong1.Rotate (Vector3.forward, -armSpeed); });
				return true;
			} else if (rep_LongBwd) {
				robotArm.Movement (delegate() { tLong1.Rotate (Vector3.forward, armSpeed); });
				return true;
			} else if (rep_ArmFwd) {
				robotArm.Movement (delegate() { tLong2.Rotate (Vector3.forward, -armSpeed); });
				return true;
			} else if (rep_ArmBwd) {
				robotArm.Movement (delegate() { tLong2.Rotate (Vector3.forward, armSpeed); });
				return true;
			} else if (rep_ArmWristFwd) { //90*2 dof
				robotArm.Movement (delegate() { tWrist.Rotate (Vector3.forward, -armSpeed); });
				return true;
			} else if (rep_ArmWristBwd) {
				robotArm.Movement (delegate() { tWrist.Rotate (Vector3.forward, armSpeed); });
				return true;
			} else if (rep_ArmWristRotateFwd) { //90*2 dof
				robotArm.Movement (delegate() { tWristRot.Rotate (Vector3.up, -armSpeed); });
				return true;
			} else if (rep_ArmWristRotateBwd) {
				robotArm.Movement (delegate() { tWristRot.Rotate (Vector3.up, armSpeed); });
				return true;
			} else if (rep_ArmHandFwd) { //90*2 dof
				robotArm.Movement (delegate() { tHand.Rotate (Vector3.forward, -armSpeed); });
				return true;
			} else if (rep_ArmHandBwd) {
				robotArm.Movement (delegate() { tHand.Rotate (Vector3.forward, armSpeed); });
				return true;
			}else
				return false;
		}

		public void Gui(){
			GUILayout.BeginHorizontal(Util.Space_Compact);
			GUILayout.FlexibleSpace();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_basePos = GUILayout.RepeatButton("►", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_baseNeg = GUILayout.RepeatButton("◄", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.FlexibleSpace();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_RotLeft = GUILayout.RepeatButton("►", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_RotRight = GUILayout.RepeatButton("◄", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.FlexibleSpace();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_LongBwd = GUILayout.RepeatButton("▲", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_LongFwd = GUILayout.RepeatButton("▼", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.FlexibleSpace();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_ArmBwd = GUILayout.RepeatButton("▲", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_ArmFwd = GUILayout.RepeatButton("▼", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.FlexibleSpace();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_ArmWristBwd = GUILayout.RepeatButton("▲", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_ArmWristFwd = GUILayout.RepeatButton("▼", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.FlexibleSpace();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_ArmWristRotateFwd = GUILayout.RepeatButton("►", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_ArmWristRotateBwd = GUILayout.RepeatButton("◄", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.FlexibleSpace();

			GUILayout.BeginVertical(Util.Space_Compact);
			rep_ArmHandFwd = GUILayout.RepeatButton("Ͽ", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			rep_ArmHandBwd = GUILayout.RepeatButton("Ͼ", Util.Buttons_arrows_mainPanel, GUILayout.ExpandWidth(true), GUILayout.Width(21));
			GUILayout.EndVertical();

			GUILayout.FlexibleSpace();
			GUILayout.EndHorizontal();
		}

		protected override void onFlightStart(){
			base.onFlightStart ();
			robotArm = Modules["LazorRoboticArmCanadarm"] as LazorRoboticArmCanadarm;
			tBase = transform.Find("model/GameObject/Arm_base/arm_base_rot");
			tShoulder = tBase.Find("arm_shoulder");
			tLong1 = tShoulder.Find("Arm_long1");
			tLong2 = tLong1.Find("Arm_long2");
			tWrist = tLong2.Find("Arm_wrist");
			tWristRot = tWrist.Find("Arm_wrist_rot");
			tHand = tWristRot.Find("Arm_hand");
			robotArm.Initialize(this);
		}
	}

	class LazorRoboticArmCanadarm : LazorRoboticArm{
		[KSPField (isPersistant=true)] public Vector3 tBase = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tShoulder = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tLong1 = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tLong2 = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tWrist = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tWristRot = Vector3.zero;
		[KSPField (isPersistant=true)] public Vector3 tHand = Vector3.zero;
	}




	///////////////////////////////////////////////////////////////////////////////////////////////////
	/// MASTER MODULE
	///////////////////////////////////////////////////////////////////////////////////////////////////



	class LazorRoboticArm : PartModule{

		public IRoboticArmControl control;
		private LazorCam lazorCam;
		private LazorCamGui lazorGui;
		private PartRoboticArm partRoboticArm;
		private int camHeight = 200;
		private int camWidth = 200;

		private bool magnet;
		private Color blinkColor = new Color(0, 1, 0, 0.5f);
		private int blink = 0;
		private bool blinking;
		private Material material;
		private GameObject go_om;
		private GameObject go_sm;
		private bool connectable;
		private DoubleSwitch moving;

		private Part ref_hitPart;
		private FixedJoint joint_static;

		[KSPField (isPersistant=true)] public bool cameraOn;
		private Rect windowPos;
		[KSPField (isPersistant=true)] public Vector2 windowPosition;
		private DoubleSwitch magnetize; 
		[KSPField (isPersistant=true)] public bool isMagnetized; //persistent version of magnetize
		[KSPField (isPersistant=true)] public bool isMainConnection;
		[KSPField (isPersistant=true)] public string partRef;
		[KSPField] public DockedVesselInfo vesselRef;
		[KSPField] public PartRefs partRefs;

		private GameObject go_joint;
		private ConfigurableJoint joint_rotate;
		Quaternion initialRotation;
		private ConfigurableJoint joint_pos;

		public delegate void Move();
		private Rigidbody hitBody = null; //current hit reference
		private Part hitPart = null; //current hit reference
		private bool startFromSaveMagnetized;
		//GameObject debug;
		//GameObject debug2;


		[KSPEvent(guiActive = true, guiName = "Activate Camera")]
		public void ActivateLazorCamera() {
			Events["DeactivateLazorCamera"].active = true;
			Events["ActivateLazorCamera"].active = false;
			cameraOn = true;
		}

		[KSPEvent(guiActive = true, guiName = "Deactivate Camera")]
		public void DeactivateLazorCamera() {
			Events["DeactivateLazorCamera"].active = false;
			Events["ActivateLazorCamera"].active = true;
			cameraOn = false;
		}

		public override void OnStart(PartModule.StartState state) {
			if (vessel != null) {
				Events["DeactivateLazorCamera"].active = false;
				Events["ActivateLazorCamera"].active = true;
			}
		}

		public override void OnLoad(ConfigNode node){
			if(node.HasNode("vesselRef")) {
				vesselRef = new DockedVesselInfo();
				vesselRef.Load(node.GetNode("vesselRef"));
			}
			if(node.HasNode("partRefs")) {
				partRefs = new PartRefs();
				partRefs.Load(node.GetNode("partRefs"));
			}
		}

		public override void OnSave(ConfigNode node){
			if(vesselRef != null)
				vesselRef.Save(node.AddNode("vesselRef"));
			if(partRefs != null)
				partRefs.Save(node.AddNode("partRefs"));
		}

		public void Initialize(IRoboticArmControl control){
			this.control = control;
			lazorCam = new LazorCam(256, 256, 24, RenderTextureFormat.Default, RenderTextureReadWrite.sRGB, 60, 557059); //32771
			lazorGui = new LazorCamGui(lazorCam);
			partRoboticArm = (part as PartRoboticArm);
			RenderingManager.AddToPostDrawQueue(0, new Callback(CamGUI));
			part.OnJustAboutToBeDestroyed += Dispose;
			material = new Material(Shader.Find("Particles/Additive"));

			control.Initialize ();

			windowPos.x = windowPosition.x;
			windowPos.y = windowPosition.y;

			//re-attach to saved grabbed object
			if(isMagnetized) {
				magnet = true;
				magnetize.set = true;
				magnetize.In();
				connectable = true;
				ref_hitPart = vessel[uint.Parse(partRef)];
				startFromSaveMagnetized = true;
			}
			GameEvents.onTimeWarpRateChanged.Add(onTimeWarpChangeRate);
			GameEvents.onSameVesselDock.Add(onSameVesselDock);
			GameEvents.onSameVesselUndock.Add(onSameVesselUndock);
		}

		private void onSameVesselDock(GameEvents.FromToAction<ModuleDockingNode, ModuleDockingNode> action) {
			if(action.from.vessel == vessel && isMainConnection) {
				print("same vessel dock");
				MakeStatic();
			}
		}
		private void onSameVesselUndock(GameEvents.FromToAction<ModuleDockingNode, ModuleDockingNode> action) {
			if(action.from.vessel == vessel && !isMainConnection) {
				print("same vessel Undock");
				//SearchForArmsOnDisconnectedVessel();
				MakeDynamic();
			}
		}


		public void OnDestroy(){
			GameEvents.onTimeWarpRateChanged.Remove(onTimeWarpChangeRate);
			GameEvents.onSameVesselDock.Remove(onSameVesselDock);
			GameEvents.onSameVesselUndock.Remove(onSameVesselUndock);
			cameraOn = false;
		}

		private void Dispose() {
			RenderingManager.RemoveFromPostDrawQueue(0, new Callback(CamGUI));
			partRoboticArm.lazorSound.StopLoop();
		}

		private void ResetPositions(PartRefs partRefs, Part root) {
			using (List<uint>.Enumerator enumerator = partRefs.GetEnumerator())
				while (enumerator.MoveNext()){
					Part part = ref_hitPart.vessel[enumerator.Current];
					if(part != null)
						part.UpdateOrgPosAndRot(root);
				}
		}

		/*private void onPartCouple(GameEvents.FromToAction<Part,Part> action) {
			MonoBehaviour.print("Arm detected couple");
			if(action.from.Modules.Contains("ModuleDockingNode"))
			   MonoBehaviour.print("from is dockingnode");
			if(action.to.Modules.Contains("ModuleDockingNode"))
				MonoBehaviour.print("to is dockingnode");
		}*/

		//*********************************************************
		// Movement
		//*********************************************************

		virtual public void Movement(Move move){
			if(TimeWarp.CurrentRate == 1) { //arm can only move when you are not warping
				moving.set = true;
				if(isMagnetized) {
					if (joint_rotate != null) { //FIX: maybe add a flag for when the joint is movable
						move ();
						joint_rotate.transform.parent = partRoboticArm.transform;
						go_om.transform.parent = partRoboticArm.transform;
						joint_rotate.SetTargetRotationLocal (go_om.transform.localRotation, initialRotation); //local space
						go_om.transform.parent = partRoboticArm.armTip;
						joint_rotate.transform.parent = null;

						joint_pos.targetPosition = -go_sm.transform.InverseTransformPoint (go_om.transform.position);

						ResetPositions (partRefs, vessel.rootPart);
					}
				} else
					move();
				part.RequestResource("ElectricCharge", 0.5d * Time.deltaTime * TimeWarp.CurrentRate);
			}
		}

		private void SetupDummy_ObjectMarker(){
			go_om = new GameObject("Robotic Joint ObjectMarker");
			go_om.transform.rotation = ref_hitPart.transform.rotation;
			go_om.transform.position = ref_hitPart.transform.position;
			go_om.transform.parent = partRoboticArm.armTip;
		}

		private void SetupDummy_StationaryMarker(){
			go_sm = new GameObject("Robotic Joint StationaryMarker");
			go_sm.transform.rotation = ref_hitPart.transform.rotation;
			go_sm.transform.position = ref_hitPart.transform.position;
			go_sm.transform.parent = partRoboticArm.transform;
		}

		// joint connector, arm - position joint - this - rotation joint - grabbed object
		private void SetupDummy_Joint(){
			go_joint = new GameObject("Robotic Joint");
			go_joint.transform.rotation = ref_hitPart.transform.rotation;
			go_joint.transform.position = ref_hitPart.transform.position;
			go_joint.AddComponent<Rigidbody>();
			go_joint.rigidbody.mass = 0.1f;
		}

		private void SetupJointRotation(){
			//GameObject.Destroy(ref_hitPart.attachJoint);
			//ref_hitPart.attachJoint = null;
			ref_hitPart.attachJoint.DestroyJoint();
			joint_rotate = ref_hitPart.gameObject.AddComponent<ConfigurableJoint>();
			joint_rotate.name = "lazor_rotation_joint";
			joint_rotate.transform.parent = partRoboticArm.transform;
			initialRotation = joint_rotate.transform.localRotation; //local space
			joint_rotate.SetTargetRotationLocal (initialRotation, initialRotation);
			joint_rotate.transform.parent = null;
			joint_rotate.breakForce = float.PositiveInfinity;
			joint_rotate.breakTorque = float.PositiveInfinity;
			joint_rotate.xMotion = ConfigurableJointMotion.Locked;
			joint_rotate.yMotion = ConfigurableJointMotion.Locked;
			joint_rotate.zMotion = ConfigurableJointMotion.Locked;
			joint_rotate.angularXMotion = ConfigurableJointMotion.Free;
			joint_rotate.angularYMotion = ConfigurableJointMotion.Free;
			joint_rotate.angularZMotion = ConfigurableJointMotion.Free;
			joint_rotate.rotationDriveMode = RotationDriveMode.XYAndZ;
			JointDrive drv_rot = new JointDrive();
			drv_rot.mode = JointDriveMode.Position;
			drv_rot.positionSpring = 10;
			drv_rot.positionDamper = 0;
			drv_rot.maximumForce = 1;
			joint_rotate.slerpDrive = joint_rotate.angularXDrive = joint_rotate.angularYZDrive = drv_rot;
			joint_rotate.connectedBody = go_joint.rigidbody;
			//ref_hitPart.attachJoint = joint_rotate;

			JointBreakCallback jbc = ref_hitPart.gameObject.AddComponent<JointBreakCallback>() as JointBreakCallback;
			jbc.AddCallback(Decouple);
		}

		private void SetupJointPosition(){
			joint_pos = go_joint.AddComponent<ConfigurableJoint>();
			joint_pos.configuredInWorldSpace = false;
			joint_pos.breakForce = float.PositiveInfinity;
			joint_pos.breakTorque = float.PositiveInfinity;
			joint_pos.anchor = Vector3.zero;
			joint_pos.xMotion = ConfigurableJointMotion.Free;
			joint_pos.yMotion = ConfigurableJointMotion.Free;
			joint_pos.zMotion = ConfigurableJointMotion.Free;
			joint_pos.angularXMotion = ConfigurableJointMotion.Locked;
			joint_pos.angularYMotion = ConfigurableJointMotion.Locked;
			joint_pos.angularZMotion = ConfigurableJointMotion.Locked;
			JointDrive drv = new JointDrive();
			drv.mode = JointDriveMode.Position;
			drv.positionSpring = 100000;
			drv.positionDamper = 0;
			drv.maximumForce = 1000;
			joint_pos.xDrive = joint_pos.yDrive = joint_pos.zDrive = drv;
			joint_pos.targetVelocity = Vector3.zero;
			joint_pos.connectedBody = partRoboticArm.Rigidbody;
		}


		//*********************************************************
		// Update
		//*********************************************************

		public override void OnUpdate(){
			//if grabbed part was destroyed
			if (isMagnetized && ref_hitPart == null) {
				magnetize.set = false;
				magnetize.Out();
				isMagnetized = false;
				if(!isMainConnection){
					isMainConnection = false;
					DestroyStaticJoint();
				}else{
					isMainConnection = false;
					DestroyMultiJoint();
				}
			}
			//debug = Util.Cross(Color.red);
			//debug2 = Util.Cross(Color.yellow);
			//debug2.transform.position = go_om.transform.position;
			/*if(isMagnetized)
				debug.transform.position = go_joint.transform.position;*/

			//camPos = partRoboticArm.armTip.position + partRoboticArm.armTip.up.normalized * 0.4f;

			isMagnetized = magnetize;
			if (magnetize.In()) {
				//print("BURAN magnetize.In()");
				//debug = Util.Cross(Color.red);
				//debug2 = Util.Cross(Color.yellow);

				ref_hitPart = hitPart;
				partRef = ref_hitPart.flightID.ToString();
				vesselRef = new DockedVesselInfo();
				vesselRef.name = ref_hitPart.vessel.vesselName;
				vesselRef.vesselType = ref_hitPart.vessel.vesselType;
				vesselRef.rootPartUId = ref_hitPart.vessel.rootPart.flightID;

				if(ref_hitPart.vessel != vessel){
					print("BURAN magnetize.In() other vessel "+ref_hitPart.name);
					Couple();
				}else{
					print("BURAN magnetize.In() same vessel "+ref_hitPart.name);
					CoupleOwnVessel();
				}
			}else if(magnetize.Out()){
				if(!isMainConnection){
					print("BURAN magnetize.Out() : dropping joint");
					DestroyStaticJoint();
				}else{
					print("BURAN magnetize.Out() : restoring vessel");
					/*print("BURAN ref_hitPart "+ref_hitPart.flightID);
					print("BURAN partRef "+partRef.ToString());
					print("BURAN vesselRef name "+vesselRef.name);
					print("BURAN vesselRef vesselType "+vesselRef.vesselType);
					print("BURAN vesselRef rootPartUId "+vesselRef.rootPartUId);*/
					/*isMainConnection = false;
					DestroyMultiJoint();
					Undock(this, ref_hitPart);
					partRef = "-1";
					partRefs = null;*/
					Decouple();
				}
			}
		}


		#region statics

		static public void Undock(LazorRoboticArm arm, Part undockPart){
			undockPart.Undock(arm.vesselRef);

			//search de-grabbed vessel for another arm connected to this vessel.
			using (List<uint>.Enumerator p = arm.partRefs.GetEnumerator())
				while (p.MoveNext()){
					Part part = undockPart.vessel[p.Current];
					if(part != null && part.rigidbody != null){
						Joint[] js = part.rigidbody.GetComponents<Joint>();
						foreach(Joint j in js){
							Part pra = j.connectedBody.GetComponent<Part>();
							if(pra is PartRoboticArm && pra.vessel == arm.vessel){
								LazorRoboticArm lra = j.connectedBody.GetComponent<Part>().Modules[(pra as PartRoboticArm).robotArm.control.moduleName] as LazorRoboticArm;
								lra.DestroyStaticJoint();
								lra.Couple();
								return;
							}
						}
					}
				}
		}

		static private PartRoboticArm FindMainArmOnPart(Part part) {
			if(part.rigidbody != null){
				Joint[] js = part.rigidbody.GetComponents<Joint>();
				foreach(Joint j in js){
					if(j.connectedBody.name == "Robotic Joint"){
						PartRoboticArm robotarmPart = j.connectedBody.GetComponent<ConfigurableJoint>().connectedBody.GetComponent<Part>() as PartRoboticArm;
						if(robotarmPart.robotArm.isMainConnection)
							return robotarmPart;
					}
				}
			}
			return null;
		}

		//searches through all children for a part that's grabbed by a MAIN robotic arm, and returns the arm
		static public PartRoboticArm FindMainArmInHierarchy(Part part, bool up = false) {
			PartRoboticArm arm = FindMainArmOnPart(part);
			if(arm != null)
				return arm;
			if(up) {
				if(part.parent != null){
					arm = FindMainArmInHierarchy(part.parent, true);
					if(arm != null)
						return arm;
				}
			} else {
				foreach(Part child in part.children) {
					arm = FindMainArmInHierarchy(child);
					if(arm != null)
						return arm;
				}
			}
			return null;
		}

		static public bool FindFlightIdInHierarchy(uint flightID, Part part, bool up = false) {
			if(part.flightID == flightID)
				return true;
			if(up) {
				if(part.parent != null && FindFlightIdInHierarchy(flightID, part.parent, true))
					return true;
			} else {
				foreach(Part child in part.children)
					if(FindFlightIdInHierarchy(flightID, child))
						return true;
			}
			return false;
		}

		//decouples part from vessel but not robotarm
		static public bool DecoupleHeldPart(Part part) {
			PartRoboticArm robotarmPart = FindMainArmInHierarchy(part, true);
			if(robotarmPart == null) {
				part.decouple();
				return true;
			}else{ 
				if(FindMainArmOnPart(part) != null)
					return false;
				if(FindFlightIdInHierarchy(robotarmPart.robotArm.vesselRef.rootPartUId, part)) //we need to undock the part because the vessel the arm grabbed, has it's rootpart in the hierarchy we are decoupling
					Undock (robotarmPart.robotArm, part);
				else
					part.decouple();
				return true;
			}
		}

		static public Part GetDominantPart(Part p1, Part p2) {
			Vessel v1 = p1.vessel;
			Vessel v2 = p2.vessel;
			bool b1 = FindMainArmInHierarchy(p1, true) != null;
			bool b2 = FindMainArmInHierarchy(p2, true) != null;
			if(b1 && !b2)
				return p1;
			else if(!b1 && b2)
				return p2;
			else {
				if(Mathf.Abs(v1.GetTotalMass() - v2.GetTotalMass()) < 0.01f)
					return p1;
				else if(v1.GetTotalMass() > v2.GetTotalMass())
					return p1;
				else
					return p2;
			}
		}

		#endregion




		//Searches a vessel for connected robotic arms.
		//assuming all arms are static,
		//couples the new vessel to the arm - i.e. go from static to multi joint
		private void SearchForArmsOnDisconnectedVessel() {
			using (List<Part>.Enumerator p = ref_hitPart.vessel.parts.GetEnumerator())
				while (p.MoveNext()){
					Part part = p.Current;
					if(part != null && part.rigidbody != null){
						Joint[] js = part.rigidbody.GetComponents<Joint>();
						foreach(Joint j in js){
							Part pra = j.connectedBody.GetComponent<Part>();
							if(pra is PartRoboticArm && pra.vessel == vessel){
								print("static to dynamic joint");
								LazorRoboticArm lra = j.connectedBody.GetComponent<Part>().Modules[control.moduleName] as LazorRoboticArm;
								//print("1");
								lra.DestroyStaticJoint();
								//print("2");
								lra.Couple();
								//print("3");
								return;
							}
						}
					}
				}
		}

		//if main connection, drop magnetized object.
		private void Decouple() {
			MonoBehaviour.print("Decouple");
			magnetize.set = false;
			magnetize.Out();
			isMagnetized = false;

			isMainConnection = false;
			DestroyMultiJoint();
			Undock(this, ref_hitPart);
			partRef = "-1";
			partRefs = null;
		}

		//IMPORTANT: remember the case where magnetized but not coupled...maybe =p
		//also fix onload (magnetized grayed out)
		public void Couple(){
			if(partRefs == null)
				partRefs = new PartRefs();
			using(List<Part>.Enumerator p = ref_hitPart.vessel.parts.GetEnumerator())
				while(p.MoveNext())
					partRefs.Add(p.Current.flightID);

			isMainConnection = true;

			this.vessel.SetRotation(this.vessel.transform.rotation);

			ref_hitPart.vessel.SetRotation(ref_hitPart.vessel.transform.rotation);

			ref_hitPart.vessel.SetPosition(ref_hitPart.vessel.transform.position, true);

			part.attachMethod = AttachNodeMethod.FIXED_JOINT;
			Vessel vessel = this.vessel;

			ref_hitPart.Couple(part);

			FlightGlobals.ForceSetActiveVessel (vessel);

			vessel.MakeActive (); 

			//FlightGlobals.ForceSetActiveVessel(vessel); //fixed the bug with scenario (canadarm buran surface multi) but drops grabbed if you lift the lifting vessel
			//vessel.MakeActive(); //bug with scenario (canadarm buran surface multi)

			FlightInputHandler.SetNeutralControls();
			//NotificationManager.OnVesselChange_Fire(vessel);
			CreateMultiJoint();

			//debug2.transform.position = go_om.transform.position;
		}

		private void CoupleOwnVessel(){
			isMainConnection = false;
			CreateStaticJoint();
		}

		public void DestroyStaticJoint(){
			UnityEngine.Object.Destroy(joint_static);
		}

		public void CreateStaticJoint(){
			joint_static = ref_hitPart.gameObject.AddComponent<FixedJoint>();
			joint_static.breakForce = float.PositiveInfinity;
			joint_static.breakTorque = float.PositiveInfinity;
			joint_static.connectedBody = partRoboticArm.Rigidbody;
		}

		private void DestroyMultiJoint(){
			UnityEngine.Object.Destroy(joint_rotate);
			UnityEngine.Object.Destroy(joint_pos);
			UnityEngine.Object.Destroy(go_om);
			UnityEngine.Object.Destroy(go_sm);
			UnityEngine.Object.Destroy(go_joint);
		}

		private void CreateMultiJoint(){
			SetupDummy_ObjectMarker();
			SetupDummy_StationaryMarker();
			SetupDummy_Joint();
			SetupJointPosition();
			SetupJointRotation();
		}

		public void MakeStatic() {
			isMainConnection = false;
			DestroyMultiJoint();
			CreateStaticJoint();
		}

		public void MakeDynamic() {
			isMainConnection = true;
			DestroyStaticJoint();
			CreateMultiJoint();
		}

		private void LockJoint(){
			DestroyMultiJoint ();
			CreateStaticJoint();
			//ref_hitPart.attachJoint = joint_static;
		}

		private void UnlockJoint(){
			UnityEngine.Object.Destroy(joint_static);
			CreateMultiJoint();
		}

		public void onPack(){
		}

		public void onUnpack(){

			if(startFromSaveMagnetized) {
				//print("BURAN Initialize() isMagnetized");
				//ref_hitPart = vessel[uint.Parse(partRef)];

				/*print("BURAN Initialize() partRef "+partRef.ToString());
				print("BURAN Initialize() ref_hitPart.name "+ref_hitPart.name);
				print("BURAN Initialize() ref_hitPart.flightID "+ref_hitPart.flightID.ToString());
				print("BURAN Initialize() vesselRef name "+vesselRef.name);
				print("BURAN Initialize() vesselRef vesselType "+vesselRef.vesselType);
				print("BURAN Initialize() vesselRef rootPartUId "+vesselRef.rootPartUId);*/

				if(isMainConnection){
					CreateMultiJoint();
				}else{
					CoupleOwnVessel();
				}

				startFromSaveMagnetized = false;
			}else if(go_joint == null && magnetize && isMainConnection) {
				UnlockJoint();
			}
		}

		private DoubleSwitch warpToggle;
		private void onTimeWarpChangeRate(){
			/*print ("TIMEWARP "+ 
			       TimeWarp.CurrentRate+
			       " "+TimeWarp.CurrentRateIndex+
			       " "+TimeWarp.deltaTime+
			       " "+TimeWarp.fixedDeltaTime+
			       " "+TimeWarp.WarpMode.ToString()+
			       "");*/
			warpToggle.set = magnetize && isMainConnection && TimeWarp.CurrentRate == 1 && TimeWarp.CurrentRateIndex == 1;

			if (warpToggle.In ()) {
				LockJoint();
			} else if(warpToggle.Out ()) {}
		}

		//*********************************************************
		// FixedUpdate
		//*********************************************************

		public override void OnFixedUpdate(){
			/*if(magnetize)
				MonoBehaviour.print(" jo l "+joint_rotate.transform.localRotation.eulerAngles+
				                    " jo r "+joint_rotate.transform.rotation.eulerAngles+
				                    " go l "+go_om.transform.localRotation.eulerAngles+
				                    " go r "+go_om.transform.rotation.eulerAngles+
									// go "+go.transform.position+
				                    //" goL "+go.transform.localPosition+
				                    //" joP "+joint_swap.transform.position+
				                    //" joL "+joint_swap.transform.localPosition+
				                    //" joJ "+joint_swap.targetPosition+
				                    //" re "+ref_hitPart.rigidbody.transform.position+
				                    //" re "+ref_hitPart.rigidbody.transform.localPosition+
									"");*/
			if (isMagnetized) {
				//check if arm was connected to self and that part has been disconnected.
				if (!isMainConnection && vessel != ref_hitPart.vessel){
					print ("SearchForArmsOnDisconnectedVessel");
					SearchForArmsOnDisconnectedVessel();
				}
				//make static when connected vessel
				/*if (isMainConnection && vessel == ref_hitPart.vessel){
					print ("SearchForArmsOnDisconnectedVessel");
					SearchForArmsOnDisconnectedVessel ();
					return;
				}*/
				//drain energy
				part.RequestResource("ElectricCharge", 2.5d * Time.deltaTime * TimeWarp.CurrentRate);
			}
			//

			RaycastHit hit;
			if(magnet && !magnetize && Physics.Raycast(control.cameraPosition, control.cameraForward, out hit, control.magnetDistance)) {
				hitBody = hit.collider.attachedRigidbody;
				if(connectable = hitBody != null)
					hitPart = hitBody.GetComponent<Part>();
			} else
				connectable = false;
			if(cameraOn) {
				part.RequestResource("ElectricCharge", 0.1d * Time.deltaTime * TimeWarp.CurrentRate);
				if(control.Fixed()){
				}else
					moving.set = false;
				if(moving.In()){
					partRoboticArm.lazorSound.PlayLoop(SoundType.SERVO);
					partRoboticArm.lazorSound.SetVolumeLoop(0.3f * GameSettings.SHIP_VOLUME);
				}else if(moving.Out()){
					partRoboticArm.lazorSound.StopLoop();
				}
			}
		}







		public void CamGUI() {
			//print(Time.frameCount);
			UnityEngine.GUI.skin = Util.LazorSkin;
			if (!FlightDriver.Pause && cameraOn && !startFromSaveMagnetized){
				windowPos = Util.ClampToScreen(GUILayout.Window(GetInstanceID(), windowPos, WindowGUI, control.title));
				windowPos.width = camWidth;
				windowPosition.x = windowPos.x;
				windowPosition.y = windowPos.y;
			}
			if (Event.current.type == EventType.Repaint)
				windowPos.height = 100;
		}

		public void WindowGUI(int windowID) {
			if(GUI.Button(new Rect(windowPos.width - 20, 0, 20, 20), "", Util.Button_header_close)) { //red close button
				cameraOn = false;
				Events["ActivateLazorCamera"].active = true;
				Events["DeactivateLazorCamera"].active = false;
			}
			/*if (partRoboticArm == null) {  //JUST USE THE TRY CLAUSE
				cameraOn = false;
				return;
			}*/

			//try {
			lazorGui.CameraGui(CameraModes.Normal, out camWidth, out camHeight, control.cameraForward, control.cameraPosition, control.cameraNormal, control.cameraNormal, null);

			//blinking green square
			if(connectable && !magnetize) {
				if(!blinking) {
					blink++;
					if(blink > 10)
						blinking = true;
				} else {
					blink--;
					if(blink < 0)
						blinking = false;
					GL.PushMatrix();
					material.SetPass(0);
					GL.LoadPixelMatrix();
					GL.Begin(GL.LINES);
					GL.Color(blinkColor);
					float yf = Screen.height - windowPos.y - lazorGui.RectBounds.height - lazorGui.RectBounds.y + 11;
					float xf = windowPos.x + lazorGui.RectBounds.x + 1;
					GL.Vertex3(xf, yf, 0); //bot left
					GL.Vertex3(xf + lazorGui.RectBounds.width - 12, yf, 0); //bot right
					GL.Vertex3(xf, yf, 0); //bot left
					GL.Vertex3(xf, yf + lazorGui.RectBounds.height - 10, 0); //top left
					GL.Vertex3(xf + lazorGui.RectBounds.width - 12, yf, 0); //bot right
					GL.Vertex3(xf + lazorGui.RectBounds.width - 12, yf + lazorGui.RectBounds.height - 10, 0); //top right
					GL.Vertex3(xf, yf + lazorGui.RectBounds.height - 10, 0); //top left
					GL.Vertex3(xf + lazorGui.RectBounds.width - 12, yf + lazorGui.RectBounds.height - 10, 0); //top right
					GL.End();
					GL.PopMatrix();
				}
			}
			if(connectable && lazorGui.RectBounds.Contains(Event.current.mousePosition) && Input.GetMouseButtonDown(0))
				magnetize.set = true;
			GUILayout.BeginHorizontal();
			GUILayout.Label("Zoom");
			lazorGui.sliderZoom = GUILayout.HorizontalSlider(lazorGui.sliderZoom, 0, 12);
			GUILayout.EndHorizontal();

			GUILayout.Space(5);

			bool enableGui = TimeWarp.CurrentRate == 1;

			GUI.enabled = enableGui && (isMainConnection || !magnetize);
			control.Gui ();

			GUI.enabled = enableGui;
			GUILayout.BeginHorizontal(Util.Space_Compact);
			magnet = GUILayout.Toggle(magnet, "Magnet");
			if(magnet) {
				GUILayout.Space(5);
				GUI.enabled = enableGui && (magnetize || connectable);
				magnetize.set = GUILayout.Toggle(magnetize, "Magnetize");
			}
			GUILayout.EndHorizontal();

			GUI.enabled = true;

			if(magnet) {
				GUILayout.Space(5);
				GUILayout.BeginVertical(Util.ListItemBG_compact);
				if(magnetize){
					GUILayout.Label("Grabbed: "+ref_hitPart.name, Util.StdLabelSmallCustom(Color.green));
				}else{
					if(connectable){
						GUILayout.Label("Object in range", Util.StdLabelSmallCustom(Color.green));
						GUILayout.Label(hitPart.name, Util.StdLabelSmallCustom(Color.green));
					}else
						GUILayout.Label("Nothing in range ", Util.StdLabelSmallCustom(Color.gray));
				}
				GUILayout.EndVertical();
			} else
				magnetize.set = false;
			GUI.enabled = true;



			//DEBUG STUFF ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::


			/*
			GUILayout.Space(10);

			if(GUILayout.Button("DOCK LISTENER")) {
				NotificationManager.OnVesselDock_Add(delegate(Vessel v) {
					print (v.vesselName);
				});
			}
			*/

			/*
			if(GUILayout.Button("WARP LISTENER")) {
				Messenger.AddListener ("onTimeWarpChangeRate", onTimeWarpChangeRate);
			}

			GUILayout.Space(5);*/

			//works only for immediate warp
			/*if(GUILayout.Button("WARP IN")) {
				joint_pos.xMotion = ConfigurableJointMotion.Locked;
				joint_pos.yMotion = ConfigurableJointMotion.Locked;
				joint_pos.zMotion = ConfigurableJointMotion.Locked;
				joint_pos.angularXMotion = ConfigurableJointMotion.Locked;
				joint_pos.angularYMotion = ConfigurableJointMotion.Locked;
				joint_pos.angularZMotion = ConfigurableJointMotion.Locked;
				JointDrive drv = new JointDrive();
				drv.mode = JointDriveMode.Position;
				drv.positionSpring = 0;
				drv.positionDamper = 0;
				drv.maximumForce = 0;
				joint_pos.xDrive = joint_pos.yDrive = joint_pos.zDrive = drv;
				joint_rotate.xMotion = ConfigurableJointMotion.Locked;
				joint_rotate.yMotion = ConfigurableJointMotion.Locked;
				joint_rotate.zMotion = ConfigurableJointMotion.Locked;
				joint_rotate.angularXMotion = ConfigurableJointMotion.Locked;
				joint_rotate.angularYMotion = ConfigurableJointMotion.Locked;
				joint_rotate.angularZMotion = ConfigurableJointMotion.Locked;
				joint_rotate.rotationDriveMode = RotationDriveMode.XYAndZ;
				JointDrive drv_rot = new JointDrive();
				drv_rot.mode = JointDriveMode.Position;
				drv_rot.positionSpring = 0;
				drv_rot.positionDamper = 0;
				drv_rot.maximumForce = 0;
				joint_rotate.slerpDrive = joint_rotate.angularXDrive = joint_rotate.angularYZDrive = drv_rot;

				go_joint.rigidbody.isKinematic = true;
				TimeWarp.SetRate(1, true);
			}*/

			//works only for immediate warp
			/*if (GUILayout.Button ("WARP OUT")) {
				joint_pos.xMotion = ConfigurableJointMotion.Free;
				joint_pos.yMotion = ConfigurableJointMotion.Free;
				joint_pos.zMotion = ConfigurableJointMotion.Free;
				joint_pos.angularXMotion = ConfigurableJointMotion.Locked;
				joint_pos.angularYMotion = ConfigurableJointMotion.Locked;
				joint_pos.angularZMotion = ConfigurableJointMotion.Locked;
				JointDrive drv = new JointDrive();
				drv.mode = JointDriveMode.Position;
				drv.positionSpring = 100000;
				drv.positionDamper = 0;
				drv.maximumForce = 1000;
				joint_pos.xDrive = joint_pos.yDrive = joint_pos.zDrive = drv;
				joint_rotate.xMotion = ConfigurableJointMotion.Locked;
				joint_rotate.yMotion = ConfigurableJointMotion.Locked;
				joint_rotate.zMotion = ConfigurableJointMotion.Locked;
				joint_rotate.angularXMotion = ConfigurableJointMotion.Free;
				joint_rotate.angularYMotion = ConfigurableJointMotion.Free;
				joint_rotate.angularZMotion = ConfigurableJointMotion.Free;
				joint_rotate.rotationDriveMode = RotationDriveMode.XYAndZ;
				JointDrive drv_rot = new JointDrive();
				drv_rot.mode = JointDriveMode.Position;
				drv_rot.positionSpring = 10;
				drv_rot.positionDamper = 0;
				drv_rot.maximumForce = 1;
				joint_rotate.slerpDrive = joint_rotate.angularXDrive = joint_rotate.angularYZDrive = drv_rot;

				go_joint.transform.position = ref_hitPart.transform.position;
				go_joint.rigidbody.isKinematic = false;
				TimeWarp.SetRate(0, true);
			}*/

			//GUILayout.Space(5);

			/*if(GUILayout.Button("seudo couple")) {
				print("seudo couple");
				ref_hitPart = hitPart;
				vesselRef = new DockedVesselInfo();
				vesselRef.name = ref_hitPart.vessel.vesselName;
				vesselRef.vesselType = ref_hitPart.vessel.vesselType;
				vesselRef.rootPartUId = ref_hitPart.vessel.rootPart.flightID;
				vessel.SetRotation(vessel.transform.rotation);
				ref_hitPart.vessel.SetRotation(ref_hitPart.vessel.transform.rotation);
				ref_hitPart.vessel.SetPosition(ref_hitPart.vessel.transform.position, true);
				part.attachMethod = AttachNodeMethod.FIXED_JOINT;
				ref_hitPart.Couple(part);
				vessel.MakeActive();
				FlightInputHandler.SetNeutralControls();
				NotificationManager.OnVesselChange_Fire(vessel);
			}

			if(GUILayout.Button("seudo joint attach")) {
				print("seudo joint attach");
				CreateMultiJoint();
			}

			if(GUILayout.Button("seudo joint remove")) {
				print("seudo joint remove");
				DestroyMultiJoint();
			}
			
			if(GUILayout.Button("seudo undock")) {
				print("seudo undock");
				ref_hitPart.Undock(vesselRef);
			}*/




			/*
			GUILayout.BeginHorizontal();
			if(GUILayout.RepeatButton("up")){
				joint_swap.targetPosition = Vector3.zero;
				joint_swap.anchor += Vector3.up * 0.05f;//joint_swap.targetPosition += Vector3.up * 0.05f;
			}if(GUILayout.RepeatButton("down")){
				joint_swap.targetPosition = Vector3.zero;
				joint_swap.anchor -= Vector3.up * 0.05f;
			}
			GUILayout.EndHorizontal();
			GUILayout.BeginHorizontal();
			if(GUILayout.RepeatButton("right"))
				joint_swap.anchor += Vector3.right * 0.05f;

			if(GUILayout.RepeatButton("left"))
				joint_swap.anchor -= Vector3.right * 0.05f;

			GUILayout.EndHorizontal();
			GUILayout.BeginHorizontal();
			if(GUILayout.RepeatButton("fwd"))
				joint_swap.anchor += Vector3.forward * 0.05f;

			if(GUILayout.RepeatButton("back"))
				joint_swap.anchor -= Vector3.forward * 0.05f;
			GUILayout.EndHorizontal();



			GUILayout.BeginHorizontal();
			if(GUILayout.Button("R up"))
				joint_swap.targetRotation *= Quaternion.Euler (0, 5, 0);
			if(GUILayout.Button("R down"))
				joint_swap.targetRotation *= Quaternion.Euler (0, 355, 0);

			GUILayout.EndHorizontal();
			GUILayout.BeginHorizontal();
			if(GUILayout.Button("R right"))
				joint_swap.targetRotation *= Quaternion.Euler (5, 0, 0);
			if(GUILayout.Button("R left"))
				joint_swap.targetRotation *= Quaternion.Euler (355, 0, 0);

			GUILayout.EndHorizontal();
			GUILayout.BeginHorizontal();
			if(GUILayout.Button("R fwd"))
				joint_swap.targetRotation *= Quaternion.Euler (0, 0, 5);
			if(GUILayout.Button("R back"))
				joint_swap.targetRotation *= Quaternion.Euler (0, 0, 355);
			GUILayout.EndHorizontal();
			*/

			/*} catch (Exception) { 
			cameraOn = false;
			partRoboticArm.lazorSound.StopLoop();
		}*/

		GUI.DragWindow(new Rect(0, 0, windowPos.width, 30));
	}

}


//Made by mstevenson : http://forum.unity3d.com/threads/8919-Quaternion-Wizardry-for-ConfigurableJoint
public static class ConfigurableJointExtensions {

	/// <summary>
	/// Sets a joint's targetRotation to match a given local rotation.
	/// The joint transform's local rotation must be cached on Start and passed into this method.
	/// </summary>
	public static void SetTargetRotationLocal (this ConfigurableJoint joint, Quaternion targetLocalRotation, Quaternion startLocalRotation){
		if (joint.configuredInWorldSpace)
			Debug.LogError ("SetTargetRotationLocal should not be used with joints that are configured in world space. For world space joints, use SetTargetRotation.", joint);
		SetTargetRotationInternal (joint, targetLocalRotation, startLocalRotation, Space.Self);
	}

	/// <summary>
	/// Sets a joint's targetRotation to match a given world rotation.
	/// The joint transform's world rotation must be cached on Start and passed into this method.
	/// </summary>
	public static void SetTargetRotation (this ConfigurableJoint joint, Quaternion targetWorldRotation, Quaternion startWorldRotation){
		if (!joint.configuredInWorldSpace)
			Debug.LogError ("SetTargetRotation must be used with joints that are configured in world space. For local space joints, use SetTargetRotationLocal.", joint);
		SetTargetRotationInternal (joint, targetWorldRotation, startWorldRotation, Space.World);
	}

	static void SetTargetRotationInternal (ConfigurableJoint joint, Quaternion targetRotation, Quaternion startRotation, Space space){
		// Calculate the rotation expressed by the joint's axis and secondary axis
		var right = joint.axis;
		var forward = Vector3.Cross (joint.axis, joint.secondaryAxis).normalized;
		var up = Vector3.Cross (forward, right).normalized;
		Quaternion worldToJointSpace = Quaternion.LookRotation (forward, up);
		// Transform into world space
		Quaternion resultRotation = Quaternion.Inverse (worldToJointSpace);
		// Counter-rotate and apply the new local rotation.
		// Joint space is the inverse of world space, so we need to invert our value
		if (space == Space.World)
			resultRotation *= startRotation * Quaternion.Inverse (targetRotation);
		else
			resultRotation *= Quaternion.Inverse (targetRotation) * startRotation;
		// Transform back into joint space
		resultRotation *= worldToJointSpace;
		// Set target rotation to our newly calculated rotation
		joint.targetRotation = resultRotation;
	}
}




[System.Serializable]
public class PartRefs : IConfigNode {
	[SerializeField]
	private List<uint> partRefs;

	public PartRefs(){
		partRefs = new List<uint>();
	}

	public void Add(uint r){
		partRefs.Add(r);
	}

	public List<uint>.Enumerator GetEnumerator(){
		return partRefs.GetEnumerator();
	}

	public void Load(ConfigNode node){
		string[] values = node.GetValues("part");
		for (int i = 0; i < values.Length; i++)
			Add(uint.Parse(values[i]));
	}

	public void Save(ConfigNode node){
		for (int i = 0; i < partRefs.Count; i++)
			node.AddValue("part", partRefs[i].ToString());
	}
}



public class JointBreakCallback : MonoBehaviour {
	Axtion action;
	public void AddCallback(Axtion action) {
		this.action = action;
	}
	public void OnJointBreak(float breakForce) {
		print("JOINTBREAK");
		action();
		Destroy(this);
	}
}


}

