/* Author: Jim-Kyrre Benjaminsen AKA Romfarer
 * License: This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivs 3.0 Unported License. 
 * http://creativecommons.org/licenses/by-nc-nd/3.0/
*/
//#define LAZORDEBUG


using System.Collections;



using System;
using System.Linq;
using System.Collections.Generic;
using UnityEngine;
using System.Text.RegularExpressions;



namespace Romfarer
{

	[KSPAddon(KSPAddon.Startup.MainMenu, true)]
	public class RuntimeLogic : MonoBehaviour {

		public static RuntimeLogic Instance;
		public static string ApplicationPath = "Romfarer/";
		private bool inFlight { get { return HighLogic.LoadedSceneIsFlight && FlightGlobals.fetch != null && FlightGlobals.ready; } }

		//DEBUG
		private bool DEBUG_gui_on;
		private bool DEBUG_hierarchy;
		private bool DEBUG_log;
		private DoubleSwitch DEBUG_joints;
		private List<Joint> joints = new List<Joint>();
		private Rect windowPos_DEBUG = new Rect(160,0,500,100);
		private Vector2 scrollArea = Vector2.zero;

		private bool showSpaceCenterGUI;
		public Texture applicationLauncherIcon;

		public void Awake() {
			if (Instance != null) {
				GameObject.DestroyImmediate (gameObject);
				return;
			}
			Instance = this;
			UnityEngine.GameObject.DontDestroyOnLoad(gameObject);
		}

		public void Start() {
			//find the application sub path within the GameData folder
			string platform = "Windows";
			string subPath = AssemblyLoader.GetPathByType(typeof(RuntimeLogic)).Substring(Application.dataPath.Length + 1);
			string pattern = "(.*\\n?)(?=PluginData.Romfarer)";
			Match match = Regex.Match(subPath, pattern, RegexOptions.IgnoreCase);

			if(Application.platform != RuntimePlatform.OSXPlayer) {
				if(match.Success)
					ApplicationPath = match.ToString().Replace("\\", "/");                                                  
			} else
				platform = "OSX";
			if(Application.platform == RuntimePlatform.LinuxPlayer)
				platform = "Linux";

			//unity debugging
			if(Application.isEditor)
				ApplicationPath = "Romfarer/";

			MonoBehaviour.print("LazorSystem RuntimeLogic Started (" + platform + ")\n" +
				"systempath:: " + ApplicationPath + "\n" +
				"path:: " + AssemblyLoader.GetPathByType (typeof(RuntimeLogic)) + "\n" +
				"datapath::    " + Application.dataPath + "\n" +
				"subpath:: " + subPath );

			//add LazorEVA partmodule to EVA prefab
			foreach (Part p in Resources.FindObjectsOfTypeAll(typeof(Part)))
				if(p.name == "kerbalEVA_RD")
					p.gameObject.AddComponent<LazorEVA>();

			applicationLauncherIcon = Util.LoadTexture(38, 38, "applauncher_lazoricon");

			GameEvents.onVesselChange.Add(OnVesselChange);
			GameEvents.onGameSceneLoadRequested.Add(GameEvents_onGameSceneLoadRequested);
			GameEvents.onPartRemove.Add(GameEvents_onPartRemove);
			GameEvents.onPartCouple.Add(GameEvents_onPartCouple);
			GameEvents.onTimeWarpRateChanged.Add(GameEvents_onTimeWarpRateChanged);
			GameEvents.onFlightReady.Add(GameEvents_onFlightReady);

			//RuntimeLogic should never be destroyed, but in case it does, these events should not be unsubscribed
			//SubscribeToApplauncher();
			//GameEvents.onGUIApplicationLauncherDestroyed.Add(SubscribeToApplauncher);
			GameEvents.onGUIApplicationLauncherReady.Add(OnApplauncherReady);

			//GameEvents.OnTechnologyResearched.Add(OnTechnologyResearched);
		}

		/*private void SubscribeToApplauncher(){
			GameEvents.onGUIApplicationLauncherReady.Add(AddConfigToApplauncher);
			GameEvents.onGUIApplicationLauncherReady.Add(AddLazorSystemPartsToApplauncher);
		}*/
			
		public void OnDestroy() {
			GameEvents.onVesselChange.Remove(OnVesselChange);
			GameEvents.onGameSceneLoadRequested.Remove(GameEvents_onGameSceneLoadRequested);
			GameEvents.onPartRemove.Remove(GameEvents_onPartRemove);
			GameEvents.onPartCouple.Remove(GameEvents_onPartCouple);
			GameEvents.onTimeWarpRateChanged.Remove(GameEvents_onTimeWarpRateChanged);
			GameEvents.onFlightReady.Remove(GameEvents_onFlightReady);
			//GameEvents.OnTechnologyResearched.Remove(OnTechnologyResearched);
		}

		public void Update(){
			if(inFlight) {
				LazorGlobals.LastVessel = FlightGlobals.ActiveVessel;
			}
		}

		#region Application Launcher workaround

		/*private bool overrideAddLazorButton;
		private void OnTechnologyResearched(GameEvents.HostTargetAction<RDTech, RDTech.OperationResult> action){
			print("LazorSystem Unlocked and Enabled");
			if(action.host.techID.Equals("experimentalElectrics") && action.target == RDTech.OperationResult.Successful && launcherButton == null) {
				overrideAddLazorButton = true;
			}
		}*/

		private void OnApplauncherReady(){
			if(HighLogic.LoadedScene == GameScenes.SPACECENTER)
				AddConfigToApplauncher();
			//In career mode only add lazor button to application launcher if the tech is unlocked
			else if(HighLogic.LoadedScene == GameScenes.FLIGHT){
				if(HighLogic.CurrentGame.Mode != Game.Modes.CAREER) {
					if(launcherButton == null)
						AddLazorSystemPartsToApplauncher();
				} else {
					if(launcherButton == null)
						StartCoroutine(DelayedAdd());
				}
			}
		}

		private void GameEvents_onGameSceneLoadRequested(GameScenes scene) {
			Util.ResetGUI();

			if(ApplicationLauncher.Instance != null){
				if(launcherButtonConfig != null){
					ApplicationLauncher.Instance.RemoveModApplication(launcherButtonConfig);
					launcherButtonConfig = null;
				}
				if(launcherButton != null) {
					ApplicationLauncher.Instance.RemoveModApplication(launcherButton);
					launcherButton = null;
				}
			}
		}

		private IEnumerator DelayedAdd(){
			yield return null;
			if(HighLogic.CurrentGame.Mode == Game.Modes.CAREER && ResearchAndDevelopment.GetTechnologyState("experimentalElectrics") == RDTech.State.Available) 
				AddLazorSystemPartsToApplauncher();
			print("RuntimeLogic Enabling LazorButton " + (HighLogic.LoadedScene == GameScenes.FLIGHT) + " " + (HighLogic.CurrentGame.Mode == Game.Modes.CAREER) + " " + (ResearchAndDevelopment.GetTechnologyState("experimentalElectrics") == RDTech.State.Available) + " " + ResearchAndDevelopment.GetTechnologyState("experimentalElectrics").ToString() + " " + (ResearchAndDevelopment.Instance == null));

		}

		#endregion

		#region Application Launcher Config Screen

		private ApplicationLauncherButton launcherButtonConfig;
		private bool showSpaceCenterGUIwithApplauncher;

		private void AddConfigToApplauncher(){
			launcherButtonConfig = ApplicationLauncher.Instance.AddModApplication(ShowConfigUI, HideConfigUI, null, null, ShowConfigUIWithLauncher, HideConfigUIWithLauncher, 
											ApplicationLauncher.AppScenes.SPACECENTER, applicationLauncherIcon);
			ApplicationLauncher.Instance.AddOnShowCallback(ShowConfigUIWithLauncher);
			ApplicationLauncher.Instance.AddOnHideCallback(HideConfigUIWithLauncher);
			HideConfigUI();
			//GameEvents.onGUIApplicationLauncherReady.Remove(AddConfigToApplauncher);
		}

		private void ShowConfigUIWithLauncher(){
			if(showSpaceCenterGUIwithApplauncher)
				showSpaceCenterGUI = true;
		}
		private void HideConfigUIWithLauncher(){
			showSpaceCenterGUI = false;
		}
		private void ShowConfigUI(){
			showSpaceCenterGUI = true;
			showSpaceCenterGUIwithApplauncher = true;
		}
		private void HideConfigUI(){
			showSpaceCenterGUI = false;
			showSpaceCenterGUIwithApplauncher = false;
		}

		#endregion

		#region Application Launcher LAZOR SYSTEM Screen

		//For lazor system on vessels
		private ApplicationLauncherButton launcherButton;

		private void AddLazorSystemPartsToApplauncher(){
			launcherButton = ApplicationLauncher.Instance.AddModApplication(ShowLazorUI, HideLazorUI, null, null, null, null, 
				ApplicationLauncher.AppScenes.FLIGHT | ApplicationLauncher.AppScenes.MAPVIEW, applicationLauncherIcon);
			//ApplicationLauncher.Instance.AddOnShowCallback(ShowLazorUI);
			//ApplicationLauncher.Instance.AddOnHideCallback(HideLazorUI);
			//GameEvents.onGUIApplicationLauncherReady.Remove(AddLazorSystemPartsToApplauncher);
		}

		private void ShowLazorUI(){
			if(LazorSystemControllerGlobal.ActiveLazorSystem != null)
				LazorSystemControllerGlobal.ActiveLazorSystem.lazorOn.set = true;
		}

		private void HideLazorUI(){
			if(LazorSystemControllerGlobal.ActiveLazorSystem != null)
				LazorSystemControllerGlobal.ActiveLazorSystem.lazorOn.set = false;
		}

		public void AppLauncherButtonSetState(){
			if(LazorSystemControllerGlobal.ActiveVesselHasLazorSystem()) {
				launcherButton.Enable();
				if(LazorSystemControllerGlobal.ActiveLazorSystem.lazorOn)
					launcherButton.toggleButton.SetTrue(false);
				else
					launcherButton.toggleButton.SetFalse(false);
			}else
				launcherButton.Disable();
		}

		private IEnumerator WaitForLauncherAndSetButtonState(){
			while(this) {
				if(launcherButton == null)
					yield return null;
				else {
					AppLauncherButtonSetState();
					break;
				}
			}
		}

		public void AppLauncherButtonDisable(){
			if(launcherButton != null)
				launcherButton.Disable();
		}

		#endregion

		#region Events

		private void OnVesselChange(Vessel v = null){
			Log("LazorSystem : RuntimeLogic : OnVesselChange()", true);
			if (FlightGlobals.ActiveVessel.isEVA && LazorSystemControllerGlobal.LastVesselWithLazor != null && FlightGlobals.ActiveVessel.rootPart.Modules["LazorEVA"].Fields.GetValue<string>("myVessel") == "-1") {
				print("[Adding memory to " + FlightGlobals.ActiveVessel.vesselName + "]");
				PartModule mem = FlightGlobals.ActiveVessel.rootPart.Modules["LazorEVA"];
				mem.Fields.SetValue("myVessel", LazorSystemControllerGlobal.LastVesselWithLazor.id.ToString());
			}

			//add docking cam to all docking ports
			foreach (Part p in FlightGlobals.ActiveVessel.parts) {
				if (p.Modules.Contains("ModuleDockingNode")) {
					if (!p.Modules.Contains("LazorSystemDockingCamera")) {
						MonoBehaviour.print("[Adding docking camera to " + p.name + "]");
						PartModule pm = p.AddModule("LazorSystemDockingCamera");
					}
				}
			}

			//AppLauncherButtonSetState();
			StartCoroutine(WaitForLauncherAndSetButtonState());
		}

		public void GameEvents_onTimeWarpRateChanged() {
			if(TimeWarp.CurrentRate != 1)
				LazorGlobals.SetLoadDistance();
			else 
				LazorGlobals.SetLoadDistance(LazorGlobals.MAX_DISTANCE, LazorGlobals.VIEW_DISTANCE);
		}



		private void GameEvents_onPartCouple(GameEvents.FromToAction<Part, Part> action) {
			if(inFlight && action.from.vessel != action.to.vessel) {
				RuntimeLogic.Instance.Log("(f" + Time.frameCount + ") COUPLE " + action.from + " " + action.to, false);
				LazorSystemControllerGlobal.MergeLazorControllers(action.from.vessel, action.to.vessel);
			}
		}

		private void GameEvents_onPartRemove(GameEvents.HostTargetAction<Part, Part> action) {
			if(inFlight){
				RuntimeLogic.Instance.Log("(f" + Time.frameCount + ") REMOVE " + action.host + " " + action.target + " " + action.target.parent, false);
				LazorSystemControllerGlobal.EnableSystemLock (action.target.parent.vessel, (sys) => {return sys.sysLock_WaitDisconnect;});
			}
		}

		private void GameEvents_onFlightReady(){
			LazorGlobals.Initialize();
			LazorGlobals.SetLoadDistance(LazorGlobals.MAX_DISTANCE, LazorGlobals.VIEW_DISTANCE);

			AppLauncherButtonSetState();
		}
			
		#endregion

		#region oldstuff

		/*private void SetLoadDistance(float loadDistance, float unloadDistance, float landedPartPackTreshold, float partPackTreshold){
			if (LazorGlobals.GLOBALCONFIG_CurrentLoadDistance != LazorGlobals.LoadDistance.Default) {
				MonoBehaviour.print ("--== LazorSystem RuntimeLogic SetLoadDistance ==--");
				if (LazorGlobals.manager == null) {
					LazorGlobals.manager = (OrbitPhysicsManager)GameObject.FindObjectOfType (typeof(OrbitPhysicsManager));
					MonoBehaviour.print ("[LazorSystem RuntimeLogic : New physics manager loaded]");
				}
				MonoBehaviour.print ("LazorSystem RuntimeLogic : loadDistance " + loadDistance + ", unloadDistance " + unloadDistance + ", landedPartPackTreshold " + landedPartPackTreshold + ", partPackTreshold " + partPackTreshold);
				//MonoBehaviour.print ("--== orig: " + Vessel.loadDistance + " " + Vessel.unloadDistance + " " + LazorGlobals.manager.distantLandedPartPackThreshold + " " + LazorGlobals.manager.distantPartPackThreshold);
				Vessel.loadDistance = loadDistance; //2500
				Vessel.unloadDistance = unloadDistance; //2250
				LazorGlobals.manager.distantLandedPartPackThreshold = landedPartPackTreshold; //95000; //350
				LazorGlobals.manager.distantPartPackThreshold = partPackTreshold; //95000; //5000
				//floatingOrigin = (FloatingOrigin)GameObject.FindObjectOfType(typeof(FloatingOrigin));
	            //if(floatingOrigin!=null)
	            //    floatingOrigin.threshold = 1000000000000;
				//MonoBehaviour.print ("--== new: " + Vessel.loadDistance + " " + Vessel.unloadDistance + " " + LazorGlobals.manager.distantLandedPartPackThreshold + " " + LazorGlobals.manager.distantPartPackThreshold);
			}
		}*/

		#endregion

		#region GUI

		private bool settingsLoaded;

		private Rect windowPos_SpaceCenter = new Rect(0,0,150,10);
		private bool settings_loading;
		private bool settings_lazorsystem;

		/// <summary>
		/// Romfarer settings interface in space center
		/// </summary>
		public void OnGUI(){
			if (HighLogic.LoadedScene == GameScenes.SPACECENTER){// && WaitASecond()) {
				if (!settingsLoaded) {
					LazorGlobals.LoadConfig();
					settingsLoaded = true;
				}
				GUI.skin = Util.LazorSkin;
				if(showSpaceCenterGUI)
					windowPos_SpaceCenter = Util.ClampToScreen (GUILayout.Window (GetInstanceID(), windowPos_SpaceCenter, GUIwindow_SpaceCenter, "Lazor Powered"));

				if (Event.current.type == EventType.Repaint) {
					windowPos_SpaceCenter.height = 10;
					windowPos_SpaceCenter.x = 0;
				}
			} else {
				settings_loading = false;
				settings_lazorsystem = false;
			}
				
			#if LAZORDEBUG
			if(HighLogic.LoadedSceneIsFlight && FlightGlobals.ready && FlightGlobals.ActiveVessel != null && FlightGlobals.ActiveVessel.loaded) {
				GUI.skin = Util.LazorSkin;
				windowPos_DEBUG = GUILayout.Window(FlightGlobals.ActiveVessel.GetInstanceID(), windowPos_DEBUG, GUIwindow_DEBUG, "", GUIStyle.none);
				if (Event.current.type == EventType.Repaint)
					windowPos_DEBUG.height = 100;
			}
			#endif
		}
			
		private void GUIwindow_SpaceCenter(int windowID) {
			settings_loading = GUILayout.Toggle(settings_loading, "View distance", Util.Button_Toggle);
			if(settings_loading) {
				GUILayout.Space(5);
				//GUILayout.Label("Loading distance: ", Util.StdLabelSmallCustom(Color.yellow));
				LazorGlobals.GLOBALCONFIG_CurrentLoadDistance = (LazorGlobals.LoadDistance)GUILayout.SelectionGrid((int)LazorGlobals.GLOBALCONFIG_CurrentLoadDistance, Enum.GetNames(typeof(LazorGlobals.LoadDistance)), 4);

				GUILayout.BeginHorizontal(Util.Space_Compact);
				GUILayout.Label("View distance: ", Util.StdLabelSmallCustom(Color.green));
				GUILayout.Label(LazorGlobals.MAX_DISTANCE + " m");
				GUILayout.EndHorizontal();
			}

			settings_lazorsystem = GUILayout.Toggle(settings_lazorsystem, "Lazor System Settings", Util.Button_Toggle);
			if(settings_lazorsystem) {
				GUILayout.Space(5);
				LazorGlobals.GLOBALCONFIG_SyncTargets = GUILayout.Toggle(LazorGlobals.GLOBALCONFIG_SyncTargets, "Sync targeting");
				LazorGlobals.GLOBALCONFIG_ElectricGenerator = GUILayout.Toggle(LazorGlobals.GLOBALCONFIG_ElectricGenerator, "Electric Generator");
			}

			GUILayout.Space(5);
			if(GUILayout.Button("Save"))
				LazorGlobals.SaveConfig();
		}


		#endregion

		#region DEBUG

		private void GUIwindow_DEBUG(int windowID){
			DEBUG_gui_on = GUILayout.Toggle(DEBUG_gui_on, "Romfarer Debug");
			if(DEBUG_gui_on) {
				GUILayout.BeginHorizontal();
				DEBUG_hierarchy = GUILayout.Toggle(DEBUG_hierarchy, "Hierarchy");
				DEBUG_joints.set = GUILayout.Toggle(DEBUG_joints, "Joints");
				DEBUG_log = GUILayout.Toggle(DEBUG_log, "Log");
				GUILayout.FlexibleSpace();
				GUILayout.EndHorizontal();

				if(DEBUG_hierarchy) {
					scrollArea = GUILayout.BeginScrollView(scrollArea, false, false, GUILayout.MinHeight(800), GUILayout.ExpandHeight(false));
					Part p = FlightGlobals.ActiveVessel.rootPart;
					GUILayout.Label(p.name + " " + p.children.Count);
					Element("   |", p);
					GUILayout.EndScrollView();
				}

				if(DEBUG_joints.In()) {
					joints = UnityEngine.Object.FindObjectsOfType(typeof(Joint)).Cast<Joint>().ToList();
				} else if(DEBUG_joints.Out()) {}
				if(DEBUG_joints) {
					scrollArea = GUILayout.BeginScrollView(scrollArea, false, false, GUILayout.MinHeight(800), GUILayout.ExpandHeight(false));
					PrintJoints();
					GUILayout.EndScrollView();
				}

				if(DEBUG_log){
					GUILayout.BeginHorizontal();
					if(GUILayout.Button("clear log"))
						log = "Log:\n";
					if(GUILayout.Button("-----"))
						log += "-----------------------------\n";
					GUILayout.FlexibleSpace();
					GUILayout.EndHorizontal();

					scrollArea = GUILayout.BeginScrollView(scrollArea, false, false, GUILayout.MinHeight(800), GUILayout.ExpandHeight(false));
					GUILayout.Label(log);
					GUILayout.EndScrollView();
				}
				GUILayout.BeginHorizontal();
				GUILayout.BeginVertical();
				if(LazorSystemControllerGlobal.ActiveLazorSystem != null)
					LazorSystemControllerGlobal.ActiveLazorSystem.debugAssist.TestButtons();
				GUILayout.EndVertical();
				GUILayout.FlexibleSpace();
				GUILayout.EndHorizontal();
			}

		}

		private void Element(string a, Part p){
			using(List<Part>.Enumerator enumerator = p.children.GetEnumerator()) {
				while(enumerator.MoveNext()){
					GUILayout.Label(a+enumerator.Current.name + " (" + enumerator.Current.children.Count + ") " + 
						(enumerator.Current.attachJoint==null?"null":enumerator.Current.attachJoint.Target.partName));
				//GUILayout.Label(a+enumerator.Current.name + " (" + enumerator.Current.children.Count + ") " +  //XXX
				//	                (enumerator.Current.attachJoint==null?"null":enumerator.Current.attachJoint.connectedBody.name));
					Element(a + "   |", enumerator.Current);
				}
			}
		}

		private void PrintJoints(){
			joints.RemoveAll(x => x == null);
			joints.Sort(delegate(Joint a, Joint b){return (a.transform.position - FlightGlobals.ActiveVessel.rootPart.transform.position).magnitude.CompareTo(
					(b.transform.position - FlightGlobals.ActiveVessel.rootPart.transform.position).magnitude); });
			foreach(Joint j in joints){
				if(j != null)
					GUILayout.Label(j.name + " "+j.GetType()+" "+(j.transform.position - FlightGlobals.ActiveVessel.rootPart.transform.position).magnitude);
			}
		}

		private string log = "Log:\n";
		public void Log(string line, bool print){
			#if LAZORDEBUG
				log += line + "\n";
				if (print)
					MonoBehaviour.print (line);
			#endif
		}

		#endregion


		//launch clamps
		/*private bool launchClampsDetected;
		private bool launching;
		private Dictionary<Guid, List<Part>> destroyOnLiftoff = new Dictionary<Guid, List<Part>>();
		*/

		/*//Destroy launch clamps
		if(!launching && FlightGlobals.ActiveVessel.situation == Vessel.Situations.PRELAUNCH){
			launching = true;
			List<Part> clamps = new List<Part>();
			foreach (Part p in FlightGlobals.ActiveVessel.parts)
				if (p.Modules.Contains("LaunchClamp"))
					clamps.Add(p);
			if (clamps.Count > 0){
				launchClampsDetected = true;
				destroyOnLiftoff.Add (FlightGlobals.ActiveVessel.id, clamps);
			}
		}
		if(launchClampsDetected && FlightGlobals.ActiveVessel.situation == Vessel.Situations.FLYING){
			print("[LazorSystem : Destroying launch clamps]");
			foreach (Part p in destroyOnLiftoff[FlightGlobals.ActiveVessel.id])
				p.explode();
			destroyOnLiftoff.Clear();
			launchClampsDetected = false;
		}*/
	}
		
}

